#define __TPTP_SYNTAX_H

#ifndef __PR_ABSTRACT_SYNTAX_H
 #include "PR_PARSER/pr_abstract_syntax.h"
#endif

#ifndef __SIGNATURE_H
 #include "signature.h"
#endif

#ifndef __TERM_H
 #include "term.h"
#endif

#ifndef __CLAUSE_H
 #include "clause.h"
#endif

#ifndef __RESOLVENT_H
 #include "resolvent.h"
#endif

#ifndef __OUTPUT_H
 #include "output.h"
#endif

#include <string.h>

extern RESOLVENT Resolvent;
void RegisterInputClause(Clause* cl);
extern SIGNATURE Signature;

//*************************************************************************
enum TPTP_ITEM_SORT
 {
  TPTP_INCLUDE_DIRECTIVE,
  TPTP_INPUT_CLAUSE,
  TPTP_INPUT_FORMULA,
  EXT_TPTP_SET_OPTION_DIRECTIVE,
  EXT_TPTP_HEADER_WEIGHT,
  TPTP_UNCKNOWN_ITEM_SORT
 };

enum TPTP_ITEM_TYPE
 {
  TPTP_AXIOM,
  TPTP_CONJECTURE,
  TPTP_HYPOTHESIS,
  TPTP_WRONG_ITEM_TYPE
 };

inline TPTP_ITEM_SORT TPTPClassify(const char* const pref)
 {
  if (!(strcmp(pref,"include"))) { return TPTP_INCLUDE_DIRECTIVE; };
  if (!(strcmp(pref,"input_clause"))) { return TPTP_INPUT_CLAUSE; };
  if (!(strcmp(pref,"input_formula"))) { return TPTP_INPUT_FORMULA; };
  if (!(strcmp(pref,"set"))) { return EXT_TPTP_SET_OPTION_DIRECTIVE; };
  if (!(strcmp(pref,"header_weight"))) { return EXT_TPTP_HEADER_WEIGHT; };
  return TPTP_UNCKNOWN_ITEM_SORT;
 }; // TPTP_ITEM_SORT TPTPClassify(const char* const pref)

inline TPTP_ITEM_SORT TPTPClassify(const PR_LITERAL* const lit)
 {
  return TPTPClassify(lit->Header()->Symbol());
 }; // TPTP_ITEM_SORT TPTPClassify(const PR_LITERAL* const lit)

inline TPTP_ITEM_TYPE TPTPItemType(PR_TERM* it)
 {
  if (it->Tag() != PR_TERM::ATOM_FUNCTOR) { return TPTP_WRONG_ITEM_TYPE; };
  const char* str = it->Symbol()->Symbol(); 
  if (!(strcmp(str,"axiom"))) { return TPTP_AXIOM; };
  if (!(strcmp(str,"conjecture"))) { return TPTP_CONJECTURE; };
  if (!(strcmp(str,"hypothesis"))) { return TPTP_HYPOTHESIS; };
  return TPTP_WRONG_ITEM_TYPE;
 }; // TPTP_ITEM_TYPE TPTPItemType(PR_TERM* it)


class TPTP_SYNTAX
 {
  private:
   static inline int varNum(const char *v)
    {
     return Signature.VarNum(v);
    };
 
   static inline int funNum(const char *f)
    {
     return Signature.InputFunctorNum(f);
    };
   static TERM* CreateConstant(const char* str) 
    {
     return new(0) TERM(ARITY(0),(FUNCTOR)(funNum(str)),POL_POS);
    };
   static TERM* CreateComplexTerm(PR_TERM* input_term)
    {
     int arity;
     TERM* res;
     PR_TERM_LIST* input_args;
     TERM* tmp_term;
     switch (input_term->Tag())
      {
       case PR_TERM::ATOM_FUNCTOR : 
        res = CreateConstant(input_term->Symbol()->Symbol());
        break;
       case PR_TERM::ATOM_NUMBER : 
        res = CreateConstant(input_term->Symbol()->Symbol());
        break;
       case PR_TERM::COMPLEX : 
        input_args = input_term->Arguments();
        arity = PR_TERM_LIST::Length(input_args);
        res = new(arity) TERM(ARITY(arity),(FUNCTOR)funNum(input_term->Symbol()->Symbol()),POL_POS);
        for (TERM* arg = res->Args(); arg != res; arg = arg->Next())
         {
          switch (input_args->Term()->Tag())
           {
            case PR_TERM::ATOM_FUNCTOR : 
             arg->SetFirst(CreateConstant(input_args->Term()->Symbol()->Symbol()));
             break;
            case PR_TERM::ATOM_NUMBER :
             arg->SetFirst(CreateConstant(input_args->Term()->Symbol()->Symbol()));
             break;
            case PR_TERM::VAR :
             arg->MakeVar(varNum(input_args->Term()->Symbol()->Symbol()));
             break;
            case PR_TERM::COMPLEX :
             tmp_term = CreateComplexTerm(input_args->Term());
             if (!tmp_term) { return (TERM*)0; };
             arg->SetFirst(tmp_term);
             break;
            default : return (TERM*)0;
           };
          input_args = input_args->Next();
         };
        break;
       default : return (TERM*)0;
      };
     return res;
    };
   
   static inline bool CreateClause(PR_TERM_LIST* lit_list,TPTP_ITEM_TYPE cl_type)
    {
     ANCESTORS::BACKGR backgr = ANCESTORS::UNCKNOWN; // initialized to surpress warnings
     
     switch (cl_type)
      {
       case TPTP_AXIOM : backgr = ANCESTORS::AXIOM; break;
       case TPTP_CONJECTURE : backgr = ANCESTORS::CONJECTURE; break;
       case TPTP_HYPOTHESIS : backgr = ANCESTORS::CONJECTURE; break;
       case TPTP_WRONG_ITEM_TYPE : return false;
      };
     
     Resolvent.Reset(); 
 
     TERM* lit; 
     POLARITY polarity;
     while (lit_list)
      {
       if (lit_list->Term()->Tag() != PR_TERM::COMPLEX) { return false; };
       if (PR_TERM_LIST::Length(lit_list->Term()->Arguments()) != 1) { return false; };
       if (strcmp(lit_list->Term()->Symbol()->Symbol(),"++"))
        {
         if (strcmp(lit_list->Term()->Symbol()->Symbol(),"--"))
          {
           return false;
          }
         else
          {
           polarity = POL_NEG;
          };
        }
       else
        {
         polarity = POL_POS;
        }; 
       
       lit = CreateComplexTerm(lit_list->Term()->Arguments()->Term());
       
       if (!(lit))
        {
         return false;
        };
       
       lit->SetPolarity(polarity);
       
       //if (!(TmpClause.PushLiteral(lit))) 
       if (!(Resolvent.PushLiteral(lit))) 
        {
         cout << "WARNING: INPUT CLAUSE REJECTED.\n";
         return true; 
        };
       lit_list = lit_list->Next();
      };
     
     Signature.ResetVariables();
     
     Clause* res_cl = Resolvent.MakeInputClause(backgr);
     
     //cout << "RES CL\n       ";
     //Output(res_cl); cout << "\n";
     
     
     RegisterInputClause(res_cl);
     return true;
    };
  public:
   static inline bool CreateClause(PR_LITERAL* input_cl)
    {
     if (PR_TERM_LIST::Length(input_cl->Arguments()) != 3) { return false; };
     if (PR_TERM_LIST::Nth(input_cl->Arguments(),0)->Tag() != PR_TERM::LIST)
      {
       return false;
      };
     TPTP_ITEM_TYPE cl_type = TPTPItemType(PR_TERM_LIST::Nth(input_cl->Arguments(),1));
     if (cl_type == TPTP_WRONG_ITEM_TYPE) { return false; };
     return CreateClause(PR_TERM_LIST::Nth(input_cl->Arguments(),0)->Arguments(),cl_type);
    };
  static inline bool CreateHeader(TERM& header,PR_TERM* t)
   {
    if (t->Tag() != PR_TERM::COMPLEX) { return false; };
    if (PR_TERM_LIST::Length(t->Arguments()) != 1) { return false; };
    if (strcmp(t->Symbol()->Symbol(),"++"))
     {
      if (strcmp(t->Symbol()->Symbol(),"--"))
       {
        return false;
       }
      else
       {
        header = TERM((ARITY)0,(FUNCTOR)0,POL_NEG);
       };
     }
    else
     {
      header = TERM((ARITY)0,(FUNCTOR)0,POL_POS);
     };  
    PR_TERM* fun = t->Arguments()->Term();
    if (fun->Tag() != PR_TERM::ATOM_FUNCTOR) { return false; };
    header.SetFunctor(funNum(fun->Symbol()->Symbol()));
    return true;
   }; 
 }; // class TPTP_SYNTAX
 


//***********************************************************************
