#define __CLAUSE_H

#ifndef __DEBUG_H
 #include "debug.h"
#endif 
#ifndef __STDDEF_H
 #include <stddef.h>
#endif
#ifndef __MEMORY_H
 #include "memory.h"
#endif
#ifndef __TERM_H
 #include "term.h"
#endif
#ifndef __ANCESTORS_H
 #include "ancestors.h"
#endif 


struct TERM;


enum CLAUSE_STORE
	{
	CS_USED = 0,
	CS_KEPT = 1,
	CS_KEPT_SIMPLIFIER = 2,
        CS_KEPT_RESERVED = 3,
        CS_KEPT_SIMPLIFIER_RESERVED = 4,
	CS_TRASH_USED = 5,
	CS_TRASH_KEPT = 6,
	CS_TRASH_KEPT_SIMPLIFIER = 7,
	CS_HISTORY = 8,
	CS_HYPER_RULE = 9
	};


class Clause
{
  public: // memory
    inline void* operator new(size_t);
    inline void operator delete(void* obj);

  private:  // structure
    Clause* next;
    Clause* previous;
    TERM* literals;
    ANCESTORS* ancestors;
    long number;
    long weight;
    long min_res_weight;
    //long weight_of_nonmax;
    CLAUSE_STORE store;
    long num_of_pos_eq;
  public:  // structure
    inline Clause* Next() const
      { return next; }
    inline void SetNext(Clause* c)
      { next = c; }
    inline Clause* Previous() const
      { return previous; }
    inline void SetPrevious(Clause* p)
      { previous = p; }
    inline TERM* Literals() const
      { return literals; };
    inline ANCESTORS* Ancestors() const { return ancestors; };  
    inline void SetAncestors(ANCESTORS* a) { ancestors = a; };
    inline long Number() const
      { return number; }
    inline void SetNumber(long n)
      { number = n; }
    inline long Weight() const
     { 
      if (weight < 0) return ((-1)*weight);
      return weight;
     };
    
    inline void AssignMaxPriority() { if (weight > 0) weight = (-1)*weight; };
    inline bool MaxPriority() const { return (weight < 0); };    

    inline long NumOfPosEq() const { return num_of_pos_eq; };
    inline void SetNumOfPosEq(long n) { num_of_pos_eq = n; };
    
    inline long MinResWeight() const { return min_res_weight; };
    inline void SetMinResWeight(long mrw) { min_res_weight = mrw; };
    
    //inline long WeightOfNonmax() const { return weight_of_nonmax; };
    //inline void SetWeightOfNonmax(long wnm) { weight_of_nonmax = wnm; };
    
    
    inline long Length() const
      { return Literals()->Arity(); };
   
    inline void SetWeight(long w)
      { weight = w; }
    inline CLAUSE_STORE Store() const
      { return store; }
    inline bool Trash() const
     {
      return ((Store() == CS_TRASH_KEPT) 
              || (Store() == CS_TRASH_USED)
              || (Store() == CS_TRASH_KEPT_SIMPLIFIER));
     };
    inline bool Active()
     {
      return ((Store() == CS_KEPT) 
              || (Store() == CS_USED)
              || (Store() == CS_KEPT_SIMPLIFIER));
     };

    inline void MakeTrash()
     {
      switch (Store())
       {
        case CS_KEPT: SetStore(CS_TRASH_KEPT); return;
        case CS_KEPT_SIMPLIFIER: SetStore(CS_TRASH_KEPT_SIMPLIFIER); return;
        case CS_KEPT_RESERVED: SetStore(CS_TRASH_KEPT); return;
        case CS_KEPT_SIMPLIFIER_RESERVED: SetStore(CS_TRASH_KEPT_SIMPLIFIER); return;
        case CS_USED: SetStore(CS_TRASH_USED); return; 
        #ifdef _DEBUG_
         default: DebugError("A0 in Clause::MakeTrash()"); return; 
        #else
         #ifdef _SUPRESS_WARNINGS_
          default: return; 
         #endif 
        #endif
       };
     };
   
    inline void SetStore(CLAUSE_STORE s)
     { store = s; }
    inline void MakeUsed() { store = CS_USED; };  
    inline void MarkAsSimplifier() 
     {
      switch (Store())
       {
        case CS_KEPT: SetStore(CS_KEPT_SIMPLIFIER); return;
        case CS_KEPT_SIMPLIFIER: return;
        case CS_KEPT_RESERVED: SetStore(CS_KEPT_SIMPLIFIER_RESERVED); return;
        case CS_KEPT_SIMPLIFIER_RESERVED: return;
        case CS_USED: return;
        case CS_TRASH_KEPT: SetStore(CS_TRASH_KEPT_SIMPLIFIER); return;
        case CS_TRASH_KEPT_SIMPLIFIER: return;
        case CS_TRASH_USED: return;
        case CS_HISTORY: return;
        case CS_HYPER_RULE: 
         #ifdef _DEBUG_
          DebugError("A0 in Clause::MarkAsSimplifier()");
         #endif
         return;
       };
     };
    inline int NumberOfLiteral(TERM* lit)
     {
      int i = 0;
      for (TERM* l = literals->Args();;l = l->Next())
       {
        if (l->First() == lit)
         {
          return i;
         };
        i++;
       };
     };
    inline bool Empty() const { return (!(Length())); };
    inline bool Unit() const { return (Length() == 1); };
    inline bool IsAChildOf(const Clause* c)
     {
      return Ancestors()->OneOfAncestors(c);
     };
    
    inline bool HornClause() const
     {
      if (Unit()) { return true; };
      int num_of_pos = 0;
      for (TERM* lit = Literals()->Args(); !(lit->IsNull()); lit = lit->Next())
       {
        if (lit->First()->Polarity() == POL_POS) 
         {
          num_of_pos++;
          if (num_of_pos > 1) { return false; };
         };
       };
      return true;
     };
   
    inline bool ContainsEliminated(bool* eliminated_pred)
     {
      for (TERM* lit_ref = Literals()->Args(); !lit_ref->IsNull(); lit_ref = lit_ref->Next())
       if (eliminated_pred[lit_ref->First()->Functor()]) return true;
      return false;
     }; 
   
  public:  // constructors

    Clause(TERM* ts) : literals (ts), num_of_pos_eq(0) {};

    // for making dummy clauses in the databases
    Clause(long w,CLAUSE_STORE s) :
      next (this),
      previous (this),
      weight (w),
      store (s),
      num_of_pos_eq(0)
      {};
     
    // for making arrays of dummy clauses in KEPT 
    Clause() : next(this), previous(this) {};

  public:
    inline void Destroy();
    static inline void InsertAfter(Clause* c,Clause* db);
    static inline void InsertBefore(Clause* c,Clause* db);
    static inline void Remove(Clause* c);
    inline int IsEmptyDb() const
      { return next == this; }
    //Clause* DBNth(unsigned long n);
    static inline Clause* MakeEmptyClause();
    /**********
    inline Clause* CopyTopLevel() 
     {
      Clause* res =  new Clause(); 
      res->SetNumber(number);
      res->SetWeight(weight);
      res->SetAncestors(ancestors);
      res->literals = literals->CopyTopLevel();
      return res; 
     };
    ***********/ 
    /******* 
    inline Clause* Copy()
     {
      Clause* res = CopyTopLevel();
      for (TERM* lit = res->Literals()->Args(); !(lit->IsNull()); lit = lit->Next())
       {
        lit->SetFirst(lit->First()->CopyRecursively());
       }; 
      return res;     
     };
   **********/  
};

inline void* Clause::operator new(size_t)
 {
  return (Clause*)Memory::Allocate(sizeof(Clause)/4);
 };


inline void Clause::operator delete(void* obj)
 {
  Memory::Deallocate(obj,sizeof(Clause)/4);
 };

inline void Clause::InsertAfter(Clause* c,Clause* db)
 {
  Clause* next_cl = db->next;
  c->next = next_cl;
  next_cl->previous = c;
  db->next = c;
  c->previous = db;
 }  // Clause::InsertAfter()


inline void Clause::InsertBefore(Clause* c,Clause* db)
 {
  Clause* previous_cl = db->previous;
  previous_cl->next = c;
  c->previous = previous_cl;
  c->next = db;
  db->previous = c;
 }  // Clause::InsertBefore()

inline void Clause::Destroy()
 {
  //CHECK0(Store() == CS_TRASH_KEPT,"A0 in Clause::Destroy()");
  
  delete literals;
  //parents->Delete(); // Corrected.
  delete ancestors;
  delete this;
 };  // Clause::Destroy()

inline void Clause::Remove(Clause* c)
 {
  Clause* previous_cl = c->previous;
  Clause* next_cl = c->next;

  previous_cl->next = next_cl;
  next_cl->previous = previous_cl;
 };  // Clause::Remove()

inline Clause* Clause::MakeEmptyClause()
 {
  return new Clause(new(0) TERM(nil));
 };


//******************* Lists representing sets of clauses: *****************

class CLAUSE_LIST
 {
  public: // memory
   inline void* operator new(size_t)
    {
     return Memory::Allocate(sizeof(CLAUSE_LIST)/4);
    };

   inline void operator delete(void* obj)
    {
     Memory::Deallocate(obj,sizeof(CLAUSE_LIST)/4);
    };

  private:  // structure
   Clause* first;
   CLAUSE_LIST* next;

  public:
   CLAUSE_LIST(Clause* c,CLAUSE_LIST* nxt) : first(c), next(nxt) {};
   inline void SetNext(CLAUSE_LIST* nxt) { next = nxt; };
   inline CLAUSE_LIST* Next() const { return next; };
   inline Clause* First() const { return first; };
   inline bool ThisClause(const Clause* c) const {  return first == c; };
  
   /****
   static inline void Remove(CLAUSE_LIST** l,Clause* c)
    {
     while (*l)
      {
       if ((*l)->first == c)
        {
         CLAUSE_LIST* tmp = (*l)->next;
         delete (*l);
         *l = tmp;
         return;
        };
       l = &((*l)->next);
      };
    }; 
   ********/
    
   static int Length(CLAUSE_LIST* l) // for debugging
    {
     if (l)   
      {
       return (1 + Length(l->Next()));
      } 
     else
      {
       return 0;
      };
    };
   /****** 
   static inline void DestroyList(CLAUSE_LIST*& lst)
    {
     CLAUSE_LIST* tmp;
     while (lst)
      {
       tmp = lst;
       lst = lst->Next();
       delete tmp; 
      };     
    };
   static inline void Destroy(CLAUSE_LIST* lst) { DestroyList(lst); };
   ********/
   
   static inline void OutputList(CLAUSE_LIST* lst)
    {
     while (lst)
      {
       Output(lst->First());
       Output("\n");
       lst = lst->Next();
      };
    };
 };  // class CLAUSE_LIST
 
/******** 
class ORD_CLAUSE_LIST 
 {
  // ordered by pointers, not by numbers
  private:
   Clause* clause;
   ORD_CLAUSE_LIST* next;
  public:
   ORD_CLAUSE_LIST(Clause* const cl,ORD_CLAUSE_LIST* const lst)
    {
     clause = cl;
     next = lst;
    };
   inline void* operator new(size_t)
    {
     return Memory::Allocate(sizeof(ORD_CLAUSE_LIST)/4);
    };
   inline void operator delete(void* obj)
    {
     Memory::Deallocate(obj,sizeof(ORD_CLAUSE_LIST)/4);
    };

   inline Clause* First() const { return clause; };
   inline ORD_CLAUSE_LIST*& Next() { return next; }; 
   static inline void Insert(Clause* const cl,ORD_CLAUSE_LIST** lst_addr)
    {
     compare:
      if ((*lst_addr) && ((*lst_addr)->First() > cl))
       {
        lst_addr = &((*lst_addr)->Next());
        goto compare;
       }
      else
       {
        *lst_addr = new ORD_CLAUSE_LIST(cl,*lst_addr);
       };
    };
   static inline void Remove(Clause* const cl,ORD_CLAUSE_LIST** lst_addr)
    {
     compare:
      if (*lst_addr)
       {
        if ((*lst_addr)->First() > cl)
         {
          lst_addr = &((*lst_addr)->Next());
          goto compare;
         }
        else
         {
          if ((*lst_addr)->First() == cl)
           {
            ORD_CLAUSE_LIST* cons_to_delete = *lst_addr;
            *lst_addr = cons_to_delete->Next();
            delete cons_to_delete;
           };
         };
       };
    };
   static inline void Destroy(ORD_CLAUSE_LIST* lst) 
    {
     while (lst)
      {
       ORD_CLAUSE_LIST* tmp = lst;
       lst = lst->Next();
       delete tmp;
      };
    }; 
 }; // class ORD_CLAUSE_LIST

*************/

#ifndef __FIFO_H
 #include "fifo.h"
#endif

class CLAUSE_QUEUE
 { 
  private:
   FIFO<Clause*> queue;
  public:
   CLAUSE_QUEUE() : queue() {};
   inline void Enqueue(Clause* cl) { queue.Enqueue(cl); };
   inline bool Dequeue(Clause*& cl) { return queue.Dequeue(cl); };
 }; // class CLAUSE_QUEUE 


//****************************************************************************

class INFERENCE
 {
  private:
   static STACK_WITH_ITERATOR<Clause*,MAX_FINAL_PROOF_SIZE> allClauses;
   
  private:
   static inline bool PushIfAbsent(Clause* cl)
    {
     allClauses.ResetIterator();
     Clause** c = allClauses.Next();
     while (c)
      {
       if ((*c) == cl) { return false; };
       c = allClauses.Next();
      };
     allClauses.Push(cl);
     return true;
    };
   static void CollectSubinference(Clause* root)
    {
     if (PushIfAbsent(root))
      {
       ANCESTORS* ancestors = root->Ancestors();
       if (ancestors->Number()) // not an input clause
        {
         int numOfAncestors = ancestors->Number();
         Clause** c = ancestors->Clauses();
         while (numOfAncestors)
          {
           CollectSubinference(*c);
           c++;
           numOfAncestors--;
          };
        };
      };
    };
  public:
   static inline void CollectInference(Clause* root)
    {
     allClauses.Reset();
     CollectSubinference(root);
    };
   static inline void ResetIterator() { allClauses.ResetIterator(); };
   static inline Clause** Next() { return allClauses.Next(); };
 }; // class INFERENCE

//****************************************************************************

