#define __DTREE_H

#ifndef __VAMPIRE_H
 #include "vampire.h"
#endif

#ifndef __DEBUG_H
 #include "debug.h"
#endif
 
#ifndef __TERM_H
 #include "term.h"
#endif

//***************************************************************************

class DTREE_TAG
 {
  private:
   TERM symbol;
  public:
   DTREE_TAG() {};
   DTREE_TAG(TERM* p) { symbol.SetFirst(p); }; // for leaves
   inline DTREE_TAG& operator= (TERM* p) // for leaves
    {
     symbol.SetFirst(p);
     return *this; 
    }; 
   
   DTREE_TAG(const TERM& sym,const unsigned depth) // for intermediate nodes
    : symbol(sym) 
    { symbol.SetReserved(depth); };
    
   inline DTREE_TAG& operator= (const DTREE_TAG& tag) // // for intermediate nodes
    {
     symbol = tag.symbol;
     return *this;
    };
    
   inline bool IsLeaf() const { return symbol.IsReference(); };
   inline unsigned Depth() const { return symbol.Reserved(); };
   inline bool Equal(const TERM& sym) const // for intermediate nodes only
    {
     return ((symbol.ComplexFunctor() == sym.ComplexFunctor()) 
             && (symbol.Flag() == sym.Flag()));
    };
   inline bool Greater(const TERM& sym) const // for intermediate nodes only
    {
     return ((symbol.Flag() > sym.Flag())
             || ((symbol.Flag() == sym.Flag()) 
                 && (symbol.ComplexFunctor() > sym.ComplexFunctor())));
    };
   
   inline bool Equal(const DTREE_TAG& tag) const // for intermediate nodes only
    {
     return Equal(tag.symbol);
    };
   inline bool Greater(const DTREE_TAG& tag) const // for intermediate nodes only
    {
     return Greater(tag.symbol);
    }; 
    
   inline TERM* Term() const { return symbol.First(); }; // for leaves
   inline void SetTerm(TERM* t) { symbol.SetFirst(t); }; // for leaves
   inline void Output() 
    {
     if (symbol.IsVariable()) { OutputVar(symbol.Var()); }
     else { OutputFunctor(symbol.Functor()); };
    };
 }; // class DTREE_TAG

class DTREE_NODE
 {
  private:
   DTREE_TAG tag;
   DTREE_NODE* next;
   DTREE_NODE* less;
   DTREE_NODE* greater;
   DTREE_NODE() {};
  public:
   DTREE_NODE(TERM* p) : tag(p), next((DTREE_NODE*)0) {}; // for making leaves
   DTREE_NODE(const TERM& sym,const unsigned depth) // for intermediate nodes
    : tag(sym,depth), 
      less((DTREE_NODE*)0), 
      greater((DTREE_NODE*)0)
    {};
   
   inline void Copy(const DTREE_NODE& node) // for intermediate nodes
    {
     tag = node.tag;
     next = node.next;
     less = node.less;
     greater = node.greater;
    };
    
   inline void* operator new(size_t)
    {
     return Memory::Allocate(sizeof(DTREE_NODE)/4);
    };
   inline void operator delete(void* obj)
    {
     Memory::Deallocate(obj,sizeof(DTREE_NODE)/4);
    };
   inline DTREE_TAG& Tag() { return tag; };
   inline TERM* Term() const { return tag.Term(); }; // for leaves
   inline void SetTerm(TERM* t) { tag.SetTerm(t); };
   inline bool IsLeaf() { return tag.IsLeaf(); };
   inline unsigned Depth() const { return tag.Depth(); }; // for intermediate
   inline DTREE_NODE*& Next() { return next; }; // for intermediate 
   //inline ORD_CLAUSE_LIST*& Clauses() { return (ORD_CLAUSE_LIST*&)next; }; // for leaves
   inline long& NumOfClauses() { return (long&)next; }; // for leaves
   inline DTREE_NODE*& Less() { return less; }; // for intermediate
   inline DTREE_NODE*& Greater() { return greater; }; // for intermediate
   
   inline void*& Reserved1() { return (void*&)less; };
   inline void*& Reserved2() { return (void*&)greater; };
   
   inline bool Fork() { return (Less() || Greater()); }; 
   inline void Output() 
    {
     ::Output("["); ::Output((long)this); ::Output("] ");
     if (IsLeaf()) 
      {
       ::Output("LEAF "); ::Output(Term());
      }
     else
      {
       Tag().Output();
       ::Output(" N["); ::Output((long)Next()); ::Output("] ");   
       ::Output("L["); ::Output((long)Less()); ::Output("] ");
       ::Output("G["); ::Output((long)Greater()); ::Output("] ");
      };
    };
 }; // class DTREE_NODE

class DTREE
 {
  private:
   static inline DTREE_NODE* Minimal(DTREE_NODE* node);
   static inline DTREE_NODE* Merge(DTREE_NODE* less,DTREE_NODE* greater);
  public:
   static inline DTREE_NODE** Track(const TERM& sym,DTREE_NODE** node);
   static inline DTREE_NODE* Track(const TERM& sym,DTREE_NODE* node);
   static inline DTREE_NODE* Track(const DTREE_TAG& tag,DTREE_NODE* node);
   static inline DTREE_NODE** Track(TERM::ITERATOR_WITH_DEPTH& trm,DTREE_NODE** node);
   static inline void InfoForDeletion(TERM::ITERATOR& trm,
                                      DTREE_NODE** tree,
                                      DTREE_NODE**& branch_to_destroy,
                                      DTREE_NODE*& leaf,
                                      DTREE_NODE*& last_fork);
   static inline DTREE_NODE* CreateBranch(TERM::ITERATOR_WITH_DEPTH& trm,
                                          DTREE_NODE*& leaf);
   static inline void DestroyBranchWithoutForks(DTREE_NODE* branch);
   static inline void ReduceFork(DTREE_NODE* fork);
   static inline void Destroy(DTREE_NODE* tree,void (*DestroyLeafContent)(DTREE_NODE* leaf));
  public: 
   class SURFER; 
  public: // for debugging
   static bool BranchWithNoForks(DTREE_NODE* branch)
    {
     // branch must be nonnull here
     CHECK2(branch,"A1 in DTREE::BranchWithNoForks()");
     if (branch->IsLeaf()) { return true; };
     if ((branch->Less()) || (branch->Greater())) { return false; };
     return BranchWithNoForks(branch->Next());
    };
 }; // class DTREE

class DTREE::SURFER
 {
  private:
   DTREE_NODE** node;
   SMART_DEPTH_COUNTER depth_counter;
  public:
   SURFER() : depth_counter() {};
   inline void Reset(DTREE_NODE** tree,unsigned init_depth)
    {
     node = tree;
     depth_counter.Reset(init_depth);
    };
   inline bool Read(const TERM& sym)
    {
     depth_counter.NextSym(sym);     
     node = DTREE::Track(sym,node);
     if (*node)
      {
       node = &((*node)->Next());
       return true;
      }
     else return false;
    }; 
   inline void Write(const TERM& sym)
    {
     *node = new DTREE_NODE(sym,depth_counter.Depth());
     node = &((*node)->Next());
     depth_counter.NextSym(sym);
    }; 
   inline void CreateLeaf(TERM* t)
    {
     *node = new DTREE_NODE(t);
    };
   inline DTREE_NODE* Leaf() { return *node; };
  
 }; // class DTREE::SURFER

//*************** Definitions for DTREE: *****************************

inline DTREE_NODE* DTREE::Minimal(DTREE_NODE* node)
 {
  // node->Less() must be nonnull here
  CHECK2(node->Less(),"A1 in DTREE::Minimal(DTREE_NODE* node)");
  next_less:
   node = node->Less();
   if (node->Less()) { goto next_less; }
   else { return node; };
 }; // DTREE_NODE* DTREE::Minimal(DTREE_NODE* node)

inline DTREE_NODE* DTREE::Merge(DTREE_NODE* less,DTREE_NODE* greater)
 {
  CHECK2(less,"A1 in DTREE::Merge(..)");
  CHECK2(greater,"A2 in DTREE::Merge(..)"); 
  if (less->Greater())
   {
    if (greater->Less())
     {
      Minimal(greater)->Less() = less;
      return greater;
     }
    else
     {
      greater->Less() = less;
      return greater;
     };
   }
  else
   {
    less->Greater() = greater;
    return less;
   };
 }; // DTREE_NODE* DTREE::Merge(DTREE_NODE* less,DTREE_NODE* greater)

inline void DTREE::ReduceFork(DTREE_NODE* fork)
 {
  // either fork->Less() or fork->Greater() must be nonnull here
  CHECK2(fork,"A1 in DTREE::ReduceFork(..)"); 
  CHECK2(fork->Fork(),"A2 in DTREE::ReduceFork(..)");
  
  DTREE_NODE* node_to_delete;
  if (fork->Less())
   {
    if (fork->Greater())
     {
      node_to_delete = Merge(fork->Less(),fork->Greater());
      fork->Copy(*node_to_delete);
     }
    else
     {
      node_to_delete = fork->Less();
      fork->Copy(*(fork->Less()));
     };
   }
  else 
   {
    node_to_delete = fork->Greater();
    fork->Copy(*(fork->Greater()));
   };
  delete node_to_delete;
 }; // void DTREE::ReduceFork(DTREE_NODE* fork) 

inline DTREE_NODE** DTREE::Track(const TERM& sym,DTREE_NODE** node)
 {
  check_node:
   if (*node)
    {
     if ((*node)->Tag().Equal(sym)) { return node; };
     if ((*node)->Tag().Greater(sym))
      {
       node = &((*node)->Less());
      }
     else
      {
       node = &((*node)->Greater());
      }; 
     goto check_node; 
    }
   else { return node; };
 }; // DTREE_NODE** DTREE::Track(const TERM& sym,DTREE_NODE** node)

inline DTREE_NODE* DTREE::Track(const TERM& sym,DTREE_NODE* node)
 {
  check_node:
   if (node)
    {
     if (node->Tag().Equal(sym)) { return node; };
     if (node->Tag().Greater(sym))
      {
       node = node->Less();
      }
     else
      {
       node = node->Greater();
      }; 
     goto check_node; 
    }
   else { return (DTREE_NODE*)0; };
 }; // DTREE_NODE* DTREE::Track(const TERM& sym,DTREE_NODE* node)

inline DTREE_NODE* DTREE::Track(const DTREE_TAG& tag,DTREE_NODE* node)
 {
  check_node:
   if (node)
    {
     if (node->Tag().Equal(tag)) { return node; };
     if (node->Tag().Greater(tag))
      {
       node = node->Less();
      }
     else
      {
       node = node->Greater();
      }; 
     goto check_node; 
    }
   else { return (DTREE_NODE*)0; };
 }; // DTREE_NODE* DTREE::Track(const DTREE_TAG& tag,DTREE_NODE* node)

inline DTREE_NODE** DTREE::Track(TERM::ITERATOR_WITH_DEPTH& trm,DTREE_NODE** node)
 {
  // trm must be nonempty here 
  track_sym:
   node = Track(trm.Symbol(),node);
   if (*node)
    {
     node = &((*node)->Next());
     if (trm.Next())
      {
       // *node must be an intermediate node here
       CHECK2(!((*node)->IsLeaf()),"A1 in DTREE::Track(TERM::ITERATOR_WITH_DEPTH& trm,..)");
       goto track_sym;
      }
     else 
      {
       // *node must be a leaf here
       CHECK2((*node)->IsLeaf(),"A2 in DTREE::Track(TERM::ITERATOR_WITH_DEPTH& trm,..)");
       return node; 
      }; 
    }
   else { return node; };
 }; // DTREE_NODE** DTREE::Track(TERM::ITERATOR_WITH_DEPTH& trm,DTREE_NODE** node)
 
inline DTREE_NODE* DTREE::CreateBranch(TERM::ITERATOR_WITH_DEPTH& trm,
                                       DTREE_NODE*& leaf)
 {
  // trm must be nonempty here 
  DTREE_NODE* branch = new DTREE_NODE(trm.Symbol(),trm.Depth());
  DTREE_NODE* last_node = branch;
  next_sym:
   if (trm.Next())
    {
     last_node->Next() = new DTREE_NODE(trm.Symbol(),trm.Depth());
     last_node = last_node->Next();
     goto next_sym;
    }
   else
    {
     leaf = new DTREE_NODE((TERM*)0);
     last_node->Next() = leaf;
    };   
  return branch;
 }; // DTREE_NODE* DTREE::CreateBranch(TERM::ITERATOR_WITH_DEPTH& trm,DTREE_NODE*& leaf)                                         

inline void DTREE::DestroyBranchWithoutForks(DTREE_NODE* branch)
 {
  // branch must be nonnul here
  CHECK2(branch,"A1 in DTREE::DestroyBranchWithoutForks(..)");
  DTREE_NODE* node_to_delete;
  next_node:
   if (branch->IsLeaf()) { delete branch; }
   else
    {
     node_to_delete = branch;
     branch = branch->Next();
     delete node_to_delete;
     goto next_node;
    };   
 }; // void DTREE::DestroyBranchWithoutForks(DTREE_NODE* branch)

inline void DTREE::Destroy(DTREE_NODE* tree,void (*DestroyLeafContent)(DTREE_NODE* leaf))
 {
  DCONTROL2("BEGIN in DTREE::Destroy(DTREE_NODE* tree,void (*DestroyLeafContent)(DTREE_NODE* leaf))");
  if (tree)
   {    
    if (tree->IsLeaf()) 
     {
      DCONTROL2("C10 in DTREE::Destroy(DTREE_NODE* tree,void (*DestroyLeafContent)(DTREE_NODE* leaf))");
      DestroyLeafContent(tree);
      DCONTROL2("C20 in DTREE::Destroy(DTREE_NODE* tree,void (*DestroyLeafContent)(DTREE_NODE* leaf))");
      delete tree;
      DCONTROL2("END1 in DTREE::Destroy(DTREE_NODE* tree,void (*DestroyLeafContent)(DTREE_NODE* leaf))");
     }
    else
     {
      DCONTROL2("C30 in DTREE::Destroy(DTREE_NODE* tree,void (*DestroyLeafContent)(DTREE_NODE* leaf))");
      if (tree->Less()) Destroy(tree->Less(),DestroyLeafContent);
      DCONTROL2("C40 in DTREE::Destroy(DTREE_NODE* tree,void (*DestroyLeafContent)(DTREE_NODE* leaf))");
      if (tree->Greater()) Destroy(tree->Greater(),DestroyLeafContent);
      DCONTROL2("C50 in DTREE::Destroy(DTREE_NODE* tree,void (*DestroyLeafContent)(DTREE_NODE* leaf))");
      if (tree->Next()) Destroy(tree->Next(),DestroyLeafContent);
      DCONTROL2("C60 in DTREE::Destroy(DTREE_NODE* tree,void (*DestroyLeafContent)(DTREE_NODE* leaf))");
      delete tree;
      DCONTROL2("END2 in DTREE::Destroy(DTREE_NODE* tree,void (*DestroyLeafContent)(DTREE_NODE* leaf))");
     };
   };
  DCONTROL2("END3 in DTREE::Destroy(DTREE_NODE* tree,void (*DestroyLeafContent)(DTREE_NODE* leaf))");
 }; // void DTREE::Destroy(DTREE_NODE* tree,void (*DestroyLeafContent)(DTREE* leaf))

inline void DTREE::InfoForDeletion(TERM::ITERATOR& trm,
                                   DTREE_NODE** tree,
                                   DTREE_NODE**& branch_to_destroy,
                                   DTREE_NODE*& leaf,
                                   DTREE_NODE*& last_fork)
 {
  // *tree must be nonnull here and not a leaf
  
  CHECK2(*tree,"A1 in DTREE::InfoForDeletion(..)");
  CHECK2(!((*tree)->IsLeaf()),"A2 in DTREE::InfoForDeletion(..)");
  
  // trm must be nonempty and contain all arguments of an indexed term
  
  last_fork = (DTREE_NODE*)0; 
  branch_to_destroy = tree;
  DTREE_NODE** current_branch;
  DTREE_NODE* current_node;
  
  track_sym:
   current_branch = Track(trm.Symbol(),tree);
   current_node = *current_branch;
  
   // current_node is nonnull here and contains symbol equal to trm.Symbol()
   CHECK2(current_node,"A3 in DTREE::InfoForDeletion(..)");
   CHECK2(!(current_node->IsLeaf()),"A4 in DTREE::InfoForDeletion(..)");
   CHECK2(current_node->Tag().Equal(trm.Symbol()),"A5 in DTREE::InfoForDeletion(..)");
   
   if (current_node->Fork())  
    {
     branch_to_destroy = &(current_node->Next());
     last_fork = current_node;
    }
   else
    {
     if (current_branch != tree) 
      { 
       // this means that *tree is marked
       // by a symbol different from trm.Symbol()
       CHECK2(!((*tree)->Tag().Equal(trm.Symbol())),"A6 in DTREE::InfoForDeletion(..)");
       branch_to_destroy = current_branch;
       last_fork = (DTREE_NODE*)0;
      }
     else
      {
       // the node *tree is marked by trm.Symbol()
       // nothing to do here
       CHECK2((*tree)->Tag().Equal(trm.Symbol()),"A7 in DTREE::InfoForDeletion(..)");
      };
    };
  
   tree = &(current_node->Next());
 
   CHECK2(*tree,"A8 in DTREE::InfoForDeletion(..)");

   if (trm.Next()) { goto track_sym; };
   
   leaf = *tree;
   
   // postcondition: leaf is a leaf,
   CHECK2(leaf->IsLeaf(),"A9 in DTREE::InfoForDeletion(..)");
 
   // *branch_to_destroy is nonnull and not a leaf
   // *branch_to_destroy is the beginning of a branch with no forks
   CHECK2(*branch_to_destroy,"A10 in DTREE::InfoForDeletion(..)");
   CHECK2(BranchWithNoForks(*branch_to_destroy),"A11 in DTREE::InfoForDeletion(..)");
 
   // if last_fork is nonnull:
   //  1) last_fork is an intermediate node
   //  2) last_fork is a fork
   //  3) branch_to_destroy == &(last_fork->Next())
   CHECK2((!last_fork) || (!(last_fork->IsLeaf())),"A12 in DTREE::InfoForDeletion(..)");
   CHECK2((!last_fork) || (last_fork->Fork()),"A13 in DTREE::InfoForDeletion(..)");
   CHECK2((!last_fork) || (branch_to_destroy == &(last_fork->Next())),"A14 in DTREE::InfoForDeletion(..)");  

   // if (*branch_to_destroy)->IsLeaf(), then 
   // last_fork is nonnull and &(last_fork->Next()) == branch_to_destroy
   CHECK2((*branch_to_destroy)->IsLeaf() ? (bool)last_fork : true,"A15 in DTREE::InfoForDeletion(..)");
   CHECK2((*branch_to_destroy)->IsLeaf() ? (&(last_fork->Next()) == branch_to_destroy) : true,"A16 in DTREE::InfoForDeletion(..)");

 }; // void DTREE::InfoForDeletion(TERM::ITERATOR& trm,...


//***********************************************************************















