#define __FLAT_TERM_H
#ifndef __DEBUG_H
 #include "debug.h"
#endif 
#ifndef __VAMPIRE_H
 #include "vampire.h"
#endif 
#ifndef __OUTPUT_H
 #include "output.h"
#endif 
#ifndef __TERM_H
 #include "term.h"
#endif 
#ifndef __VARIABLES_H
 #include "variables.h"
#endif  

//********************************************************************************

class FLAT_TERM    
 {
  private:
   //TERM symbol;
   const FLAT_TERM* next_term;
   TERM symbol;
  public:
   FLAT_TERM() {};
   inline TERM Symbol() const { return symbol; };
   inline TERM& SymbolRef() { return symbol; };
   inline void SetSymbol(const TERM& sym) { symbol = sym; };
   inline void SetVar(int var) { symbol.TERM(TERM_VAR,var); next_term = this + 1; };
   inline void SetVar(const TERM& var) { symbol = var; next_term = this + 1; };
   inline void SetConst(const TERM& csym) { symbol = csym; next_term = this + 1; }; 
   inline const FLAT_TERM* NextTerm() const { return next_term; };
   inline void SetNextTerm(const FLAT_TERM* nt) { next_term = nt; };
   inline long Weight() const { return (next_term - this); };
   inline unsigned Depth();
   inline bool Equal(const FLAT_TERM* t) const;
   inline bool EqualArguments(const FLAT_TERM* t) const;
   inline bool Complementary(const FLAT_TERM* t) const;
   inline COMPARISON CompareLex(const FLAT_TERM* t) const;
   static inline COMPARISON KBCompare(const FLAT_TERM* t1,VAR_MULTISET* vars1,const FLAT_TERM* t2,VAR_MULTISET* vars2);
   inline void RenameVariables(INT_VAR_RENAMING& var_renaming);
   inline TERM StandardRepCopy() const;
   inline bool IsBackJump() const { return next_term < this; };
   inline void SetBackJump(const FLAT_TERM* jmp) { next_term = jmp; };
  public: // iterators
   class ITER;
   class OPEN;
   class OPEN_WITH_MEMORY;
  public: // methods for debugging
   inline bool CheckIntegrity() { return (NextTerm() > this); };
   inline void OutputSymbol() const; 
   inline void OutputTerm() const;
   inline void OutputStructure() const;
 }; // class FLAT_TERM

class FLAT_TERM::OPEN
 {
  private:
   FLAT_TERM* curr_pos;
   STACK<FLAT_TERM*,MAX_TERM_WEIGHT> open_parentheses;
   BRACKET_COUNTER bracket_counter;
   DEPTH_COUNTER depth_counter;
   #ifdef _DEBUG9_
    FLAT_TERM* debug_fst_pos;
   #endif 
  private: // auxilliary methods
   inline void CloseNParentheses(unsigned n)
    {
     depth_counter.Up(n);
     while (n)
      {
       open_parentheses.Pop()->SetNextTerm(curr_pos);
       n--;
      };
    }; 
  public:
   OPEN() 
    : open_parentheses("MAX_TERM_WEIGHT in FLAT_TERM::OPEN::open_parentheses"),
      bracket_counter(),
      depth_counter()
     {};
   OPEN(FLAT_TERM* mem) 
    : open_parentheses("MAX_TERM_WEIGHT in FLAT_TERM::OPEN::open_parentheses"),
      bracket_counter()
     { 
      Reset(mem);
     };
   inline void Reset(FLAT_TERM* mem) 
    {
     #ifdef _DEBUG9_
      debug_fst_pos = mem;
     #endif 
     curr_pos = mem; 
     open_parentheses.Reset();
     bracket_counter.Reset();
     depth_counter.Reset(0);
    };

   inline unsigned Depth() const { return depth_counter.Depth(); };
   inline void PushVar(int var)
    {
     CHECK9(curr_pos - debug_fst_pos < MAX_TERM_WEIGHT,"A0 in FLAT_TERM::OPEN::PushVar(int var)");
     curr_pos->SetVar(var);
     curr_pos++;
     CloseNParentheses(bracket_counter.Close());
    }; 
    
   inline void PushVar(const TERM& var)
    {
     CHECK9(curr_pos - debug_fst_pos < MAX_TERM_WEIGHT,"A0 in FLAT_TERM::OPEN::PushVar(TERM var)");
     curr_pos->SetVar(var);
     curr_pos++;
     CloseNParentheses(bracket_counter.Close());
    };
   inline void PushConst(const TERM& csym)
    {
     CHECK9(curr_pos - debug_fst_pos < MAX_TERM_WEIGHT,"A0 in FLAT_TERM::OPEN::PushConst(TERM csym)");  
     curr_pos->SetConst(csym);
     curr_pos++;
     CloseNParentheses(bracket_counter.Close());
    }; 
   inline void PushNonconstFun(const TERM& fsym)
    {
     CHECK9(curr_pos - debug_fst_pos < MAX_TERM_WEIGHT,"A0 in FLAT_TERM::OPEN::PushNonconstFun(TERM fsym)");  
     curr_pos->SetSymbol(fsym);
     // fsym.Arity() > 0 here
     bracket_counter.Open(fsym.Arity());
     open_parentheses.Push(curr_pos);
     depth_counter.Down();
     curr_pos++;
    }; 
   inline void Push(const TERM& sym)
    {
     if (sym.IsVariable()) { PushVar(sym); }
     else if (sym.Arity()) { PushNonconstFun(sym); }
           else PushConst(sym);
    };
   inline FLAT_TERM* CurrPos() const { return curr_pos; }; 
 }; // class FLAT_TERM::OPEN

class FLAT_TERM::OPEN_WITH_MEMORY : public FLAT_TERM::OPEN
 {
  private:
   FLAT_TERM memory[MAX_TERM_WEIGHT+2];
  public:
   OPEN_WITH_MEMORY() : OPEN(memory) {};
   inline void Reset() { OPEN::Reset(memory); };
   inline const FLAT_TERM* Term() const { return memory; }; 
   inline long CurrWeight() const { return CurrPos() - memory; };
 }; // class FLAT_TERM::OPEN_WITH_MEMORY : public FLAT_TERM::OPEN

class FLAT_TERM::ITER
 {
  private:
   const FLAT_TERM* curr_pos;
   const FLAT_TERM* end_of_term;
  public:
   ITER() {};
   ITER(const FLAT_TERM* t) { Reset(t); };
   inline void Reset(const FLAT_TERM* t)
    {
     curr_pos = t;
     end_of_term = t->NextTerm();
    };
   inline const FLAT_TERM* CurrPos() const { return curr_pos; }; 
   inline bool EndOfTerm() const { return (curr_pos == end_of_term); };
   inline bool NotEndOfTerm() const { return (curr_pos < end_of_term); };
   inline void NextSym() { curr_pos++; };
   inline void NextTerm() { curr_pos = curr_pos->NextTerm(); }; 
 }; // class FLAT_TERM::ITER

//********************  Definitions for FLAT_TERM:  ********************************

inline unsigned FLAT_TERM::Depth() 
 {
  if (!(Symbol().Arity())) return 0;
  ITER iter(this);
  SMART_DEPTH_COUNTER depth_counter;
  depth_counter.Reset(0);
  unsigned depth = 0;
  do
   {
    depth_counter.NextSym(SymbolRef());   
    if (depth_counter.Depth() > depth) depth = depth_counter.Depth();
    iter.NextSym();
   }  
  while (iter.NotEndOfTerm());
  return depth;
 }; // unsigned FLAT_TERM::Depth()

inline TERM FLAT_TERM::StandardRepCopy() const
 {
  // very inefficient but can be used for input terms
  if (Symbol().IsVariable()) return Symbol();
  ITER iter(this);
  iter.NextSym();
  TERM result;
  result.SetFirst(new(Symbol().Arity()) TERM((ARITY)Symbol().Arity(),(FUNCTOR)Symbol().Functor(),(POLARITY)Symbol().Polarity()));
  TERM* arg = result.First()->Args();
  while (iter.NotEndOfTerm())
   {
    *arg = iter.CurrPos()->StandardRepCopy();
    iter.NextTerm();
    arg = arg->Next();
   };
  return result;
 }; // TERM FLAT_TERM::StandardRepCopy()

inline bool FLAT_TERM::Equal(const FLAT_TERM* t) const
 {
  // can be applied to variables
  if (Symbol() != t->Symbol()) return false;
  //if (Weight() != t->Weight()) return false;]
  const FLAT_TERM* sym = this + 1;
  if (sym != next_term)
   {
    t++;
    do
     {
      if (sym->Symbol() != t->Symbol()) return false;
      sym++;
      t++;
     }
    while (sym != next_term);
   };  
  return true; 
 }; // bool FLAT_TERM::Equal(const FLAT_TERM* t) const

inline bool FLAT_TERM::EqualArguments(const FLAT_TERM* t) const
 {
  //if (Weight() != t->Weight()) return false;
  const FLAT_TERM* sym1 = this + 1;
  const FLAT_TERM* sym2 = t + 1;
  while (sym1 != next_term)
   {
    if (sym1->Symbol() != sym2->Symbol()) return false;
    sym1++;
    sym2++;
   };
  return true; 
 }; // bool FLAT_TERM::EqualArguments(const FLAT_TERM* t) const

inline bool FLAT_TERM::Complementary(const FLAT_TERM* t) const
 {
  return (Symbol().ComplementaryHeader(t->Symbol()))
          && (EqualArguments(t));
 }; // bool FLAT_TERM::Complementary(const FLAT_TERM* t) const

inline COMPARISON FLAT_TERM::CompareLex(const FLAT_TERM* t) const
 {
  const FLAT_TERM* sym1 = this;
  const FLAT_TERM* sym2 = t;
  while (sym1 != next_term)
   {
    if (sym1->Symbol() != sym2->Symbol())
     {
      if ((sym1->Symbol().IsVariable()) || (sym2->Symbol().IsVariable()))
       {
        return INCOMPARABLE;
       }
      else
       {
        return sym1->Symbol().CompareFunctor(sym2->Symbol());
       }; 
     };
    sym1++;
    sym2++;
   };
  return EQUAL; 
 }; // COMPARISON FLAT_TERM::CompareLex(const FLAT_TERM* t) const


inline COMPARISON FLAT_TERM::KBCompare(const FLAT_TERM* t1,VAR_MULTISET* vars1,
                                       const FLAT_TERM* t2,VAR_MULTISET* vars2)
 {
  if (t1->Weight() == t2->Weight())
   {
    if (vars1->Equal(*vars2)) return t1->CompareLex(t2);
    return INCOMPARABLE;
   };
  // different weights 
  if (t1->Weight() > t2->Weight())
   {
    return vars1->GreaterOrEqual(*vars2) ? GREATER : INCOMPARABLE;
   }
  else // t1->Weight() < t2->Weight()
   {
    return vars2->GreaterOrEqual(*vars1) ? LESS : INCOMPARABLE;
   };
 }; // COMPARISON FLAT_TERM::KBCompare(const FLAT_TERM* t1,VAR_MULTISET* vars1,...


inline void FLAT_TERM::RenameVariables(INT_VAR_RENAMING& var_renaming)
 {
  for (FLAT_TERM* sym = this; sym < next_term; sym++)
   if (sym->Symbol().IsVariable())
    sym->SymbolRef().SetVar(var_renaming.Rename(sym->Symbol().Var()));
 }; // void FLAT_TERM::RenameVariables(INT_VAR_RENAMING& var_renaming)

 
inline void FLAT_TERM::OutputSymbol() const
 {
  ::Output(symbol);   
 };

inline void FLAT_TERM::OutputTerm() const
 {
  ITER iter(this);
  while (iter.NotEndOfTerm())
   {
    iter.CurrPos()->OutputSymbol();
    ::Output(" ");
    iter.NextSym();
   };     
 }; // void FLAT_TERM::OutputTerm() 

inline void FLAT_TERM::OutputStructure() const
 {
  ITER iter(this);
  while (iter.NotEndOfTerm())
   {
    ::Output("[");
    ::Output(iter.CurrPos() - this);
    ::Output("] ");
    iter.CurrPos()->OutputSymbol();
    ::Output(" n[");
    ::Output(iter.CurrPos()->NextTerm() - this);
    ::Output("]\n");
    iter.NextSym();
   };     
 }; // void FLAT_TERM::OutputStructure() 




//********************************************************************************
