#define __FORMULA_H

#ifndef __DEBUG_H
 #include "debug.h"
#endif
#ifndef __VAMPIRE_H
 #include "vampire.h"
#endif 
#ifndef __CLAUSE_H
 #include "clause.h"
#endif
#ifndef __TERM_H
 #include "term.h"
#endif
#ifndef __VARIABLES_H
 #include "variables.h"
#endif


//**************************************************************

class FORMULA
 {
  public:
   enum TAG { ATOM, NEG, CONJ, DISJ, IMPL, UNI, EXI };
  private:
   TAG tag;
   union
    {
     TERM* atom;
     FORMULA* form;
     int var; 
    } arg1;
   FORMULA* arg2; 
  public:
   FORMULA() {};
   FORMULA(TERM* at) : tag(ATOM) { arg1.atom = at; };
   FORMULA(TAG con,FORMULA* f) : tag(con), arg2(f) {};
    
   FORMULA(TAG con,FORMULA* f1,FORMULA* f2)
    : tag(con),
      arg2(f2)
     {
      arg1.form = f1;
     };  
   FORMULA(TAG quant,int v,FORMULA* f)
    : tag(quant),
      arg2(f) 
     {
      arg1.var = v;
     };  
   inline void* operator new(size_t)
    {
     return Memory::Allocate(sizeof(FORMULA)/4);
    };
   inline void operator delete(void* obj)
    {
     Memory::Deallocate(obj,sizeof(FORMULA)/4);
    }; 
   
   inline TAG Tag() const { return tag; };
   inline TERM* Atom() const { return arg1.atom; };
   inline FORMULA* Form1() const { return arg1.form; };
   inline FORMULA* Form2() const { return arg2; };
   inline int Var() const { return arg1.var; };
   static inline FORMULA* NewAtom(TERM* at) 
    { 
     return new FORMULA(at);
    };
   static inline FORMULA* NewNeg(FORMULA* f) 
    { 
     return new FORMULA(NEG,f);
    }; 
   static inline FORMULA* NewConj(FORMULA* f1,FORMULA* f2)
    {
     return new FORMULA(CONJ,f1,f2);
    };   
   static inline FORMULA* NewDisj(FORMULA* f1,FORMULA* f2)
    {
     return new FORMULA(DISJ,f1,f2);
    }; 
   static inline FORMULA* NewImpl(FORMULA* f1,FORMULA* f2)
    {
     return new FORMULA(IMPL,f1,f2);
    }; 
   static inline FORMULA* NewUni(int v,FORMULA* f)
    {
     return new FORMULA(UNI,v,f);
    }; 
   static inline FORMULA* NewExi(int v,FORMULA* f)
    {
     return new FORMULA(EXI,v,f);
    };
   static inline FORMULA* UQuantify(VAR_MULTISET& vars,FORMULA* f);
   static inline FORMULA* Translate(TERM* lit)
    {
     if (lit->Positive()) return NewAtom(lit);
     return NewNeg(NewAtom(lit));
    };
   static inline FORMULA* Translate(Clause* cl);
   
   inline void Destroy();
    
 }; // class FORMULA

//************** Definitions for FORMULA: ***************************

inline FORMULA* FORMULA::UQuantify(VAR_MULTISET& vars,FORMULA* f)
 {
  FORMULA* res = f;
  vars.ResetIterator();
  for (int* v = vars.NextVar();v;v = vars.NextVar()) res = NewUni(*v,res);
  return res;
 }; // FORMULA* FORMULA::UQuantify(VAR_MULTISET& vars,FORMULA* f)

inline FORMULA* FORMULA::Translate(Clause* cl)
 {
  static VAR_MULTISET vars; 
  vars.Reset();
  for (TERM* lit_ref = cl->Literals()->Args();!lit_ref->IsNull();lit_ref = lit_ref->Next())
   {
    TERM::ITERATOR iter;
    iter.Reset(lit_ref->First());
    while (iter.Next()) 
     if (iter.CurrentSymIsVar()) vars.Register(iter.Symbol());
   };
  
  // variables collected
  FORMULA* res = Translate(cl->Literals()->Args()->First());
  for (TERM* lit_ref = cl->Literals()->Args()->Next();!lit_ref->IsNull();lit_ref = lit_ref->Next())
   res = NewDisj(Translate(lit_ref->First()),res);
  
  return UQuanify(vars,res);
 }; // FORMULA* FORMULA::Translate(Clause* cl)

inline void FORMULA::Destroy()
 {
  switch (tag)
   {
    case ATOM: delete this; return;
    case NEG: Form1()->Destroy(); delete this; return; 
    case CONJ: 
    case DISJ:
    case IMPL:
     Form1()->Destroy();
     Form2()->Destroy();  
     delete this;
     return;
    case UNI:
    case EXI:
     Form2()->Destroy();  
     delete this;
     return;
   };
 }; // void FORMULA::Destroy()

//**************************************************************