#define __FS_COMMAND_H

#ifndef __DEBUG_H
 #include "debug.h"
#endif

#ifndef __VAMPIRE_H
 #include "vampire.h"
#endif

#ifndef __MEMORY_H
 #include "memory.h"
#endif 

#ifndef __TERM_H
 #include "term.h"
#endif

#ifndef __CLAUSE_H
 #include "clause.h"
#endif 

#ifndef __FLAT_TERM_H
 #include "flat_term.h"
#endif 

#ifndef __OUTPUT_H
 #include "output.h"
#endif  

//**************************************************************************

const int MAX_FS_CODE_DEPTH = MAX_CLAUSE_WEIGHT*2 + 1;

class FS_COMMAND
 {
  public:
   enum TAG 
    {
     LIT_HEADER,
     PROP_LIT,
     UNORD_EQ, 
     ORD_EQ, 
     SECOND_EQ_ARG, 
     FUNC, 
     VAR, 
     COMPARE, 
     SUCCESS 
    }; 
  private: // structure
   union 
    {
     int header_num;
     long func; // to be casted into TERM 
     const FLAT_TERM** var1;
     Clause* clause;
    } arg1;
   union
    { 
     const FLAT_TERM** var2;
     bool swap_flag;
    } arg2;
   FS_COMMAND* fork; 
   FS_COMMAND* next; 
   TAG tag;
  public:
   FS_COMMAND(TAG t) : fork((FS_COMMAND*)0), next((FS_COMMAND*)0), tag(t) {};
   inline void* operator new(size_t)
    {
     return Memory::Allocate(sizeof(FS_COMMAND)/4);
    };
   inline void operator delete(void* obj)
    { 
     Memory::Deallocate(obj,sizeof(FS_COMMAND)/4);
    };
   inline TAG Tag() const { return tag; };
   inline bool IsLitHeader() const { return (tag == LIT_HEADER); };
   inline bool IsPropLit() const { return (tag == PROP_LIT); };
   inline bool IsFunc() const { return (tag == FUNC); };
   inline bool IsVar() const { return (tag == VAR); };
   inline bool IsCompare() const { return (tag == COMPARE); };
   inline bool IsSuccess() const { return (tag == SUCCESS); };
   
   inline FS_COMMAND* Next() const { return next; };
   inline void SetNext(FS_COMMAND* c) { next = c; };
   inline FS_COMMAND** NextAddr() { return &(next); };
   
   inline FS_COMMAND* Fork() const { return fork; };
   inline void SetFork(FS_COMMAND* c) { fork = c; };
   inline FS_COMMAND** ForkAddr() { return &(fork); };
   
       // Methods for LIT_HEADER, PROP_LIT, UNORD_EQ and ORD_EQ
   inline int HeaderNum() const { return arg1.header_num; };
   inline void SetHeader(TERM h) { arg1.header_num = h.HeaderNum(); }; 
   static inline FS_COMMAND* NewLIT_HEADER(TERM header)
    {
     FS_COMMAND* res = new FS_COMMAND(LIT_HEADER);
     res->SetHeader(header);
     return res;
    };
   static inline FS_COMMAND* NewPROP_LIT(TERM header)
    {
     FS_COMMAND* res = new FS_COMMAND(PROP_LIT);
     res->SetHeader(header);
     return res;
    }; 
   static inline FS_COMMAND* NewUNORD_EQ(TERM header)
    {
     FS_COMMAND* res = new FS_COMMAND(UNORD_EQ);
     res->SetHeader(header);
     return res;
    }; 
   static inline FS_COMMAND* NewORD_EQ(TERM header)
    {
     FS_COMMAND* res = new FS_COMMAND(ORD_EQ);
     res->SetHeader(header);
     return res;
    };
      
       // for UNORD_EQ only
   inline bool SwapFlag() const { return arg2.swap_flag; };
   inline void SwapOn() { arg2.swap_flag = true; };
   inline void SwapOff() { arg2.swap_flag = false; };    
     
       // Methods for SECOND_EQ_ARG
   static inline FS_COMMAND* NewSECOND_EQ_ARG()
    {
     return new FS_COMMAND(SECOND_EQ_ARG);
    };
        
       // Methods for FUNC: 
   inline TERM Func() const { return TERM(arg1.func); };
   inline void SetFunc(TERM f) { arg1.func = f.Content(); };
   static inline FS_COMMAND* NewFUNC(TERM f)
    {
     FS_COMMAND* res = new FS_COMMAND(FUNC);
     res->SetFunc(f);
     return res;
    };
       // Methods for VAR:
   inline const FLAT_TERM** Var() const { return arg1.var1; };
   inline void SetVar(const FLAT_TERM** v) { arg1.var1 = v; };  
   static inline FS_COMMAND* NewVAR(const FLAT_TERM** v)
    {
     FS_COMMAND* res = new FS_COMMAND(VAR);
     res->SetVar(v);
     return res;
    };  
      
       // Methods for COMPARE:
   inline const FLAT_TERM** Var1() const { return arg1.var1; };
   inline void SetVar1(const FLAT_TERM** v) { arg1.var1 = v; };    
   inline const FLAT_TERM** Var2() const { return arg2.var2; };
   inline void SetVar2(const FLAT_TERM** v) { arg2.var2 = v; };
   inline bool SameVars(const FLAT_TERM** v1,const FLAT_TERM** v2)
    {
     return ((Var1() == v1) && (Var2() == v2));
    }; 
   static inline FS_COMMAND* NewCOMPARE(const FLAT_TERM** v1,const FLAT_TERM** v2)
    {
     CHECK7(v1 < v2,"ARGUMENTS in FS_COMMAND::NewCOMPARE(..)");
     FS_COMMAND* res = new FS_COMMAND(COMPARE);
     res->SetVar1(v1);
     res->SetVar2(v2);
     return res;
    }; 
      
      // Methods for SUCCESS:
   inline Clause* TheClause() const { return arg1.clause; };
   inline void SetClause(Clause* cl) { arg1.clause = cl; };
   static inline FS_COMMAND* NewSUCCESS(Clause* cl)
    {
     FS_COMMAND* res = new FS_COMMAND(SUCCESS);
     res->SetClause(cl);
     return res;
    };  
   
     // Destruction of a tree
   static void DestroyTree(FS_COMMAND* tree) 
    {
     DCONTROL7("BEGIN in FS_COMMAND::DestroyTree(FS_COMMAND* tree)");
     CHECK7(tree,"Wrong argument in FS_COMMAND::DestroyTree(FS_COMMAND* tree)");
     if (tree->Fork()) DestroyTree(tree->Fork());
     if (tree->IsSuccess()) 
      {
       delete tree; 
      }
     else
      {
       DestroyTree(tree->Next());
       delete tree; 
      };
     DCONTROL7("END in FS_COMMAND::DestroyTree(FS_COMMAND* tree)"); 
    }; 
  
  public: // methods for debugging
   inline void Output(const FLAT_TERM** vars)
    {
     ::Output("["); ::Output((long)this); ::Output("] ");
     TERM header;
     switch (Tag())
      {
       case LIT_HEADER : 
        header.MakeHeader(HeaderNum());
        ::Output("LIT_HEADER "); ::Output(header); break;
       case PROP_LIT :
        header.MakeHeader(HeaderNum());
        ::Output("PROP_LIT "); ::Output(header); break;
       case UNORD_EQ :
        header.MakeHeader(HeaderNum());
        ::Output("UNORD_EQ "); ::Output(header); break; 
       case ORD_EQ :
        header.MakeHeader(HeaderNum());
        ::Output("ORD_EQ "); ::Output(header); break; 
       case SECOND_EQ_ARG :
        ::Output("SECOND_EQ_ARG ");  
        break;
       case FUNC :
        ::Output("FUNC "); 
        ::Output(Func());
        break;
       case VAR :
        ::Output("VAR "); ::Output(Var() - vars); break;
       case COMPARE : 
        ::Output("COMPARE "); 
        ::Output(Var1() - vars); 
        ::Output(" ");
        ::Output(Var2() - vars); 
        break;
       case SUCCESS :
        ::Output("SUCCESS "); ::Output(TheClause()); break;
      }; 
     //::Output(" ["); ::Output((long)Fork()); ::Output("]"); 
    };
   static inline void OutputTree(FS_COMMAND* tree,const FLAT_TERM** subst);
 }; // class FS_COMMAND

//**************** Definitions for FS_COMMAND: ***************************

inline void FS_COMMAND::OutputTree(FS_COMMAND* tree,const FLAT_TERM** subst)
 {
  long fork_num = 0;
  if (!tree) { return; };
  STACK<FS_COMMAND*,MAX_FS_CODE_DEPTH> forks("MAX_FS_CODE_DEPTH in FS_CODE::Output():forks");
  STACK<long,MAX_FS_CODE_DEPTH> fork_nums("MAX_FS_CODE_DEPTH in FS_CODE::Output():fork_nums");
 
  FS_COMMAND* node = tree;
 
  next_branch:
   while (node)
    {
     node->Output(subst);
     if (node->Fork()) 
      {
       forks.Push(node); 
       fork_nums.Push(fork_num); 
       ::Output(" ["); ::Output(fork_num); ::Output("]");
       fork_num++;
      };
     ::Output("\n");
     node = node->Next();
    };
   if (forks.Empty()) { return; };
   node = forks.Pop();
   ::Output("\nFORK ["); ::Output(fork_nums.Pop()); ::Output("]\n");
   node = node->Fork();
   goto next_branch;
 }; // void FS_COMMAND::OutputTree(FS_COMMAND* tree) 


//*****************************************************************************
