
/*******************  Discrimination trees for matching. ****************/

#define __MI_TREE_H
#ifndef __DEBUG_H
 #include "debug.h"
#endif
#ifndef __VAMPIRE_H
 #include "vampire.h"
#endif
#ifndef __MEMORY_H
 #include "memory.h"
#endif
#ifndef __STACK_H
 #include "stack.h"
#endif
#ifndef __TERM_H
 #include "term.h"
#endif
#ifndef __CLAUSE_H
 #include "clause.h"
#endif 
#ifndef __VARIABLES_H
 #include "variables.h"
#endif 
#ifndef __FLAT_TERM_H
 #include "flat_term.h"
#endif 
#ifndef __OUTPUT_H
 #include "output.h"
#endif  
#ifndef __EQUIVALENCE_RELATION_H
 #include "equivalence_relation.h"
#endif  

//*********************************************************************

class MI_TREE
 {
  public:
   class NODE;
   class INTEGRATOR;
   class REMOVAL;
   class RETRIEVAL;
   class OUTPUT;
  public:
   static inline void Destroy(NODE* tree);  
 }; // class MI_TREE 

class MI_TREE::NODE
 {
  public: 
   enum TAG { FUNC, VAR, COMPARE, LEAF };  
  private:
   #ifdef _DEBUG19_
    long id;
    static long next_id;
   #endif        
   union 
    {
     long func; // to be casted into TERM
     const FLAT_TERM** var1;
     void* leaf_info1; 
    } arg1;
   union 
    {
     const FLAT_TERM** var2;
     void* leaf_info2;
    } arg2;   
   MI_TREE::NODE* fork;
   MI_TREE::NODE* next;
   TAG tag; 

  public:
   NODE(TAG t)  : fork((NODE*)0), next((NODE*)0), tag(t)
    {
     #ifdef _DEBUG19_
      id = next_id;
      next_id++;
     #endif
    };
   inline void* operator new(size_t)
    {
     return Memory::Allocate(sizeof(NODE)/4);
    };
   inline void operator delete(void* obj)
    { 
     Memory::Deallocate(obj,sizeof(NODE)/4);
    };
  
   #ifdef _DEBUG19_
    inline long Id() const { return id; };
   #endif


   inline TAG Tag() const { return tag; };
   inline bool IsFunc() const { return (tag == FUNC); };
   inline bool IsVar() const { return (tag == VAR); };
   inline bool IsCompare() const { return (tag == COMPARE); };
   inline bool IsLeaf() const { return (tag == LEAF); };    
 
   inline NODE* Next() const { return next; };
   inline NODE** NextAddr() { return &(next); };
   inline void SetNext(NODE* n) { next = n; };
   inline NODE* Fork() const { return fork; }; 
   inline NODE** ForkAddr() { return &(fork); };
   inline void SetFork(NODE* f) { fork = f; };

       // Methods for FUNC:
   inline TERM Func() const { return TERM(arg1.func); };
   inline void SetFunc(TERM f) { arg1.func = f.Content(); };
   static inline NODE* NewFUNC(TERM f)
    {
     NODE* res = new NODE(FUNC);
     res->SetFunc(f);
     return res;
    };

       // Methods for VAR:
   inline const FLAT_TERM** Var() const { return arg1.var1; };
   inline void SetVar(const FLAT_TERM** v) { arg1.var1 = v; };      
   static inline NODE* NewVAR(const FLAT_TERM** v)
    {
     NODE* res = new NODE(VAR);
     res->SetVar(v);
     return res;
    };

       // Methods for COMPARE:
   inline const FLAT_TERM** Var1() const { return arg1.var1; };
   inline void SetVar1(const FLAT_TERM** v) { arg1.var1 = v; };    
   inline const FLAT_TERM** Var2() const { return arg2.var2; };
   inline void SetVar2(const FLAT_TERM** v) { arg2.var2 = v; };
   inline void BothVars(const FLAT_TERM**& v1,const FLAT_TERM**& v2) { v1 = arg1.var1; v2 = arg2.var2; };
   static inline NODE* NewCOMPARE(const FLAT_TERM** v1,const FLAT_TERM** v2)
    {
     CHECK19(v1 < v2,"ARGUMENTS in MI_TREE::NODE::NewCOMPARE(..)");
     NODE* res = new NODE(COMPARE);
     res->SetVar1(v1);
     res->SetVar2(v2);

     /****** debug *******
     
     if (res->Id() == 1992)
      {
	::Output("CREATED IN "); REPORT_CONTROL_POINT(); ::Output("\n");
      };

     ************/        

     return res;
    }; 

     // Methods for LEAF:
   inline void*& LeafInfo1() { return arg1.leaf_info1; };
   inline void*& LeafInfo2() { return arg2.leaf_info2; };
   static inline NODE* NewLEAF()
    {
     NODE* res = new NODE(LEAF);
     res->LeafInfo1() = (void*)0;
     res->LeafInfo2() = (void*)0;
     return res;
    };
  public: // methods for debugging
   inline void Output(const FLAT_TERM** subst);    
 }; // class MI_TREE::NODE

typedef EQ_REL<MAX_TERM_WEIGHT> MI_EQ_CONSTR;

class MI_TREE::INTEGRATOR // objects must be static - too costly constructor
 {
  private:
   MI_TREE::NODE** node_addr;
   
   MI_EQ_CONSTR constr_to_insert;
   MI_EQ_CONSTR constr_to_shift;
   MI_EQ_CONSTR constr_passed;
   
   const FLAT_TERM** mi_subst;
  
   const FLAT_TERM** first_occ[MAX_NUMBER_OF_VARS];
   VAR_MULTISET vars; 
   const FLAT_TERM** mi_var;
   bool read_mode;
  public:
   INTEGRATOR(const FLAT_TERM** subst)
    : constr_to_insert(),
      constr_to_shift(),
      constr_passed(),
      mi_subst(subst),
      vars()      
    {};

   inline void Reset(MI_TREE::NODE** n)
    {
     node_addr = n;
     constr_to_insert.Reset();
     constr_to_shift.Reset();
     constr_passed.Reset();
     vars.Reset();
     mi_var = mi_subst;
     read_mode = (bool)(*node_addr);
    };

   inline VAR_MULTISET& Vars() { return vars; };

   inline const FLAT_TERM** MapVar(int v) const { return first_occ[v]; }; 

   inline void Push(const TERM& sym)
    { if (sym.IsVariable()) { PushVar(sym); } else PushFunc(sym); };  
   inline MI_TREE::NODE* Leaf()
    {
     read_mode = read_mode && ReadCollectedComps();        
     if (read_mode) 
      {
       CHECK19((*node_addr)->IsLeaf(),"A0 in MI_TREE::INTEGRATOR::Leaf()");
       return *node_addr;
      };
     // write mode
     *node_addr = MI_TREE::NODE::NewLEAF();
     return *node_addr;
    };

  private: 
   void PushVar(const TERM& sym);
   void PushFunc(const TERM& sym);

  private:
   inline void MoveDown() { node_addr = (*node_addr)->NextAddr(); };
   inline void MoveRight() { node_addr = (*node_addr)->ForkAddr(); };
   inline int VarNum(const FLAT_TERM** v) const { return v - mi_subst; };
   inline const FLAT_TERM** Var(int var_num) const { return mi_subst + var_num; };

     // Read mode
  private:
   bool ReadFunc(const TERM& func);
   bool ReadVar();
   void ReadCompare(const FLAT_TERM** v1,const FLAT_TERM** v2);
   bool ReadCollectedComps();
    
    // Write mode 
  private:
   void EnterWriteMode();
   void WriteFunc(const TERM& func);
   void WriteVar(int v);
   void WriteCompare(const FLAT_TERM** v1,const FLAT_TERM** v2);
   void InsertVar(const FLAT_TERM** v);
   void InsertFunc(const TERM& func);
      
    // Insertion of constraints 
  private:
   void CopyComps(const MI_EQ_CONSTR& constr,MI_TREE::NODE** addr);
   void InsertComps(MI_EQ_CONSTR& constr,MI_TREE::NODE** addr);
   void MoveComps(MI_EQ_CONSTR& constr,MI_TREE::NODE**& addr); 
 }; // class MI_TREE::INTEGRATOR

class MI_TREE::REMOVAL // objects must be static - too costly constructor
 {
  private:
   MI_TREE::NODE** node_addr;
   MI_TREE::NODE** branch_to_destroy;
   MI_EQ_CONSTR constr; 
   MI_EQ_CONSTR constr_passed;

   const FLAT_TERM** mi_subst;  
   const FLAT_TERM** first_occ[MAX_NUMBER_OF_VARS];
   VAR_MULTISET vars; 
   const FLAT_TERM** mi_var;
  public:
   REMOVAL(const FLAT_TERM** subst)
    : constr(),
      constr_passed(),
      mi_subst(subst),
      vars()   
    {};
   inline void Reset(MI_TREE::NODE** n)
    {
     // here *n must be nonnull!
     CHECK19(*n,"A0 in MI_TREE::REMOVAL::Reset(..)");
     node_addr = n;
     branch_to_destroy = n;
     constr.Reset();
     constr_passed.Reset();
     vars.Reset();
     mi_var = mi_subst; 
    };
   inline bool Skip(const TERM& sym)
    {
     if (sym.IsVariable()) { return SkipVar(sym.Var()); }
     else return SkipFunc(sym);
    };    
   inline bool SkipComps(); 
   inline MI_TREE::NODE* Leaf() const 
    {
     CHECK19((*node_addr)->IsLeaf(),"A0 in MI_TREE::REMOVAL::Leaf()");
     return *node_addr; 
    };
   inline MI_TREE::NODE** BranchToDestroy() const { return branch_to_destroy; }; 
  private:
   inline void MoveDown() { node_addr = (*node_addr)->NextAddr(); };
   inline void MoveRight() { node_addr = (*node_addr)->ForkAddr(); }; 
   inline int VarNum(const FLAT_TERM** v) const { return v - mi_subst; };
   inline bool SkipVar(int var);
   inline bool SkipVar();
   inline bool SkipFunc(const TERM& func);
 }; // class MI_TREE::REMOVAL

class MI_TREE::RETRIEVAL
 {  
  public:
   RETRIEVAL() {};
   MI_TREE::NODE* Match(const FLAT_TERM* query,MI_TREE::NODE* tree);
 }; // class MI_TREE::RETRIEVAL


class MI_TREE::OUTPUT
 {
  private: 
   const FLAT_TERM** vars;
   void (*LeafInfo)(MI_TREE::NODE*); 
  public:
   OUTPUT()
    : vars(0), 
      LeafInfo(&LeafInfoDefault)
     {};
   inline void Init() { vars = 0; LeafInfo = &LeafInfoDefault; };
  private:
   static void LeafInfoDefault(MI_TREE::NODE*); 
   inline void Var(const FLAT_TERM** v) { ::Output(v - vars); };
  public:
   inline void SetSubst(const FLAT_TERM** subst) { vars = subst; };
   inline void SetLeafInfo(void (*f)(MI_TREE::NODE*)) { LeafInfo = f; };
   inline void Node(MI_TREE::NODE* node);
   inline void Tree(MI_TREE::NODE* tree);    
 }; // class MI_TREE::OUTPUT


//****************** Definitions for MI_TREE: *******************************
 
inline void MI_TREE::Destroy(NODE* tree)
 {
  if (!tree) return;
  if (!tree->IsLeaf())
   {
    Destroy(tree->Fork());
    Destroy(tree->Next());
   }; 
  delete tree;    
 }; // void MI_TREE::Destroy(NODE* tree)

//****************** Definitions for MI_TREE::REMOVAL: ***********************

inline bool MI_TREE::REMOVAL::SkipVar(int var)
 {
  if (vars.Occurences(var))
   {
    constr.MakeEquivalent(VarNum(first_occ[var]),VarNum(mi_var)); 
   }
  else // first occurence
   {
    vars.Register(var);
    first_occ[var] = mi_var;
   };

  mi_var++;

  return SkipVar();    
 }; // bool MI_TREE::REMOVAL::SkipVar(int var) 

inline bool MI_TREE::REMOVAL::SkipVar()
 {
  CHECK19(*node_addr,"A10 in MI_TREE::REMOVAL::SkipFunc(..)");
  const FLAT_TERM** var1;
  const FLAT_TERM** var2;
  int var_num1;
  int var_num2;
  MI_TREE::NODE* curr_node;  
  check_node:
   curr_node = *node_addr;
   switch (curr_node->Tag())
    {
     case MI_TREE::NODE::FUNC: return false; // no functor can preceed a variable
     
     case MI_TREE::NODE::VAR:    
      if (curr_node->Fork()) branch_to_destroy = node_addr;
      MoveDown();
      return true;     

     case MI_TREE::NODE::COMPARE:
      curr_node->BothVars(var1,var2);     
      var_num1 = VarNum(var1);
      var_num2 = VarNum(var2);
      if (constr.Equivalent(var_num1,var_num2))
       {
        CHECK19(!curr_node->Fork(),"A30 in MI_TREE::REMOVAL::SkipFunc(..)");
        constr_passed.MakeEquivalent(var_num1,var_num2);
        MoveDown();
        goto check_node;
       }       
      else return false;
     
     #ifdef _DEBUG19_
      default: 
       DebugError("X0 inMI_TREE::REMOVAL::SkipVar()");
       return false;
     #else 
      #ifdef _SUPPRESS_WARNINGS_
       default: return false;
      #endif
     #endif 
    };
 }; // bool MI_TREE::REMOVAL::SkipVar() 


inline bool MI_TREE::REMOVAL::SkipFunc(const TERM& func)
 {
  CHECK19(*node_addr,"A10 in MI_TREE::REMOVAL::SkipFunc(..)");
  const FLAT_TERM** var1;
  const FLAT_TERM** var2;
  int var_num1;
  int var_num2;
  MI_TREE::NODE* curr_node;  
  check_node:
   curr_node = *node_addr;
   switch (curr_node->Tag())
    {
     case MI_TREE::NODE::FUNC:
      if (curr_node->Func() == func) 
       {
        if (curr_node->Fork()) branch_to_destroy = node_addr;
        MoveDown();
        return true;
       };
      try_fork:
       if (curr_node->Fork())
        {
         MoveRight();
         branch_to_destroy = node_addr;
         goto check_node;
        }        
       else return false; 

     case MI_TREE::NODE::COMPARE:
      curr_node->BothVars(var1,var2);     
      var_num1 = VarNum(var1);
      var_num2 = VarNum(var2);
      if (constr.Equivalent(var_num1,var_num2))
       {
        CHECK19(!curr_node->Fork(),"A30 in MI_TREE::REMOVAL::SkipFunc(..)");
        constr_passed.MakeEquivalent(var_num1,var_num2);
        MoveDown();
        goto check_node;
       }       
      else return false;  

    case MI_TREE::NODE::VAR: goto try_fork;

    #ifdef _DEBUG19_
      default: 
       DebugError("X0 inMI_TREE::REMOVAL::SkipFunc(..)");
       return false;
     #else 
      #ifdef _SUPPRESS_WARNINGS_
       default: return false;
      #endif
     #endif        
    };  
 }; // bool MI_TREE::REMOVAL::SkipFunc(const TERM& func)




inline bool MI_TREE::REMOVAL::SkipComps()
 {
  CHECK19(*node_addr,"A0 in MI_TREE::REMOVAL::SkipComps()");
  const FLAT_TERM** var1;
  const FLAT_TERM** var2;
  int var_num1;
  int var_num2;
  MI_TREE::NODE* curr_node;
  check_tag:
   curr_node = *node_addr;
 
   if (curr_node->IsCompare())
    {
     curr_node->BothVars(var1,var2);
     var_num1 = VarNum(var1);
     var_num2 = VarNum(var2);
     if (constr.Equivalent(var_num1,var_num2))
      {
       constr_passed.MakeEquivalent(var_num1,var_num2);
       if (curr_node->Fork()) branch_to_destroy = node_addr;       
       MoveDown();       
       goto check_tag;
      };
     if (curr_node->Fork())
      {
       MoveRight();
       branch_to_destroy = node_addr;
       goto check_tag;
      };
     return false;
    };
   // !curr_node->IsCompare()
   constr.Subtract(constr_passed);  
   return constr.Identity();  
 }; // bool MI_TREE::REMOVAL::SkipComps()




//********************** Definitions for MI_TREE::OUTPUT: ********************
 
inline void MI_TREE::OUTPUT::Node(MI_TREE::NODE* node)
 {
  ::Output("["); ::Output((long)node); ::Output("] ");
  switch (node->Tag())
   {
    case MI_TREE::NODE::VAR:
     ::Output("VAR ");
     Var(node->Var());
     break;
    case MI_TREE::NODE::FUNC:
     ::Output("FUNC ");
     ::Output(node->Func());   
     break;
    case MI_TREE::NODE::COMPARE:
     ::Output("COMPARE ");
     Var(node->Var1());
     ::Output(" ");
     Var(node->Var2());
     break;
    case MI_TREE::NODE::LEAF:
     ::Output("LEAF ");
     LeafInfo(node);
   };
  ::Output("["); ::Output((long)node->Next()); ::Output("] ");   
  ::Output("["); ::Output((long)node->Fork()); ::Output("] ");
 }; // void MI_TREE::OUTPUT::Node(MI_TREE::NODE* node) 

inline void MI_TREE::OUTPUT::Tree(MI_TREE::NODE* tree)
 {
  STACK<MI_TREE::NODE*,MAX_TERM_WEIGHT> forks("MAX_TERM_WEIGHT in MI_TREE::OUTPUT::Tree(..)");    
  next_node:
   Node(tree);
   ::Output("\n");
   if (tree->Fork()) forks.Push(tree);
   tree = tree->Next();
   if (tree) goto next_node;
   if (forks.Empty()) return;
   tree = forks.Pop();
   ::Output("fork from [");
   ::Output((long)tree);
   ::Output("]\n");
   tree = tree->Fork();
   goto next_node;   
 }; // void MI_TREE::OUTPUT::Tree(MI_TREE::NODE* tree) 


//********************* Methods for MI_TREE::NODE: **************************

inline void MI_TREE::NODE::Output(const FLAT_TERM** subst)
 {
  ::Output("["); ::Output((long)this); ::Output("] ");
  #ifdef _DEBUG19_
   ::Output(" "); ::Output(Id()); ::Output(" ");
  #endif

  switch (Tag())
   {
    case VAR:
     ::Output("VAR ");
     ::Output(Var() - subst);
     break;
    case FUNC:
     ::Output("FUNC ");
     ::Output(Func());   
     break;
    case MI_TREE::NODE::COMPARE:
     ::Output("COMPARE ");
     ::Output(Var1() - subst);
     ::Output(" ");
     ::Output(Var2() - subst);
     break;
    case MI_TREE::NODE::LEAF:
     ::Output("LEAF ");
   }; 
  ::Output(" n["); ::Output((long)Next()); ::Output("] ");
  ::Output("f["); ::Output((long)Fork()); ::Output("] ");
 }; // void MI_TREE::NODE::Output(const FLAT_TERM** subst)



//*****************************************************************************
