#define __NEW_CLAUSE_H
#ifndef __DEBUG_H
 #include "debug.h"
#endif
#ifndef __VAMPIRE_H
 #include "vampire.h"
#endif
#ifndef __TERM_H
 #include "term.h"
#endif
#ifndef __STACK_H
 #include "stack.h"
#endif
#ifndef __VARIABLES_H
 #include "variables.h"
#endif
#ifndef __WEIGHTING_H
 #include "weighting.h"
#endif
#ifndef __STATISTICS_H
 #include "statistics.h"
#endif
#ifndef __FLAT_TERM_H
 #include "flat_term.h"
#endif 
#ifndef __TMP_LITERAL_H
 #include "tmp_literal.h"
#endif 
#ifndef __TMP_LIT_LIST_H
 #include "tmp_lit_list.h"
#endif 

//**********************************************************************


class NEW_CLAUSE
 {
  private:
   TMP_LITERAL literal[MAX_NUMBER_OF_LITERALS + 1];
   TMP_LITERAL* next_literal;
   const TMP_LITERAL* last_possible_literal; 
   TMP_LIT_LIST all_literals;
   TMP_LITERAL* spare_literal;
   long weight;
   bool was_tautology;
   int num_of_unordered_eq_lits;
  private: // limits
   long max_weight;
   long max_lit_weight;
   long max_length;  
  public:
   NEW_CLAUSE() 
    :
     last_possible_literal(literal + (MAX_NUMBER_OF_LITERALS - 1)),
     all_literals(), 
     max_weight(MAX_CLAUSE_WEIGHT),
     max_lit_weight(MAX_TERM_WEIGHT),
     max_length(MAX_NUMBER_OF_LITERALS)
    {
    }; 
   inline void Reset()
    {
     next_literal = literal;
     spare_literal = literal + MAX_NUMBER_OF_LITERALS;
     weight = 0;
     was_tautology = false;
     num_of_unordered_eq_lits = 0;
    }; 
   
   inline void SetMaxWeight(long mw) { max_weight = (mw > MAX_CLAUSE_WEIGHT) ? MAX_CLAUSE_WEIGHT : mw; };
   inline void SetMaxLitWeight(long mlw) { max_lit_weight = (mlw > MAX_TERM_WEIGHT) ? MAX_TERM_WEIGHT : mlw; };
   inline void SetMaxLength(long ml) { max_length = (ml > MAX_NUMBER_OF_LITERALS) ? MAX_NUMBER_OF_LITERALS : ml; };  
   inline void SetMaxDepth(long md) 
    {
     for (TMP_LITERAL* lit = literal; lit <= last_possible_literal; lit++) lit->SetMaxDepth(md);
    }; 
    
   inline long Length() const { return next_literal - literal; };
   inline bool Empty() const { return next_literal == literal; };
   inline bool Unit() const { return (Length() == 1); };
   inline long Weight() const { return weight; };
   inline bool WasTautology() const { return was_tautology; };
     
   inline TMP_LIT_LIST& AllLiterals() { return all_literals; };
   //inline TMP_LITERAL* SpareTmpLit() { return literal + MAX_NUMBER_OF_LITERALS; };
   inline TMP_LITERAL*& SpareLiteral() { return spare_literal; }; 
     
   // collecting a new clause
   
   //inline void OpenLiteral() { next_literal->Reset(); };
   
   inline bool OpenNonpropLiteral(TERM header)
    {
     // header can't be a propositional literal
     weight += (header.Arity() + 1);
     if (weight > max_weight)  
      {
       STATISTICS_ON_GENERATED::AnotherTooHeavyClause();
       return false;
      };
     next_literal->Reset();  
     next_literal->PushNonpropHeader(header); 
     return true;
    };
   
   inline bool CloseEqLiteral()
    {
     CHECK10(next_literal->IsEquality(),"A0 in NEW_CLAUSE::CloseEqLiteral()");
     // try to reduce equality or detect a tautology
     if (next_literal->EqEqualArgs())
      {
       if (next_literal->Positive())
        {
         STATISTICS_ON_GENERATED::AnotherTautology();
         was_tautology = true;
         return false;
        }
       else 
        {
         weight -= next_literal->Weight();
         return true;
        }; 
      };
     for (TMP_LITERAL* lit = literal; lit < next_literal; lit++)
      {
       if (lit->IsEquality())
        {
         if (next_literal->EqualEqLit(*lit)) 
          {
           weight -= next_literal->Weight();
           return true;
          };
          
         if (next_literal->ComplementaryEqLit(*lit))
          {
           STATISTICS_ON_GENERATED::AnotherTautology();
           was_tautology = true;
           return false;
          };
        };  
      }; 
     if ((next_literal->Weight() > max_lit_weight) || (weight > max_weight))
      {
       STATISTICS_ON_GENERATED::AnotherTooHeavyClause();
       return false;
      };
     if (Length() >= max_length) 
      {
       STATISTICS_ON_GENERATED::AnotherTooLongClause();
       return false;
      }; 
     if (!next_literal->IsOrderedEq()) num_of_unordered_eq_lits++;
     next_literal++;
     return true; 
    }; // bool CloseEqLiteral()
   
   inline bool CloseOrdinaryLiteral()
    {
     // check for identical or complementary literals
     for (TMP_LITERAL* lit = literal; lit < next_literal; lit++)
      {
       if (next_literal->Equal(*lit)) 
        {
         weight -= next_literal->Weight();
         return true;
        }; 
       if (next_literal->Complementary(*lit))
        {
         STATISTICS_ON_GENERATED::AnotherTautology();
         was_tautology = true;
         return false;
        };
      };
     if ((next_literal->Weight() > max_lit_weight) || (weight > max_weight))
      {
       STATISTICS_ON_GENERATED::AnotherTooHeavyClause();
       return false;
      };
     if (Length() >= max_length) 
      {
       STATISTICS_ON_GENERATED::AnotherTooLongClause();
       return false;
      }; 
     next_literal++;
     return true;
    }; // bool CloseOrdinaryLiteral()
   
   inline bool CloseLiteral()
    {
     if (next_literal->IsEquality()) 
      {
       return CloseEqLiteral();
      };
     
     return CloseOrdinaryLiteral();
    }; // bool CloseLiteral() 
   
   inline void EndOfClause()
    {
     all_literals.Reset();
     for (TMP_LITERAL* lit = literal; lit < next_literal; lit++) 
      {
      
       //lit->Output(); cout << "\n\n"; // debug
      
      
       if (lit->IsEquality()) lit->EqOrderArgs();
       all_literals.Push(lit);
      }; 
    };


  inline bool Weed()
   {
    // can be used after simplifications
    if (all_literals.Weed()) return true;
    STATISTICS_ON_GENERATED::AnotherTautology();
    return false; 
   }; 
   
   inline bool PushPropLit(const TERM header)
    {
     TERM compl_header = header;
     compl_header.InversePolarity();
     for (TMP_LITERAL* lit = literal; lit < next_literal; lit++)
      {
       if (lit->FlatTerm()->Symbol() == header) return true;
       if (lit->FlatTerm()->Symbol() == compl_header)
        {
         STATISTICS_ON_GENERATED::AnotherTautology();
         was_tautology = true;
         return false;
        };
      };
    
     weight++;
     if (weight > max_weight)
      {
       STATISTICS_ON_GENERATED::AnotherTooHeavyClause();
       return false;
      };
     if (Length() >= max_length) 
      {
       STATISTICS_ON_GENERATED::AnotherTooLongClause();
       return false;
      };
     next_literal->Reset(); 
     next_literal->PushPropLit(header); 
     next_literal++;
     return true;
    }; 
   inline bool PushVar(const int v) 
    {
     // this check is too expensive and must be eliminated in the future!
     if (v >= MAX_NUMBER_OF_VARS) 
      {
       STATISTICS_ON_GENERATED::AnotherClauseWithTooManyVariables();
       return false; 
      };
     next_literal->PushVar(v);  
     return true;  
    };
   inline void PushVar(const TERM v) { next_literal->PushVar(v); };  
   inline void PushConst(const TERM c) { next_literal->PushConst(c); };
   inline bool PushNonconstFun(const TERM f) 
    {
     weight += f.Arity();
     if (weight > max_weight) 
      {
       STATISTICS_ON_GENERATED::AnotherTooHeavyClause();
       return false;
      };
     
     if (!next_literal->PushNonconstFun(f))
      {
       STATISTICS_ON_GENERATED::AnotherTooDeep();
       return false;
      };
      
     if (next_literal->Weight() > max_lit_weight)
      {
       STATISTICS_ON_GENERATED::AnotherTooHeavyTerm();
       return false;
      }; 
      
     return true;
    };
    
   public: // for debugging
    inline void Output()
     {
      for (TMP_LITERAL* lit = literal; lit < next_literal; lit++)
       {
        lit->Output();
        ::Output(" | ");
       };
     }; 
 }; // class NEW_CLAUSE
 

//***********************************************************************
