#define __PATH_INDEX_H

#ifndef __DEBUG_H
 #include "debug.h"
#endif

#ifndef __VAMPIRE_H
 #include "vampire.h"
#endif

#ifndef __TERM_H
 #include "term.h"
#endif

#ifndef __CLAUSE_H
 #include "clause.h"
#endif

#ifndef __OUTPUT_H
 #include "output.h"
#endif

#ifndef __PI_CLAUSE_LIST_H
 #include "pi_clause_list.h"
#endif 


//************************************************************************

class PI_NODE;
class PI_LEVEL;


class SMALL_PATH_INDEX
 {
  private:
   PI_NODE* tree;
   PI_C_LIST* all_clauses;
  public: 
   SMALL_PATH_INDEX() { Init(); };
   inline void Init() 
    {
     tree = (PI_NODE*)0; 
     all_clauses = (PI_C_LIST*)0;
    };
   inline void Destroy();  
   inline void* operator new(size_t)
    { 
     return Memory::Allocate(sizeof(SMALL_PATH_INDEX)/4);
    };
   inline void operator delete(void* obj)
    {
     Memory::Deallocate(obj,sizeof(SMALL_PATH_INDEX)/4);
    }; 
   inline PI_NODE* Tree(const TERM header); 
   inline PI_C_LIST* AllClauses() const { return all_clauses; };
   inline bool Empty() const { return !(AllClauses()); };
   inline void Integrate(const Clause* cl,int lit_num,const TERM* lit);
   inline void Integrate(const Clause* cl);
   inline void Remove(const Clause* cl,const TERM* lit);
   inline void Remove(const Clause* cl);
   inline PI_C_LIST* PropOccurences(TERM prop_lit,bool inverse_polarity);
  public:  
   static inline SMALL_PATH_INDEX** SpecIndexAddr(const TERM* lit)
    {
     return ((SMALL_PATH_INDEX**)(lit->InfoField()));       
    };   
   static inline SMALL_PATH_INDEX* SpecIndex(const TERM* lit) 
    {
     return *(SpecIndexAddr(lit));
    }; 
   static inline void InitSpecIndex(const TERM* lit) 
    {     
     *(SpecIndexAddr(lit)) = new SMALL_PATH_INDEX(); 
    };
   static inline void DestroySpecIndex(const TERM* lit) 
    {
     SpecIndex(lit)->Destroy();
     delete SpecIndex(lit);
     *(SpecIndexAddr(lit)) = (SMALL_PATH_INDEX*)0;
    };    
  public: // for debugging
   void OutputHeaders();
 }; // class SMALL_PATH_INDEX


class PATH_INDEX
 {
  private:
   PI_NODE* tree[MAX_NUMBER_OF_FUNCTORS*2];
   PI_C_LIST* prop_occurences[MAX_NUMBER_OF_FUNCTORS*2];
  public:
   PATH_INDEX() 
    {
     for (int n = 0; n < MAX_NUMBER_OF_FUNCTORS*2; n++)
      {
       tree[n] = (PI_NODE*)0;
       prop_occurences[n] = (PI_C_LIST*)0;
      };
    };
   inline PI_NODE* Tree(const TERM header);
   inline void IntegrateEq(const Clause* cl,int lit_num,int swapped_lit_num,const TERM* lit);
   inline void IntegrateNoneq(const Clause* cl,int lit_num,const TERM* lit);
   inline void Integrate(const Clause* cl)
    {        
     DCONTROL4("BEGIN in PATH_INDEX::Integrate(Clause* cl)");
     TERM* literals = cl->Literals();
     int num_of_literals = cl->Length();
     TERM* lit;   
     
     for (int lit_num = 0; lit_num < num_of_literals; lit_num++)
      {
       lit = literals->NthArg(lit_num)->First();
       if (lit->IsEquality())
        {
         IntegrateEq(cl,lit_num,lit_num + num_of_literals,lit);
        }
       else IntegrateNoneq(cl,lit_num,lit);
      };
     DCONTROL4("END in PATH_INDEX::Integrate(Clause* cl)");
    };
   
   inline void RemoveNoneq(const Clause* cl,const TERM* lit);
   inline void RemoveEq(const Clause* cl,const TERM* lit);
   inline void Remove(const Clause* cl)
    {     
     DCONTROL4("A0 in PATH_INDEX::Remove(Clause* cl)");
     TERM* literals = cl->Literals();
     int num_of_literals = cl->Length();
     TERM* lit;
     for (int lit_num = 0; lit_num < num_of_literals; lit_num++)
      {       
       lit = literals->NthArg(lit_num)->First();
       if (lit->IsEquality())
        {
         RemoveEq(cl,lit);
        }
       else  RemoveNoneq(cl,lit);
      };
     DCONTROL4("A1 in PATH_INDEX::Remove(Clause* cl)");
    };  
   inline PI_C_LIST* PropOccurences(TERM prop_lit,bool inverse_polarity);
   inline void Destroy();
 }; // class PATH_INDEX


class PI_LEVEL
 {
  private:
   PI_CLT_LIST* occurences;
   PI_NODE* subterms;
  public:
   inline void Init() { occurences = (PI_CLT_LIST*)0; subterms = (PI_NODE*)0; };
   inline PI_CLT_LIST*& Occurences() { return occurences; };
   inline PI_NODE*& ComplexSubterms() { return subterms; };
   inline void AddOccurence(const Clause* cl,int lit,const TERM t);
  public: // for debugging
   static inline bool CheckIntegrity(const PI_LEVEL* lev); 
   
 }; // class PI_LEVEL

class PI_NODE
 {
  private:
   TERM header; // possibly a predicate with polarity
   PI_NODE* alternative;
  private:
   inline void* operator new(size_t)
    {
     RuntimeError("Operator new called for PI_NODE");
     return (void*)0;
    }; 
   inline void operator delete(void* obj)
    {
     obj = (void*)0;
     RuntimeError("Operator delete called for PI_NODE");
    };
  public:
   inline TERM Header() const { return header; };
   inline bool Const() const { return !(header.Arity()); };
   inline PI_NODE*& Alternative() { return alternative; };
   inline PI_LEVEL* Subterms() const { return (PI_LEVEL*)(this + 1); };
   inline PI_CL_LIST*& ConstOccurences() const { return *((PI_CL_LIST**)(this + 1)); };
   inline PI_C_LIST*& PropOccurences() const { return *((PI_C_LIST**)(this + 1)); }; 
   static inline PI_NODE* NewNode(const TERM h)
    {
     // h.Arity() != 0 here
     CHECK2(h.Arity(),"wrong argument in PI_NODE::NewNode(..)");
     
     PI_NODE* res = (PI_NODE*)(Memory::Allocate((sizeof(PI_NODE)/4) + ((h.Arity())*(sizeof(PI_LEVEL)/4))));
     res->header = h;
     res->alternative = (PI_NODE*)0;
     PI_LEVEL* subterms = res->Subterms();
     for (PI_LEVEL* lev = subterms + h.Arity() - 1; lev >= subterms; lev--)
      {
       lev->Init();
      };
     return res;
    };
   static inline PI_NODE* NewConstNode(const TERM h) 
    {
     // h.Arity() == 0 here
     CHECK2(!(h.Arity()),"wrong argument in PI_NODE::NewConstNode(..)");
     PI_NODE* res = (PI_NODE*)(Memory::Allocate((sizeof(PI_NODE)/4) + 1));
     res->header = h;
     res->alternative = (PI_NODE*)0;
     res->ConstOccurences() = (PI_CL_LIST*)0;
     return res;
    };
   static inline PI_NODE* NewPropNode(const TERM h) 
    {
     // h.Arity() == 0 here
     CHECK2(!(h.Arity()),"wrong argument in PI_NODE::NewPropNode(..)");
     PI_NODE* res = (PI_NODE*)(Memory::Allocate((sizeof(PI_NODE)/4) + 1));
     res->header = h;
     res->alternative = (PI_NODE*)0;
     res->PropOccurences() = (PI_C_LIST*)0;
     return res;
    }; 
   static inline void Delete(PI_NODE* node)
    {
     // node->header.Arity() != 0 here
     
     CHECK2(node->header.Arity(),"wrong arg. in PI_NODE::Delete(..)");
     Memory::Deallocate((void*)node,(sizeof(PI_NODE)/4) + ((node->header.Arity())*(sizeof(PI_LEVEL)/4)));
    };  
   static inline void DeleteConst(PI_NODE* node)
    {
     // node->header.Arity() == 0 here
     
     CHECK2(!(node->header.Arity()),"wrong arg. in PI_NODE::DeleteConst(..)");
     Memory::Deallocate((void*)node,(sizeof(PI_NODE)/4) + 1);
    };  
   static inline void DeleteProp(PI_NODE* node)
    {
     // node->header.Arity() == 0 here
     
     CHECK2(!(node->header.Arity()),"wrong arg. in PI_NODE::DeleteProp(..)");
     Memory::Deallocate((void*)node,(sizeof(PI_NODE)/4) + 1);
    }; 
  public: // for debugging
   static inline bool CheckIntegrity(const PI_NODE* node)
    {
     if (!node) return true;
     #ifdef _DEBUG_MEMORY_
      if (!(Memory::Captured((void*)node))) return false; 
     #endif
     if (node->header.Flag() != TERM_COMPLEX) return false;
     if (!(CheckIntegrity(node->alternative))) return false;
     for (int i = 0; i < node->header.Arity(); i++)
      {
       if (!(PI_LEVEL::CheckIntegrity(node->Subterms() + i))) return false;
      };
     return true;  
    };
   inline void Output() 
    {
     ::Output(header);
    };     
 }; // class PI_NODE

class PI_TREE
 {
  public:
   static PI_NODE* CreateBranch(const Clause* cl,int lit,const TERM* t);
  private: 
   static void Remove(const Clause* cl,const TERM* t,PI_NODE** node);
  public:
   static void Destroy(PI_NODE* node);
   static void DestroyFromTop(PI_NODE* node);
   static void Integrate(const Clause* cl,int lit,const TERM* t_ref,PI_LEVEL* lev);
   static void Remove(const Clause* cl,const TERM* t_ref,PI_LEVEL* lev);
 }; // class PI_TREE


//***************** Definitions for PATH_INDEX: ******************************

inline void PATH_INDEX::IntegrateNoneq(const Clause* cl,int lit_num,const TERM* lit)
 {
  TERM header = *lit;
  int arity = header.Arity();
  if (arity)
   {
    PI_LEVEL* subterms = Tree(header)->Subterms();
    for (int n = 0; n < arity; n++)
     { 
      PI_TREE::Integrate(cl,lit_num,lit->NthArg(n),subterms);
      subterms++;
     };
   }
  else // propositional variable
   {
    PI_C_LIST::Insert(prop_occurences + ((MAX_NUMBER_OF_FUNCTORS*(header.Polarity())) + header.Functor()),
                      cl);
   };
 }; // void PATH_INDEX::IntegrateNoneq(Clause* cl,int lit,TERM* lit) 

inline void PATH_INDEX::IntegrateEq(const Clause* cl,int lit_num,int swapped_lit_num,const TERM* lit)
 {
  TERM header = *lit;
  header.MakeUnordered();
  PI_LEVEL* subterms = Tree(header)->Subterms();
  PI_TREE::Integrate(cl,lit_num,lit->NthArg(0),subterms);
  PI_TREE::Integrate(cl,lit_num,lit->NthArg(1),subterms+1);
  // now swap arguments
  PI_TREE::Integrate(cl,swapped_lit_num,lit->NthArg(1),subterms);
  PI_TREE::Integrate(cl,swapped_lit_num,lit->NthArg(0),subterms+1);  
 }; // void PATH_INDEX::IntegrateEq(const Clause* cl,int lit_num,int swapped_lit_num,const TERM* lit) 



inline void PATH_INDEX::RemoveNoneq(const Clause* cl,const TERM* lit)
 {
  
  DCONTROL4("BEGIN in PATH_INDEX::Remove(const Clause* cl,const TERM* lit)");
  TERM header = *lit;
  int arity = header.Arity();
  if (arity)
   {
    DCONTROL4("A0 in PATH_INDEX::Remove(const Clause* cl,const TERM* lit)");
    PI_LEVEL* subterms = Tree(header)->Subterms();
    for (int n = 0; n < arity; n++)
     {
      DCONTROL4("A1 in PATH_INDEX::Remove(const Clause* cl,const TERM* lit)"); 
      
      PI_TREE::Remove(cl,lit->NthArg(n),subterms);
      
      DCONTROL4("A2 in PATH_INDEX::Remove(const Clause* cl,const TERM* lit)");
      
      CHECK2(PI_LEVEL::CheckIntegrity(subterms),"A5 in PATH_INDEX::Remove(const Clause* cl,const TERM* lit)");
      
      subterms++;
     };
   }
  else // propositional variable
   {
    DCONTROL4("A8 in PATH_INDEX::Remove(const Clause* cl,const TERM* lit)");
    PI_C_LIST::Remove(prop_occurences + ((MAX_NUMBER_OF_FUNCTORS*(header.Polarity())) + header.Functor()),
                      cl);
   };
  DCONTROL4("END in PATH_INDEX::Remove(const Clause* cl,const TERM* lit)"); 
 }; // void PATH_INDEX::RemoveNoneq(const Clause* cl,const TERM* lit)



inline void PATH_INDEX::RemoveEq(const Clause* cl,const TERM* lit)
 {
  TERM header = *lit;   
  header.MakeUnordered();
  PI_LEVEL* subterms = Tree(header)->Subterms();

  PI_TREE::Remove(cl,lit->NthArg(1),subterms); 
  PI_TREE::Remove(cl,lit->NthArg(0),subterms+1);

  PI_TREE::Remove(cl,lit->NthArg(0),subterms); 
  PI_TREE::Remove(cl,lit->NthArg(1),subterms+1);

 }; // void PATH_INDEX::RemoveEq(const Clause* cl,const TERM* lit) 


inline PI_NODE* PATH_INDEX::Tree(const TERM header)
 {
  // header.Arity() != 0 here
  CHECK2(header.Arity(),"wrong argument in PATH_INDEX::Tree(..)");
  int n = (MAX_NUMBER_OF_FUNCTORS*(header.Polarity())) + header.Functor();
  if (!(tree[n]))
   {
    tree[n] = PI_NODE::NewNode(header);
   }; 
  return tree[n];
 }; // PI_NODE* PATH_INDEX::Tree(const TERM header)


inline PI_C_LIST* PATH_INDEX::PropOccurences(TERM prop_lit,bool inverse_polarity)
 { 
  if (inverse_polarity) prop_lit.InversePolarity();
  
  return prop_occurences[(MAX_NUMBER_OF_FUNCTORS*(prop_lit.Polarity())) + prop_lit.Functor()]; 
 }; // PI_C_LIST* PATH_INDEX::PropOccurences(TERM prop_lit,bool inverse_polarity)

inline void PATH_INDEX::Destroy()
 {
  for (int n = 0; n < MAX_NUMBER_OF_FUNCTORS*2; n++)
   {
    PI_TREE::Destroy(tree[n]);
    tree[n] = (PI_NODE*)0;
    PI_C_LIST::Destroy(prop_occurences[n]);
    prop_occurences[n] = (PI_C_LIST*)0;
   };
 }; // void PATH_INDEX::Destroy()

//***************** Definitions for SMALL_PATH_INDEX: **********************

inline void SMALL_PATH_INDEX::Remove(const Clause* cl)
 {
  DCONTROL4("A0 in SMALL_PATH_INDEX::Remove(Clause* cl)");
  TERM* literals = cl->Literals();
  int num_of_literals = cl->Length();
  for (int lit_num = 0; lit_num < num_of_literals; lit_num++)
   {       
    Remove(cl,literals->NthArg(lit_num)->First());
   };
  
  PI_C_LIST::Remove(&all_clauses,cl);  
   
   
  DCONTROL4("A1 in SMALL_PATH_INDEX::Remove(Clause* cl)");
 }; // void SMALL_PATH_INDEX::Remove(const Clause* cl)

inline void SMALL_PATH_INDEX::Remove(const Clause* cl,const TERM* lit)
 { 
  TERM header = *lit;
  int arity = header.Arity();
  PI_NODE* node = Tree(header);
  CHECK2(node,"A0 in SMALL_PATH_INDEX::Remove(const Clause* cl,const TERM* lit)");
  if (arity)
   {
    PI_LEVEL* subterms = node->Subterms();
    for (int n = 0; n < arity; n++)
     {
      PI_TREE::Remove(cl,lit->NthArg(n),subterms);
      
      CHECK2(PI_LEVEL::CheckIntegrity(subterms),"A5 in SMALL_PATH_INDEX::Remove(const Clause* cl,const TERM* lit)");
      
      subterms++;
     };
   }
  else
   {
    PI_C_LIST::Remove(&(node->PropOccurences()),cl);
   }; 
 }; // void SMALL_PATH_INDEX::Remove(const Clause* cl,const TERM* lit)

inline void SMALL_PATH_INDEX::Integrate(const Clause* cl)
 {
  DCONTROL4("BEGIN in SMALL_PATH_INDEX::Integrate(Clause* cl)");
  TERM* literals = cl->Literals();
  int num_of_literals = cl->Length();
     
  for (int lit_num = 0; lit_num < num_of_literals; lit_num++)
   {
    Integrate(cl,lit_num,literals->NthArg(lit_num)->First());
   };
  
  PI_C_LIST::Insert(&all_clauses,cl); 
   
  DCONTROL4("END in SMALL_PATH_INDEX::Integrate(Clause* cl)");
 }; // void SMALL_PATH_INDEX::Integrate(const Clause* cl)


inline void SMALL_PATH_INDEX::Integrate(const Clause* cl,int lit_num,const TERM* lit)
 {     
  TERM header = *lit;
  PI_NODE* node = Tree(header);
  int arity = header.Arity();
  if (arity)
   {
    if (node)
     {
      integrate_into_node:
       PI_LEVEL* subterms = node->Subterms();
       for (int n = 0; n < arity; n++)
        { 
         PI_TREE::Integrate(cl,lit_num,lit->NthArg(n),subterms);
         subterms++;
        };
     }
    else
     {
      PI_NODE* tmp = tree;
      //tree = PI_TREE::CreateBranch(cl,lit_num,lit);
      
      tree = PI_NODE::NewNode(header);
      
      tree->Alternative() = tmp;
      
      node = tree;
      goto integrate_into_node;
     }; 
   }
  else // propositional
   {
    if (!(node))
     {
      node = PI_NODE::NewPropNode(*lit);
      node->Alternative() = tree;
      tree = node;
     };
     
    PI_C_LIST::Insert(&(node->PropOccurences()),cl);     
   }; 
  CHECK2(Tree(header),"A10 in SMALL_PATH_INDEX::Integrate(const Clause* cl,int lit_num,const TERM* lit)");
 }; // void SMALL_PATH_INDEX::Integrate(const Clause* cl,int lit_num,const TERM* lit)
 


inline PI_NODE* SMALL_PATH_INDEX::Tree(const TERM header)
 {
  // header.Arity() == 0 is possible here!
  PI_NODE* res = tree;
  while ((res) && (res->Header() != header)) res = res->Alternative();
  return res; 
 }; // PI_NODE* SMALL_PATH_INDEX::Tree(const TERM header)

inline PI_C_LIST* SMALL_PATH_INDEX::PropOccurences(TERM prop_lit,bool inverse_polarity)
 {
  // prop_lit.Arity() = 0 here
  CHECK2(!(prop_lit.Arity()),"wrong argument in  SMALL_PATH_INDEX::PropOccurences(..)");
  if (inverse_polarity) prop_lit.InversePolarity();
  PI_NODE* node = Tree(prop_lit);
  if (node) { return node->PropOccurences(); }
  else { return (PI_C_LIST*)0; };
 }; // PI_C_LIST* SMALL_PATH_INDEX::PropOccurences(TERM prop_lit,bool inverse_polarity)

inline void SMALL_PATH_INDEX::Destroy() 
 {
  PI_TREE::DestroyFromTop(tree); 
  PI_C_LIST::Destroy(all_clauses);  
  tree = (PI_NODE*)0;
  all_clauses = (PI_C_LIST*)0;
 }; 

//***************** Definitions for PI_LEVEL: ******************************

inline void PI_LEVEL::AddOccurence(const Clause* cl,int lit,const TERM t)
 {
  PI_CLT_LIST::Insert(&occurences,cl,lit,t);
 }; // void PI_LEVEL::AddOccurence(Clause* cl,int lit,const TERM t)


inline bool PI_LEVEL::CheckIntegrity(const PI_LEVEL* lev) 
 {
 
  return ((PI_CLT_LIST::CheckIntegrity(lev->occurences)) 
         && (PI_NODE::CheckIntegrity(lev->subterms)));
 }; // bool PI_LEVEL::CheckIntegrity(const PI_LEVEL* lev) 
 

//***********************************************************************






























