#define __PREFIX_H
#ifndef __DEBUG_H
 #include "debug.h"
#endif
#ifndef __TERM_H
 #include "term.h"
#endif
#ifndef __VARIABLES_H
 #include "variables.h"
#endif

class PREF_SYM 
 {
  private:
   TERM sym;
  public:
   PREF_SYM() {};
   PREF_SYM(TERM s) 
    {
     if (s.IsVariable())
      {
       MkVar(VAR::Base(s.Var()));
      }
     else // functor
      {
       sym = s;
      };
    };
   inline bool IsFunctor() const { return (bool)(sym.Flag()); };
   inline bool IsVariable() const { return !(sym.Flag()); };
   inline TERM Func() const { return sym; };
   inline VAR* Var() const { return (VAR*)(sym.First()); };
   inline void MkFunc(const TERM& f) { sym = f; };
   inline void InversePolarity() { sym.InversePolarity(); };
   inline void MkVar(const VAR* const v) { sym.SetFirst((TERM*)v); };
   inline int operator==(const PREF_SYM& s) const 
     { return sym.Content() == s.Func().Content(); };
   inline int operator!=(const PREF_SYM& s) const 
     { return sym.Content() != s.Func().Content(); };
  public: // methods for debugging
   inline void Output()
    {
     if (IsFunctor())
      {
       if (Func().Polarity() == POL_NEG) { ::Output("~"); };
       OutputFunctor(Func().Functor());
      }
     else { Var()->Output(); };
    };
   inline void Output(unsigned var_offset)
    {
     if (IsFunctor())
      {
       if (Func().Polarity() == POL_NEG) { ::Output("~"); };
       OutputFunctor(Func().Functor());
      }
     else { (Var()+var_offset)->Output(); };
    }; 
 }; // class PREF_SYM
 
class PREFIX_SYM
 {
  private: // structure
   PREF_SYM head;
   PREFIX_SYM* bracket;
  public: 
   PREFIX_SYM() {};
   inline PREF_SYM& Head() { return head; };
   inline void SetBracket(PREFIX_SYM* const br) { bracket = br; };
   inline void ResetBracket() { bracket = 0; };
   inline PREFIX_SYM* Bracket() const { return bracket; };
   inline PREFIX_SYM* Arg1() { return this + 1; };
   inline PREFIX_SYM* Arg2() { return Arg1()->Bracket(); };
  public: // methods for debugging
   inline void Output()
    {
     Head().Output();
     ::Output(" [");
     ::Output((long unsigned int)Bracket());
     ::Output("]");
    };
 }; // class PREFIX_SYM


extern void OutputPrefix(void* word); // debug
 
class PREFIX
 {
  public:
   static inline long Weight(const PREFIX_SYM* const word) 
    {
     return (word->Bracket() - word);
    };
  public:
   class SIMPLE_ITER
    {
     private:
      PREFIX_SYM* curr_pos;
      PREFIX_SYM* end_of_word;
     public:
      SIMPLE_ITER() {};
      inline void Reset(PREFIX_SYM* const word) // word must represent a complex term here
       {
        CHECK1(PREFIX::CorrectComplexTerm(word),"bad argument in PREFIX::SIMPLE_ITER::Reset()"); 
        curr_pos = word;
        end_of_word = word->Bracket();
       };
      inline void MemorizeState(SIMPLE_ITER& memo) const
       {
        memo.curr_pos = curr_pos;
        memo.end_of_word = end_of_word;
       };
      inline void RestoreState(const SIMPLE_ITER& memo)
       {
        curr_pos = memo.curr_pos;
        end_of_word = memo.end_of_word;
       }; 
      inline PREFIX_SYM* CurrPos() const { return curr_pos; };
      inline bool Empty() const { return (curr_pos == end_of_word); };
      inline bool NotEmpty() const { return (bool)(end_of_word-curr_pos); };
      inline void SkipSym()
       {
        curr_pos++; 
        CHECK1(curr_pos <= end_of_word,"in PREFIX::SIMPLE_ITER::SkipSym()");
       };
      inline void SkipTerm()
       {
        CHECK1((curr_pos < curr_pos->Bracket()) 
                && (curr_pos->Bracket() - curr_pos <= MAX_TERM_WEIGHT),
               "strange bracket in PREFIX::SIMPLE_ITER::SkipTerm()");
        curr_pos = curr_pos->Bracket();
        CHECK1(curr_pos,"null bracket in PREFIX::SIMPLE_ITER::SkipTerm()");
       };
    }; // class PREFIX::SIMLE_ITER
  public:
   class ITER : private SIMPLE_ITER
    {
     private:
      SIMPLE_ITER* next_continuation;
      SIMPLE_ITER continuation[MAX_TERM_DEPTH];
      VAR* curr_var;
     private:
      inline void PopContinuation()
       {
        next_continuation--;
        SIMPLE_ITER::RestoreState(*next_continuation);
       };
      inline void PushContinuation()
       {
        SIMPLE_ITER::MemorizeState(*next_continuation); 
        next_continuation++;
       };
     public: 
      ITER() {};
      inline void Reset(PREFIX_SYM* const word) // word must represent a complex term here
       {
        CHECK1(word,"null argument in PREFIX::ITER::Reset()");
        CHECK1(word->Head().IsFunctor(),"strange argument in PREFIX::ITER::Reset()");
        CHECK1(word->Bracket(),"null bracket in PREFIX::ITER::Reset()");
        SIMPLE_ITER::Reset(word);
        //continuation_num = 0;
        next_continuation = continuation;
        curr_var = (VAR*)0;
       };
      inline bool Empty() const
       {
        #ifdef _DEBUG1_
         if (SIMPLE_ITER::Empty()) 
          {
           CHECK1(next_continuation == continuation,"in PREFIX::ITER::Empty()");
           return true;
          }
         else { return false; };
        #else
         return SIMPLE_ITER::Empty(); 
        #endif
       };
      inline bool NotEmpty() const { return SIMPLE_ITER::NotEmpty(); };
      inline VAR* CurrVar() const { return curr_var; };
      inline PREFIX_SYM* CurrSubterm() const { return SIMPLE_ITER::CurrPos(); };
      inline void SkipSym() // can't be applied if the iterator is empty
       {
        // SIMPLE_ITER can't be empty here
        CHECK0(!(SIMPLE_ITER::Empty()),"in PREFIX::ITER::SkipSym()");

        SIMPLE_ITER::SkipSym();
        if (SIMPLE_ITER::Empty()) 
         {
          if (next_continuation - continuation)
           {
            PopContinuation();
           }
          else
           {
            return; // empty iterator now
           };
         };
        if (SIMPLE_ITER::CurrPos()->Head().IsVariable())
         {
          curr_var = SIMPLE_ITER::CurrPos()->Head().Var();
          void* bind = curr_var->Binding();
          if (bind) // bound variable
           {
            check_bind_type: 
             if (VAR::IsVar(bind))
              {
               curr_var = (VAR*)bind;
               bind = curr_var->Binding();
               if (bind) { goto check_bind_type; };
              }
             else // variable bound by a complex term
              {
               curr_var = (VAR*)0;
               SIMPLE_ITER::SkipSym();
               if (SIMPLE_ITER::NotEmpty())
                {
                 PushContinuation();
                };
               SIMPLE_ITER::Reset((PREFIX_SYM*)bind);
              };
           };
         }
        else
         {
          curr_var = (VAR*)0;
         };
       }; // void SkipSym()  
       
      inline void SkipTerm() // can't be applied if the iterator is empty
       {
        // SIMPLE_ITER can't be empty here
        CHECK0(!(SIMPLE_ITER::Empty()),"in PREFIX::ITER::SkipSym()");
        
        SIMPLE_ITER::SkipTerm(); // the only difference with void SkipSym() 
        if (SIMPLE_ITER::Empty()) 
         {
          if (next_continuation - continuation)
           {
            PopContinuation();
           }
          else
           {
            return; // empty iterator now
           };
         };
        if (SIMPLE_ITER::CurrPos()->Head().IsVariable())
         {
          curr_var = SIMPLE_ITER::CurrPos()->Head().Var();
          void* bind = curr_var->Binding();
          if (bind) // bound variable
           {
            check_bind_type: 
             if (VAR::IsVar(bind))
              {
               curr_var = (VAR*)bind;
               bind = curr_var->Binding();
               if (bind) { goto check_bind_type; };
              }
             else // variable bound by a complex term
              {
               curr_var = (VAR*)0;
               SIMPLE_ITER::SkipSym();
               if (SIMPLE_ITER::NotEmpty())
                {
                 PushContinuation();
                };
               SIMPLE_ITER::Reset((PREFIX_SYM*)bind);
              };
           };
         }
        else
         {
          curr_var = (VAR*)0;
         };
       }; // void SkipTerm()
    }; // class PREFIX::ITER 
  public:
   class ITER_WITH_BACKTRACK : private SIMPLE_ITER
    {
     private:
      enum ACTION_TO_UNDO
       {
        NOTHING = 0,
        POP = 1, // = NOTHING + POP
        PUSH = 2, // = NOTHING + PUSH
        POP_AND_PUSH = 3 // = POP + PUSH
       };
     private:
      SIMPLE_ITER* next_continuation;
      SIMPLE_ITER continuation[MAX_TERM_DEPTH];
      VAR* current_var[MAX_TERM_WEIGHT];
      VAR** next_curr_var;
      VAR* curr_var;
     private:
      #ifdef _DEBUG_
       PREFIX_SYM* top_level;
      #endif
       
     private:
      ACTION_TO_UNDO undo[MAX_TERM_WEIGHT];
      unsigned undo_num;
      SIMPLE_ITER state[MAX_TERM_WEIGHT];
      SIMPLE_ITER used_continuation[MAX_TERM_WEIGHT];
     private:
      inline void PushCurrVar() 
       {
        *next_curr_var = curr_var;
        next_curr_var++;
        CHECK1(next_curr_var - current_var <= MAX_TERM_WEIGHT,
               "in PREFIX::ITER_WITH_BACKTRACK::PushCurrVar()");
       };   
      inline void PopCurrVar()
       {
        next_curr_var--;
        
        CHECK1(next_curr_var >= current_var,
               "in PREFIX::ITER_WITH_BACKTRACK::PopCurrVar()");
        
        curr_var = *next_curr_var;
       };   
      inline void PopContinuation()
       {
        next_continuation--;
        SIMPLE_ITER::RestoreState(*next_continuation);
       };
      inline void PushContinuation()
       {
        SIMPLE_ITER::MemorizeState(*next_continuation); 
        next_continuation++;
       };
     public:
      #ifdef _DEBUG_
       inline PREFIX_SYM* TopLevel() { return top_level; };
      #endif  
     public: 
      ITER_WITH_BACKTRACK() {};
      inline void Reset(PREFIX_SYM* const word) // word must represent a complex term here
       {
        CHECK1(word,"null agrument in PREFIX::ITER_WITH_BACKTRACK::Reset()");
        CHECK1(word->Head().IsFunctor(),"strange agrument in PREFIX::ITER_WITH_BACKTRACK::Reset()");
        CHECK1(word->Bracket(),"null bracket in PREFIX::ITER_WITH_BACKTRACK::Reset()");
        
        SIMPLE_ITER::Reset(word);
        
        #ifdef _DEBUG_
         top_level = word;
        #endif
        
        next_continuation = continuation;
        curr_var = (VAR*)0;
        next_curr_var = current_var;
        undo_num = 0;
       };
      inline bool Empty() const { return SIMPLE_ITER::Empty(); };
      inline bool NotEmpty() const { return SIMPLE_ITER::NotEmpty(); };
      inline VAR* CurrVar() const { return curr_var; };
      inline PREFIX_SYM* CurrSubterm() const { return SIMPLE_ITER::CurrPos(); };
      inline void SkipSym() // can't be applied if the iterator is empty
       {
        // SIMPLE_ITER can't be empty here
        CHECK1(!(SIMPLE_ITER::Empty())," in PREFIX::ITER_WITH_BACKTRACK::SkipSym()");
        
        SIMPLE_ITER::MemorizeState(state[undo_num]);
        PushCurrVar();
        
        unsigned undo_action = NOTHING;
        SIMPLE_ITER::SkipSym();
        if (SIMPLE_ITER::Empty()) 
         {
          if (next_continuation - continuation)
           {
            undo_action += POP;
            PopContinuation();
            SIMPLE_ITER::MemorizeState(used_continuation[undo_num]);
           }
          else
           {
            undo[undo_num] = (ACTION_TO_UNDO)undo_action;
            undo_num++;
            return; // empty iterator now
           };
         };
        if (SIMPLE_ITER::CurrPos()->Head().IsVariable())
         {
          curr_var = SIMPLE_ITER::CurrPos()->Head().Var();
          void* bind = curr_var->Binding();
          if (bind) // bound variable
           {
            check_bind_type: 
             if (VAR::IsVar(bind))
              {
               curr_var = (VAR*)bind;
               bind = curr_var->Binding();
               if (bind) { goto check_bind_type; };
              }
             else // variable bound by a complex term
              {
               curr_var = (VAR*)0;
               SIMPLE_ITER::SkipSym();
               if (SIMPLE_ITER::NotEmpty())
                {
                 undo_action += PUSH;
                 PushContinuation();
                };
               SIMPLE_ITER::Reset((PREFIX_SYM*)bind);
              };
           };
         }
        else
         {
          curr_var = (VAR*)0;
         };
        undo[undo_num] = (ACTION_TO_UNDO)undo_action;
        undo_num++;
       }; // void SkipSym() 
       
      inline void SkipTerm() // can't be applied if the iterator is empty
       {
        // SIMPLE_ITER can't be empty here
        CHECK1(!(SIMPLE_ITER::Empty()),"1 in PREFIX::ITER_WITH_BACKTRACK::SkipTerm()");
        
        SIMPLE_ITER::MemorizeState(state[undo_num]);
        PushCurrVar();
        
        unsigned undo_action = NOTHING;
        SIMPLE_ITER::SkipTerm(); // the only difference with void SkipSym() 
        if (SIMPLE_ITER::Empty()) 
         {
          if (next_continuation - continuation)
           {
            undo_action += POP;
            PopContinuation();
            SIMPLE_ITER::MemorizeState(used_continuation[undo_num]);
           }
          else
           {
            undo[undo_num] = (ACTION_TO_UNDO)undo_action;
            undo_num++;
            return; // empty iterator now
           };
         };
        if (SIMPLE_ITER::CurrPos()->Head().IsVariable())
         {
          curr_var = SIMPLE_ITER::CurrPos()->Head().Var();
          void* bind = curr_var->Binding();
          if (bind) // bound variable
           {
            check_bind_type: 
             if (VAR::IsVar(bind))
              {
               curr_var = (VAR*)bind;
               bind = curr_var->Binding();
               if (bind) { goto check_bind_type; };
              }
             else // variable bound by a complex term
              {
               curr_var = (VAR*)0;
               SIMPLE_ITER::SkipSym();
               if (SIMPLE_ITER::NotEmpty())
                {
                 undo_action += PUSH;
                 PushContinuation();
                };
               SIMPLE_ITER::Reset((PREFIX_SYM*)bind);
              };
           };
         }
        else
         {
          curr_var = (VAR*)0;
         };
        undo[undo_num] = (ACTION_TO_UNDO)undo_action;
        undo_num++;
       }; // void SkipTerm() 
     
      inline void Backtrack()
       {
        CHECK1(undo_num,"A1 PREFIX::ITER_WITH_BACKTRACK::Backtrack()");

        undo_num--;
        SIMPLE_ITER::RestoreState(state[undo_num]);
        PopCurrVar();
        switch (undo[undo_num])
         {
          case NOTHING: break;
          case POP:
           next_continuation->RestoreState(used_continuation[undo_num]);
           next_continuation++;
           break;
          case PUSH:
           next_continuation--;
           break;
          case POP_AND_PUSH: 
           (next_continuation-1)->RestoreState(used_continuation[undo_num]);
           break;
          #ifdef _DEBUG1_
           default: DebugError("default in PREFIX::ITER_WITH_BACKTRACK::Backtrack()"); 
          #endif 
         };
       }; // void Backtrack()
    }; // class PREFIX::ITER_WITH_BACKTRACK
  public: // methods for debugging
   static inline bool Equal(PREFIX_SYM* word1,PREFIX_SYM* word2)
    {
     ITER iter1;
     ITER iter2;
     iter1.Reset(word1);
     iter2.Reset(word2);
     while (!(iter1.Empty()))
      {
       if (iter1.CurrVar())
        {
         if (iter1.CurrVar() != iter2.CurrVar()) { return false; };
        }
       else
        {
         if ((iter2.CurrVar()) 
             || (iter1.CurrSubterm()->Head() != iter2.CurrSubterm()->Head()))
          { return false; };
        };
       
       iter1.SkipSym();
       iter2.SkipSym();
      };
     return true;
    }; 
    
   static inline bool CorrectComplexTerm(PREFIX_SYM* word)
    {
     if (!word) { DebugWarning("word = NULL in PREFIX::CorrectComplexTerm()"); return false; };
     if (!(word->Head().IsFunctor()))
      {
       DebugWarning("word->Head() is not a functor in PREFIX::CorrectComplexTerm()");
       return false;
      };
     if (!(word->Bracket()))
      {
       DebugWarning("null bracket in PREFIX::CorrectComplexTerm()");
       return false;
      };
     if (word->Bracket() <= word) 
      {
       DebugWarning("corrupt bracket pointer (too small) in PREFIX::CorrectComplexTerm()");
       
       //cout << (long)word << " "; word->Output(); cout << "\n"; // debug
       
       return false;
      };
     if (word->Bracket() - word > MAX_TMP_TERM_WEIGHT + 1)
      {
       DebugWarning("corrupt bracket pointer (too big) in PREFIX::CorrectComplexTerm()");
       return false;
      };
     
     PREFIX_SYM* end = word->Bracket();
     int holes = word->Head().Func().Arity();
     word++;
     while (word < end)
      {
       holes--;
       if (holes < 0) 
        {
         DebugWarning("too long word in PREFIX::CorrectComplexTerm()");
         return false; 
        };
       if (word->Head().IsFunctor())
        {
         holes += word->Head().Func().Arity();
        }
       else
        {
         if (!(VAR::IsVar(word->Head().Var()))) 
          {
           DebugWarning("strange symbol in PREFIX::CorrectComplexTerm()");
           return false;
          };
        };
       word++;
      };
     if (holes) 
      {
       DebugWarning("too short word in PREFIX::CorrectComplexTerm()");
       return false;
      };
     return true;
    };
    
   static inline void OutputWord(PREFIX_SYM* word)
    {
     SIMPLE_ITER w_iter;
     w_iter.Reset(word);
     while (!(w_iter.Empty()))
      {
       Output("<"); Output((long unsigned int)w_iter.CurrPos()); Output("> ");
       w_iter.CurrPos()->Output();
       Output("\n");
       w_iter.SkipSym();
      };
    };
   static inline void OutputWordWithSubst(PREFIX_SYM* word)
    {
     ITER w_iter;
     w_iter.Reset(word);
     while (!(w_iter.Empty()))
      {
       if (w_iter.CurrVar())
        {
         w_iter.CurrVar()->Output();
        }
       else
        {
         w_iter.CurrSubterm()->Output();
        };
       Output("\n");
       w_iter.SkipSym();
      };
    };
   static inline void OutputTerm(PREFIX_SYM* word)
    {
     if (word->Head().IsVariable())
      {
       word->Head().Output();
       return;
      };
     SIMPLE_ITER w_iter;
     BRACKET_COUNTER brack_counter;
     w_iter.Reset(word);
     while (!(w_iter.Empty()))
      {
       w_iter.CurrPos()->Head().Output();
       if ((w_iter.CurrPos()->Head().IsFunctor()) 
           && (w_iter.CurrPos()->Head().Func().Arity()))
        {
         brack_counter.Open(w_iter.CurrPos()->Head().Func().Arity());
         ::Output("(");
         w_iter.SkipSym();
        }
       else
        {
         unsigned closed = brack_counter.Close();
         if (closed)
          {
           while (closed) { ::Output(")");  closed--; };
          };
         w_iter.SkipSym();
         if (!(w_iter.Empty()))
          {
           ::Output(",");
          }; 
        };
      };
    };
  public:
   class OPEN_WORD
    {
     private:
      PREFIX_SYM* begin;
      PREFIX_SYM* curr_pos;
     public:
      inline void Reset(PREFIX_SYM* const memory) { begin = memory; curr_pos = memory; };
      inline PREFIX_SYM* Beginning() const { return begin; };
      inline PREFIX_SYM* CurrPos() const { return curr_pos; };
      inline void Write(const TERM f)
       {
        curr_pos->Head().MkFunc(f);
        if (f.Arity())
         {
          curr_pos->ResetBracket();
         }
        else
         {
          curr_pos->SetBracket(curr_pos+1);
         };
        curr_pos++;
       };
      inline void Write(VAR* const v)
       {
        curr_pos->Head().MkVar(v);
        curr_pos->SetBracket(curr_pos+1);
        curr_pos++;
       };
      inline void Close() { begin->SetBracket(curr_pos); };
      inline void Backtrack(const unsigned n) 
       {
        curr_pos -= n; 
        for (PREFIX_SYM* p = begin; p<curr_pos; p++)
         {
          if (p->Bracket() > curr_pos) { p->ResetBracket(); };
         };
       };
    }; // class PREFIX::OPEN_WORD
  public:
   static inline void CorrectBracket(PREFIX_SYM* const word)
    {
     if (!(word->Bracket()))
      {
       PREFIX_SYM* curr_pos = word+1; 
       BRACKET_COUNTER brack_counter;
       brack_counter.Open(word->Head().Func().Arity());
       do
        {
         if (curr_pos->Head().IsVariable() || (!(curr_pos->Head().Func().Arity())))
          {
           brack_counter.Close();
          }
         else { brack_counter.Open(curr_pos->Head().Func().Arity()); };
         curr_pos++;
        }
       while (brack_counter.IsOpen());
       word->SetBracket(curr_pos);
      };
    }; // void CorrectBracket(PREFIX_SYM* word)
  public: 
   static inline void ResetCorruptedBrackets(PREFIX_SYM* word,
                                             PREFIX_SYM* const changed_suffix,
                                             PREFIX_SYM* const old_end_of_word,
                                             PREFIX_SYM* const new_end_of_word)
    {
     while (word < changed_suffix)
      {
       if (word->Bracket() > changed_suffix)
        {
         if (word->Bracket() == old_end_of_word)
          {
           word->SetBracket(new_end_of_word);
          }
         else
          {
           word->ResetBracket();
          };
        };
       word++;
      };
    }; // void ResetCorruptedBrackets(..........
 }; // class PREFIX
 
class OCCUR_CHECK 
 {
  public:
   static inline bool Occurs(VAR* v,PREFIX_SYM* word)
    {
     DEBUG0_OP(PREFIX_SYM* debug_top_level_word = word);
     static LARGE_VAR_COUNTER link_counter;
     link_counter.Reset();
     void* links[MAX_TMP_TERM_WEIGHT];
     void** next_link = links;
     PREFIX::SIMPLE_ITER iter;
     void* link;
     
     check_top_level:
      iter.Reset(word);
      iter.SkipSym();
     
      while (!(iter.Empty()))
       {
        PREF_SYM sym  = iter.CurrPos()->Head();
        if (sym.IsVariable())
         {
          link = sym.Var()->Binding();
          if (link) // bound variable
           {
            if (!(link_counter.Register(sym.Var())))
             {
              *next_link = link;
              next_link++;
             };
           }
          else // free variable
           {
            if (sym.Var() == v) 
             {
              CHECK0(OccursIneff(v,debug_top_level_word),"A1 in OCC_CHECK::Occurs()");
              return true; 
             };
           };
         };
        iter.SkipSym();
       }; 
     
      try_next_link:
       if (next_link == links) 
        {
         CHECK0(!(OccursIneff(v,debug_top_level_word)),"A2 in OCC_CHECK::Occurs()");
         return false; 
        };
       next_link--;
       link = *next_link;
       check_link:
        if (VAR::IsVar(link))
         {
          if (((VAR*)link)->Binding()) 
           {
            if (!(link_counter.Register((VAR*)link)))
             {
              link = ((VAR*)link)->Binding();
              goto check_link;
             }
            else
             {
              goto try_next_link;
             };
           }
          else // free var
           {
            if (((VAR*)link) == v) 
             {
              CHECK0(OccursIneff(v,debug_top_level_word),"A3 in OCC_CHECK::Occurs()");
              return true; 
             };
            goto try_next_link;
           };
         }
        else // complex term
         {
          word = (PREFIX_SYM*)link;
          goto check_top_level;
         };
    }; // bool Occurs(VAR* v,PREFIX_SYM* word)
    
   /************** Inefficient but simple version: *******/
   static inline bool OccursIneff(VAR* v,PREFIX_SYM* word) 
    // word must represent a complex term, v must be a nonnull pointer
    {
     CHECK1(v,"null argument <VAR* v> in OCCUR_CHECK::OccursIneff()");
     CHECK1(VAR::IsVar(v),"strange argument <VAR* v> in OCCUR_CHECK::OccursIneff()");
     CHECK1(v,"bound variable in OCCUR_CHECK::OccursIneff()");
     CHECK2(PREFIX::CorrectComplexTerm(word),"A1 in in OCCUR_CHECK::OccursIneff()");
     
     PREFIX::ITER iter;
     iter.Reset(word);
     next_sym:
      iter.SkipSym();
      if (iter.Empty())
       {
        return false;
       }
      else
       {
        if (iter.CurrVar() == v)
         {
          return true;
         }
        else
         {
          goto next_sym;
         };
       };
    }; // bool OccursIneff(VAR* v,PREFIX_SYM* word) 
   /********************************************************/
 }; // class OCCUR_CHECK


//*********************************************************************

class ST_TERM_WITH_SUBST
 {
  private:
   unsigned var_offset;
   PREFIX::ITER word;
   TERM::ITERATOR top;
   bool top_level;
   bool empty;
   PREF_SYM curr_sym;
  public:
   ST_TERM_WITH_SUBST() : word(), top() {};
   inline void Reset(const int index,TERM* t) 
    // t must be a complex term here 
    {
     top.Reset(t);
     var_offset = VAR::Offset(index);
     top_level = true;
     empty = false;
     curr_sym.MkFunc(top.Symbol());
    };
   inline bool Empty() const { return empty; };
   inline PREF_SYM CurrSym() const { return curr_sym; };
   inline TERM* CurrPosInBase() const { return top.CurrPos(); }; 
   
   inline void SkipTerm()
    {
     if (top_level)
      {
       if ((empty = !top.SkipTerm())) return;
       top_check_sym:
        if (top.CurrentSymIsVar())
         {
          bool free_var;
          void* bind = (VAR::Base(top.Symbol().Var()) + var_offset)->Unref(free_var);
          if (free_var)
           {
            curr_sym.MkVar((VAR*)bind);
           }
          else
           {
            top_level = false;
            word.Reset((PREFIX_SYM*)bind);
            curr_sym = word.CurrSubterm()->Head();
           };
         }
        else // functor
         {
          curr_sym.MkFunc(top.Symbol());
         };
      }
     else // in the substitution
      {
       word.SkipTerm();
       if (word.Empty())
        {
         top_level = true;
         if (!top.Next()) return;
         goto top_check_sym;
        }
       else
        {
         if (word.CurrVar())
          {
           curr_sym.MkVar(word.CurrVar());
          }
         else
          {
           curr_sym = word.CurrSubterm()->Head();
          };
        };
      };
    };
    
     
   inline void SkipSym()
    {
     bool free_var;
     void* bind;
     if (top_level)
      {
       top_next_sym:
        if (top.Next())
         {
          if (top.CurrentSymIsVar())
           {
            bind = (VAR::Base(top.Symbol().Var()) + var_offset)->Unref(free_var);
            if (free_var)
             {
              curr_sym.MkVar((VAR*)bind);
             }
            else
             {
              top_level = false;
              word.Reset((PREFIX_SYM*)bind);
              curr_sym = word.CurrSubterm()->Head();
             };
           }
          else // functor
           {
            curr_sym.MkFunc(top.Symbol());
           };
         }
        else { empty = true; };
      }
     else
      {
       word.SkipSym();
       if (word.Empty())
        {
         top_level = true;
         goto top_next_sym;
        }
       else
        {
         if (word.CurrVar())
          {
           curr_sym.MkVar(word.CurrVar());
          }
         else
          {
           curr_sym = word.CurrSubterm()->Head();
          };
        };
      };
    };
 }; // class ST_TERM_WITH_SUBST


//******************************************************************************

inline void StandardTermToPrefix(const int index,TERM* const t,PREFIX_SYM*& memory)
 // t must be a complex term here
 {
  TERM::ITERATOR t_iter;
  BRACKET_COUNTER brack_counter;
  PREFIX_SYM* left_bracket[MAX_TERM_WEIGHT];
  PREFIX_SYM** next_left_bracket = left_bracket;
  t_iter.Reset(t);
  TERM sym;
  unsigned closed;
  unsigned var_offset = VAR::Offset(index);
  next_sym:
   sym = t_iter.Symbol();
   if (sym.IsVariable())
    {
     memory->Head().MkVar(VAR::Base(sym.Var())+var_offset);
     memory->SetBracket(memory+1);
     closed = brack_counter.Close();    
     while (closed)
      {
       next_left_bracket--;
       (*next_left_bracket)->SetBracket(memory+1);
       closed--;
      };
    }
   else // functor
    {
     memory->Head().MkFunc(sym);
     if (sym.Arity())
      {
       brack_counter.Open(sym.Arity());
       *next_left_bracket = memory;
       next_left_bracket++;
      }
     else // constant
      {
       memory->SetBracket(memory+1);
       unsigned closed = brack_counter.Close();    
       while (closed)
        {
         next_left_bracket--;
         (*next_left_bracket)->SetBracket(memory+1);
         closed--;
        };  
      };
    };
   memory++;
   if (t_iter.Next())
    {
     goto next_sym;
    };
  
  CHECK1(next_left_bracket == left_bracket,"in StandardTermToPrefix(int index,...)"); 
   
 }; // void StandardTermToPrefix(int index,TERM* t,PREFIX_SYM*& memory)
 
inline void StandardTermToPrefix(int index,TERM t,PREFIX_SYM*& memory) 
 {
  if (t.IsVariable())
   {
    memory->Head().MkVar(VAR::Var(t.Var(),index));
    memory->SetBracket(memory + 1);
    memory++;
   }
  else StandardTermToPrefix(index,t.First(),memory); 
 };
 
//***************************************************************************** 
 
 
 
 
 
 
 
 
 