#define __PROBLEM_PROFILE_H

#ifndef __DEBUG_H
 #include "debug.h"
#endif
#ifndef __VAMPIRE_H
 #include "vampire.h"
#endif
#ifndef __CLAUSE_H
 #include "clause.h"
#endif
#ifndef __TERM_H
 #include "term.h"
#endif
#ifndef __OUTPUT_H
 #include "output.h"
#endif

//************************************************************

class PRED_INFO
 {
  public:
   enum STATUS { UNCKNOWN, NOT_ELIMINATED, ELIMINATED };
   static inline void OutputStatus(STATUS s)
    {
     switch (s)
      {
       case UNCKNOWN: ::Output("UNCKNOWN"); break;
       case NOT_ELIMINATED: ::Output("NOT_ELIMINATED"); break;
       case ELIMINATED: ::Output("ELIMINATED"); break;
      };
    };
  public:
   TERM Symbol;
   STATUS Status; 
   CLAUSE_LIST* Clauses;
   bool PositiveNondef;
   bool NegativeNondef;  
  public:
   PRED_INFO() {};
   inline void Init() 
    {
     Status = UNCKNOWN; 
     Clauses = (CLAUSE_LIST*)0; 
     PositiveNondef = false;
     NegativeNondef = false;
    };
   inline void Output()
    {
     ::Output(Symbol); ::Output("{ ar: "); ::Output(Symbol.Arity());
     ::Output(",st: "); OutputStatus(Status); ::Output("}");
    }; 
 }; // class PRED_INFO 


class PRED_INFO_TABLE
 {
  private:
   PRED_INFO info[MAX_NUMBER_OF_FUNCTORS];
   bool is_registered[MAX_NUMBER_OF_FUNCTORS];
   STACK_WITH_ITERATOR<int,MAX_NUMBER_OF_FUNCTORS> all_registered;
  public:
   PRED_INFO_TABLE() : all_registered("MAX_NUMBER_OF_FUNCTORS in PRED_INFO_TABLE::all_registered")
    {
     for (int p = 0; p < MAX_NUMBER_OF_FUNCTORS; p++) 
      {
       is_registered[p] = false;
      };
    };
   inline bool IsRegistered(int p) const { return is_registered[p]; };
   inline void Register(int p) 
    {
     if (!IsRegistered(p)) { is_registered[p] = true; all_registered.Push(p); info[p].Init(); };           
    };
   inline PRED_INFO& Info(int p) 
    { 
     if (!IsRegistered(p)) { Register(p); }; 
     return info[p]; 
    };
   inline void ResetIter() { all_registered.ResetIterator(); };
   inline bool Next(int& p)
    {
     int* next = all_registered.Next();
     if (next) { p = *next; return true; };
     return false;
    };
   inline void Output()
    {
     ResetIter();
     int p;
     while (Next(p)) 
      {
       Info(p).Output(); ::Output(" ");
      };
    };
 }; // class PRED_INFO_TABLE


class CLAUSE_PROFILE
 {
  public:
   long num_of_lits;
   long num_of_pos_lits;
   long num_of_neg_lits;  
   long num_of_eq_lits;
   long num_of_neq_lits;
  public:
   CLAUSE_PROFILE() { Reset(); };
   inline void Reset()
    {
     num_of_lits = 0;
     num_of_pos_lits = 0;
     num_of_neg_lits = 0;  
     num_of_eq_lits = 0;
     num_of_neq_lits = 0;   
    }; 
   inline void Analyse(Clause* cl,PRED_INFO_TABLE& pred_info_table)
    {
     Reset();
     num_of_lits = cl->Length();
     for (TERM* lit_ref = cl->Literals()->Args(); !(lit_ref->IsNull()); lit_ref = lit_ref->Next())
      {
       TERM* lit = lit_ref->First();
       if (lit->Positive()) { num_of_pos_lits++; } else num_of_neg_lits++;
       if (lit->IsEquality()) { num_of_eq_lits++; } 
       else 
        {
         TERM header = *lit;
         header.SetPolarity(POL_POS);
         int pred = header.Functor();
         pred_info_table.Info(pred).Symbol = header;
         if (!header.Arity()) pred_info_table.Info(pred).Status = PRED_INFO::NOT_ELIMINATED;
         if ((pred_info_table.Info(pred).Clauses) && (pred_info_table.Info(pred).Clauses->First() == cl))
          pred_info_table.Info(pred).Status = PRED_INFO::NOT_ELIMINATED; 
         pred_info_table.Info(pred).Clauses = new CLAUSE_LIST(cl,pred_info_table.Info(pred).Clauses);
         
         if (pred_info_table.Info(pred).Status == PRED_INFO::UNCKNOWN)
          { 
           if (lit->Positive())
            {
             pred_info_table.Info(pred).PositiveNondef =
               pred_info_table.Info(pred).PositiveNondef 
               || (!lit->Plain()); 
            }
           else
            {
             pred_info_table.Info(pred).NegativeNondef =
               pred_info_table.Info(pred).NegativeNondef 
               || (!lit->Plain()); 
            };
           if (pred_info_table.Info(pred).PositiveNondef && pred_info_table.Info(pred).NegativeNondef)
            pred_info_table.Info(pred).Status = PRED_INFO::NOT_ELIMINATED;
          };

         num_of_neq_lits++;
        };   
      };
    }; 
   // some sugar
   inline bool Horn() const { return (num_of_pos_lits <= 1); };
   inline bool Eq() const { return (bool)num_of_eq_lits; }; 
   inline bool Noneq() const { return (bool)num_of_neq_lits; };
   inline bool Unit() const { return (num_of_lits == 1); };   
 }; // class CLAUSE_PROFILE

class PROBLEM_PROFILE
 {
  public:
   enum CATEGORY { HNE, NNE, HEQ, NEQ, PEQ, UEQ };
  private:
   CLAUSE_PROFILE clause_profile;
   PRED_INFO_TABLE pred_info_table;
   bool eliminated_pred[MAX_NUMBER_OF_FUNCTORS];
  public:
   long num_of_clauses;
   long num_of_nonhorn_clauses;
   long num_of_eq_clauses;
   long num_of_noneq_clauses;
   long num_of_nonunit_clauses; 
   long num_of_eliminated_pred;  
   long num_of_not_eliminated_pred;
  public:
   PROBLEM_PROFILE() : clause_profile(), pred_info_table() { Reset(); };
   inline void Reset()
    {
     num_of_clauses = 0;
     num_of_nonhorn_clauses = 0;
     num_of_eq_clauses = 0;
     num_of_noneq_clauses = 0;
     num_of_nonunit_clauses = 0;
     num_of_eliminated_pred = 0;  
     num_of_not_eliminated_pred = 0;
     for (int p = 0; p < MAX_NUMBER_OF_FUNCTORS; p++) eliminated_pred[p] = false;
    }; 
   inline void Analyse(Clause* cl)
    {
     num_of_clauses++;
     clause_profile.Analyse(cl,pred_info_table);
     if (!clause_profile.Horn()) num_of_nonhorn_clauses++;
     if (clause_profile.Eq()) num_of_eq_clauses++;
     if (clause_profile.Noneq()) num_of_noneq_clauses++;
     if (!clause_profile.Unit()) num_of_nonunit_clauses++;
    };

   inline PRED_INFO_TABLE& PredInfoTable() { return pred_info_table; };
   inline bool* EliminationTable() { return eliminated_pred; };
   inline long NumOfEliminatedPred() const { return num_of_eliminated_pred; };  
   inline long NumOfNotEliminatedPred() const { return num_of_not_eliminated_pred; }; 
  
   inline bool Horn() const { return !num_of_nonhorn_clauses; };
   inline bool Eq() const { return (bool)num_of_eq_clauses; };
   inline bool PureEq() const { return !num_of_noneq_clauses; };
   inline bool Unit() const { return !num_of_nonunit_clauses; };  
   inline CATEGORY Category() const
    {
     if (Eq())
      {
       if (PureEq())
        {
         if (Unit()) { return UEQ; } else return PEQ; 
        }
       else
        if (Horn()) { return HEQ; } else return NEQ; 
      }
     else // no equality
      if (Horn()) { return HNE; } else return NNE;
    };


   inline void MarkEliminated()
    {
     pred_info_table.ResetIter();
     int pred;
     while (pred_info_table.Next(pred))
      {
       if (pred_info_table.Info(pred).Status == PRED_INFO::UNCKNOWN)
        {
         for (CLAUSE_LIST* clist = pred_info_table.Info(pred).Clauses; clist; clist = clist->Next())
          for (TERM* lit_ref = clist->First()->Literals()->Args(); !lit_ref->IsNull(); lit_ref = lit_ref->Next())
           if (!lit_ref->First()->IsEquality())
            pred_info_table.Info(lit_ref->First()->Functor()).Status = PRED_INFO::NOT_ELIMINATED;          

         pred_info_table.Info(pred).Status = PRED_INFO::ELIMINATED;
         eliminated_pred[pred] = true;
         num_of_eliminated_pred++;
        }
       else num_of_not_eliminated_pred++;
      };
    };

   inline void Output()
    {
     ::Output("Total number of clauses:                       "); ::Output(num_of_clauses); ::Output("\n");     
     ::Output("Number of nonhorn clauses:                     "); ::Output(num_of_nonhorn_clauses); ::Output("\n"); 
     ::Output("Number of clauses with equality:               "); ::Output(num_of_eq_clauses); ::Output("\n"); 
     ::Output("Number of clauses with nonequational literals: "); ::Output(num_of_noneq_clauses); ::Output("\n"); 
     ::Output("Number of nonunit clauses:                     "); ::Output(num_of_nonunit_clauses); ::Output("\n"); 
     ::Output("Noneq. predicates: "); pred_info_table.Output(); ::Output("\n"); 
     ::Output("Number of eliminated predicates:               "); ::Output(num_of_eliminated_pred); ::Output("\n");
     ::Output("Number of not eliminated predicates:           "); ::Output(num_of_not_eliminated_pred); ::Output("\n"); 
     ::Output("********* Category: ");
     switch (Category())
      {
       case HNE: ::Output("HNE"); break;
       case NNE: ::Output("NNE"); break;
       case HEQ: ::Output("HEQ"); break;
       case NEQ: ::Output("NEQ"); break;
       case PEQ: ::Output("PEQ"); break;
       case UEQ: ::Output("UEQ"); break; 
      };
     ::Output(" ***************************\n");
    };  
 }; // class PROBLEM_PROFILE


//***********************************************************
