#define __STACK_H

#ifndef __DEBUG_H
 #include "debug.h"
#endif
#ifndef __VAMPIRE_H
 #include "vampire.h"
#endif


template <class C,int I>

class STACK
{
  // made public temporarily, since used in
  // KBOrder, should be using friend mechanism instead
  public:  // structure
    #ifdef _DEBUG_ 
    const char* name;
    C* after;
    #endif
    C stack[I];
    C* next;
  public:
    #ifndef _DEBUG_
    STACK(const char*) :
      next(stack) {}
    #else
    STACK(const char* nm) :
      name(nm),
      after(stack+I),
      next(stack)
      {}
    #endif
    
    STACK() 
     {
      #ifdef _DEBUG_ 
      name = "uncknown name";
      #endif
      Init();
     };
    
    inline void Init() 
     {
      next = stack;
      #ifdef _DEBUG_ 
      after = stack + I;
      #endif 
     }; 
    
    inline void Initialize()
      { next = stack; }
       
    inline STACK& operator=(STACK& st)
     {
      const int size = st.NumberOfElements();
      next = stack + size;
      for (int i = 0; i < size; i++)
       {
        stack[i] = st.stack[i];
       };
      return *this;
     };  
       
    inline void Push(C elem)
    {
      #ifdef _DEBUG_
      if (next >= after) 
       { 
	DebugError("Parameter ",name," exceeded");
       };
      #endif
      *next = elem;
      next++;
    }

    inline int Empty() const
      // cast to avoid problems with const
      { return (next == (C*)stack); }

    inline bool Full() { return (next >= stack + I); };

    inline C Pop()
      {
       #ifdef _DEBUG_
       if (next == stack) 
        {
         DebugError("Poping empty stack of size ",name," ");
        };
       #endif
       
       next--;
       return *next;
      };

    inline void Pop(int n)
      { next -= n; }

    inline void Reset()
      { next = stack; }
    
    inline C& Head() 
     {
      return *next;
     };
    inline C* HeadAddr() 
     {
      return next;
     };
    
    inline C& Bottom() 
     {
      return *stack;
     };
    inline C* BottomAddr() { return stack; };
     
    inline void Next() { next++; };
    inline void Previous() { next--; };

    inline C& Top() const
      { return next[-1]; }

    inline C& Nth(int n) const // n must be non-negative
      { return next[-1 - n]; }

    inline void Replace(C newelem)
      { *(next-1)=newelem; }

    inline void MoveTopIn()
    {
      next--;
      next[-1] = *next;
    };

    inline C* Mark() const
    { return next; };

    inline bool OnMark(C* mark)
     { return (next == mark); };

    inline void RestoreMark(C* mark)
      { next = mark; };

    inline int NumberOfElements() const
      { return (int)(((C*)next) - ((C*)stack)); };
      
};  // class STACK


template <class C,int I>

class STACK_WITH_ITERATOR
  : public STACK <C,I>
{
  private:
    C* iterator;

  public:

    STACK_WITH_ITERATOR(const char* name) : STACK <C,I>(name) {};
    STACK_WITH_ITERATOR() 
     {
      STACK_WITH_ITERATOR("<< no name >>"); 
     };

    inline void ResetIterator()
    {
      iterator = stack;
    }
    
    inline void ResetBackIterator()
     {
      iterator = next;
     };

    inline void ResetIterator(C* mark)
    {
      iterator = mark;
    }

    inline C* Next()
    {
      if (iterator == next)  // no more elements
	return 0;
      
      return iterator++;
    }
    inline C* Previous()
     {
      if (iterator == stack) { return 0; };
      iterator--;
      return iterator;
     };
};

template <class C,int I>
 class STACK_AS_SET : public STACK_WITH_ITERATOR <C,I>
  {
   public:
    STACK_AS_SET(const char* name) : STACK_WITH_ITERATOR <C,I>(name) {};
    bool Contains(const C x)
     {
      ResetIterator();
      C* p = Next();
      while (p)
       {
        if (*p == x) { return true; };
        p = Next();
       };
      return false;
     };
  };


template <class C,int I>
 class UNLIMITED_STACK
  {
   private:
    UNLIMITED_STACK<C,I>* prevSegment;
    STACK<C,I>* currSegment;
   public:
    UNLIMITED_STACK()
     : prevSegment((UNLIMITED_STACK<C,I>*)0)
     {
      currSegment = new STACK<C,I>("<< unlimited >>");
     };
    ~UNLIMITED_STACK() {};
    inline bool Empty() { return ((currSegment->Empty()) && (!prevSegment)); };
    inline void Push(C x)
     {
      if (currSegment->Full())
       {
        UNLIMITED_STACK<C,I>* tmp = new UNLIMITED_STACK<C,I>();
        *tmp = *this;
        prevSegment = tmp;
        currSegment = new STACK<C,I>("<< unlimited >>");
       };
      currSegment->Push(x);
     };
    inline C Pop()
     {
      if (currSegment->Empty())
       {
        delete currSegment;
        UNLIMITED_STACK<C,I>* tmp = prevSegment;
        *this = *tmp;
        delete tmp;
       };
      return currSegment->Pop();
     };
    inline void Reset()
     {
      UNLIMITED_STACK<C,I>* tmp;
      while (prevSegment)
       {
        delete currSegment;
        tmp = prevSegment;
        *this = *tmp;
        delete tmp;
       };
      currSegment->Reset();
     };     
  }; // class UNLIMITED_STACK<class C,int I>



//*************************************************************************



