//
//      Functions of this file must be reimplemented
//

#ifdef __BORLANDC__  // PC
#  include <sys\timeb.h>
#else  // not PC
#  include <sys/time.h>
#  include <sys/resource.h>
#  include <sys/syscall.h>
#endif


#include "timer.h"
#include "output.h"


#ifdef TIMER_CALIBRATE
#  define CALLS_PER_SECOND 2337.13
#endif


enum TIMERS
	{
	BACK_SUBSUMPTION_TIMER,
	FORWARD_SUBSUMPTION_TIMER,
	RESOLUTION_TIMER,
	OUTPUT_TIMER,
	OTHER_TIMER,	// must be the last one

	NUMBER_OF_TIMERS
	};

const char* timer_names[] =
  {
  "Back subsumption",
  "Forward subsumption",
  "Resolution",
  "Output",
  "Other"
  };


struct TIMER_FULL_TIME
{
  long milliseconds;
  long seconds;
  long minutes;
  long hours;
  long days;

  TIMER_FULL_TIME(long milliseconds);
};


TIMER_FULL_TIME::TIMER_FULL_TIME(long msec)
{
  milliseconds = msec % 1000;
  seconds = msec/1000;

  minutes = seconds/60;
  seconds %= 60;

  hours = minutes/60;
  minutes %= 60;

  days = hours/24;
  hours %= 24;
}  // TIMER_FULL_TIME::TIMER_FULL_TIME()


class TIMER
{
  private:
    int stopped;
    unsigned long calls;
    long total_seconds;
    long total_milliseconds;
    long mark_seconds;
    long mark_milliseconds;

    long TotalTime();
    inline static void CpuTime(long &seconds,long &milliseconds);
    inline void Start(long sec,long msec);
    inline void Stop(long sec,long msec);

  public:
    TIMER() :
      stopped(1),
      calls(0),
      total_seconds(0),
      total_milliseconds(0)
      {}

    void Start();
    void Stop();
    void ShowTime(const char*);
    static void ShowAllTimers();
};


TIMER timers[NUMBER_OF_TIMERS];
TIMER* current_timer = &timers[OTHER_TIMER];


#ifdef __BORLANDC__  // PC

inline void TIMER::CpuTime(long &sec, long &msec)
{
  timeb tm;

  ftime(&tm);
  sec = tm.time;
  msec = tm.millitm;
}  // CpuTime()

#else // not a PC

inline void TIMER::CpuTime(long &sec, long &msec)
{
  rusage r;
  #ifdef HP_UX  // HP
  syscall(SYS_GETRUSAGE,RUSAGE_SELF, &r);
  #else // Sun
  // this use of getrusage is not valid any more
  // getrusage(RUSAGE_SELF, &r);
  #endif

  sec = r.ru_utime.tv_sec;
  msec = r.ru_utime.tv_usec / 1000;
}  // CpuTime()

#endif // PC/Unix


inline void TIMER::Start(long sec,long msec)
{
  calls++;
  mark_seconds = sec;
  mark_milliseconds = msec;
  stopped = 0;
}  // TIMER::Start()


inline void TIMER::Stop(long sec,long msec)
{
  stopped = 1;
  total_seconds += sec - mark_seconds;
  total_milliseconds += msec - mark_milliseconds;
}  // TIMER::Stop()


void TIMER::Start()
{
  if (!stopped)  // was already started
    return;

  calls++;
  stopped = 0;
  CpuTime(mark_seconds,mark_milliseconds);
  #ifndef _FAST_
  if (this != current_timer)  // the other timer should stop
    current_timer->Stop(mark_seconds,mark_milliseconds);
  #endif
}  // TIMER::Start()


void TIMER::Stop()
{
  if (stopped)  // was alredy stopped
    return;

  stopped = 1;

  long sec;
  long msec;

  CpuTime(sec,msec);
  total_seconds += sec - mark_seconds;
  total_milliseconds += msec - mark_milliseconds;
  #ifndef _FAST_
  if (this != current_timer)  // the other timer should stop
    current_timer->Start(sec,msec);
  #endif
}  // TIMER::Stop()


long TIMER::TotalTime()
{
  int timer_msec;
  timer_msec = 0;
  int was_stopped = stopped;

  if (!stopped)
    Stop();

  long result = total_milliseconds + total_seconds*1000 - timer_msec;

  if (!was_stopped)
    Start();

  return result < 0 ? 0 : result;
}  // TIMER::TotalTime()


void TIMER::ShowTime(const char* timer)
{
  long totalTime = TotalTime();

  Output(timer);
  Output(" time: ");
  Output(totalTime / 1000);
  Output('.');
  OutputWidth(3);
  char fill = OutputFill('0');
  Output(totalTime % 1000);
  Output('\n');
  OutputFill(fill);

}  // TIMER::ShowTime()


void TIMER::ShowAllTimers()
{
  long totalTime = 0;
  for (int i = 0;i < NUMBER_OF_TIMERS;i++)
    {
    long time = timers[i].TotalTime();
    if (time)
      {
      timers[i].ShowTime(timer_names[i]);
      totalTime += time;
      }
    }

  TIMER overall;
  overall.total_milliseconds = totalTime;
  overall.ShowTime("Overall");

  Output(" (");

  TIMER_FULL_TIME ftm(totalTime);

  if (ftm.days)
    {
    Output(ftm.days);
    Output(" days, ");
    Output(ftm.hours);
    Output(" hours, ");
    Output(ftm.minutes);
    Output(" minutes");
    }
  else if (ftm.hours)
    {
    Output(ftm.hours);
    Output(" hours, ");
    Output(ftm.minutes);
    Output(" minutes, ");
    Output(ftm.seconds);
    Output(" seconds");
    }
  else if (ftm.minutes)
    {
    Output(ftm.minutes);
    Output(" minutes, ");
    Output(ftm.seconds);
    Output(" seconds, ");
    Output(ftm.milliseconds);
    Output(" milliseconds");
    }
  else if (ftm.seconds)
    {
    Output(ftm.seconds);
    Output(" seconds, ");
    Output(ftm.milliseconds);
    Output(" milliseconds");
    }
  else
    {
    Output(ftm.milliseconds);
    Output(" milliseconds");
    }

  Output(")\n");
}  // TIMER::ShowAllTimers();


// All exported functions start here
void StartForwardSubsumptionTimer()
  { timers[FORWARD_SUBSUMPTION_TIMER].Start(); }

void StopForwardSubsumptionTimer()
  { timers[FORWARD_SUBSUMPTION_TIMER].Stop(); }

void StartResolutionTimer()
  { timers[RESOLUTION_TIMER].Start(); }

void StopResolutionTimer()
  { timers[RESOLUTION_TIMER].Stop(); }

void StartOutputTimer()
  { timers[OUTPUT_TIMER].Start(); }

void StopOutputTimer()
  { timers[OUTPUT_TIMER].Stop(); }

void StartBackSubsumptionTimer()
  { timers[BACK_SUBSUMPTION_TIMER].Start(); }

void StopBackSubsumptionTimer()
  { timers[BACK_SUBSUMPTION_TIMER].Stop(); }

void StartMainTimer()
  { timers[OTHER_TIMER].Start(); }

void ShowAllTimers()
  { TIMER::ShowAllTimers(); }

// current times is the one stoped by ^C
void StartCurrentTimer() {}
void StopCurrentTimer();


/*
  This is an example code for calibrating timers.
  The timer t show the total time elapsed
  Resolution and Other timers show the share of this time (50% on the PC)
  ON the Laptop PC it gives 1160 calls/second

  TIMER t("Calibrate");
  t.Start();

  Timer().Reset();
  Timer().Start();
  for (long l = 0;l < 100000;l++)
    {
    ResolutionTimer().Start();
    ResolutionTimer().Stop();
    }
  t.ShowTime();
  Timer().ShowTime();
  ResolutionTimer().ShowTime();
*/
