#define __TMP_LIT_LIST_H

#ifndef __DEBUG_H
 #include "debug.h"
#endif
#ifndef __VAMPIRE_H
 #include "vampire.h"
#endif 
#ifndef __TERM_H
 #include "term.h"
#endif
#ifndef __STACK_H
 #include "stack.h"
#endif
#ifndef __VARIABLES_H
 #include "variables.h"
#endif
#ifndef __WEIGHTING_H
 #include "weighting.h"
#endif
#ifndef __STATISTICS_H
 #include "statistics.h"
#endif
#ifndef __FLAT_TERM_H
 #include "flat_term.h"
#endif 
#ifndef __TMP_LITERAL_H
 #include "tmp_literal.h"
#endif 


//**********************************************************************

class TMP_LIT_LIST
 {
  private:
   class STAT
    {
     public: 
      int negative;
      int equations;
      int noneq_lits;
      TERM max_noneq_header;
     public:
      STAT() { Reset(); };
      inline void Reset()
       {
        negative = 0;
        equations = 0;
        noneq_lits = 0;
       }; 
    };
  public: 
   class CONCAT;
  private:
   TMP_LITERAL* literal[MAX_NUMBER_OF_LITERALS];
   TMP_LITERAL** next_literal;
  public: 
   TMP_LIT_LIST() : next_literal(literal) {};
   TMP_LIT_LIST(const TMP_LIT_LIST& lst) : next_literal(literal)
    {
     for (TMP_LITERAL* const * lit = lst.literal; lit < lst.next_literal; lit++) Push(*lit);
    };
   inline TMP_LIT_LIST& operator=(const TMP_LIT_LIST& lst)
    {
     next_literal = literal;
     for (TMP_LITERAL* const * lit = lst.literal; lit < lst.next_literal; lit++) Push(*lit);
     return *this;
    };
   
   inline void Reset() { next_literal = literal; };
   inline void Push(TMP_LITERAL* lit) { *next_literal = lit; next_literal++; };
   inline TMP_LITERAL* Pop() { next_literal--; return *next_literal; };
   inline void ReplaceByTop(TMP_LITERAL** pos) { next_literal--; *pos = *next_literal; };
   inline bool Empty() const { return next_literal == literal; };
   inline bool Nonempty() const { return next_literal != literal; };
   inline int Length() const { return next_literal - literal; };
   inline TMP_LITERAL** First() { return literal; };
   inline TMP_LITERAL** Last()  { return next_literal - 1; };
   inline long Weight() const 
    {
     long res = 0L;
     for (const TMP_LITERAL* const * lit = literal; lit < next_literal; lit++) res += (*lit)->Weight();
     return res;
    }; 

   inline bool SubstPossible(int var,unsigned subst_depth,long subst_weight)
    {
     for (TMP_LITERAL** lit = First(); lit <= Last(); lit++)
      if (!(*lit)->SubstPossible(var,subst_depth,subst_weight)) return false;     
     return true;
    };

   inline void WeightAndNumOfPosEq(long& weight,long& num_of_pos_eq)
    {
     weight = 0;
     num_of_pos_eq = 0;
     for (const TMP_LITERAL* const * lit = literal; lit < next_literal; lit++) 
      {
       const TMP_LITERAL* curr_lit = *lit;
       weight += curr_lit->Weight();
       if (curr_lit->IsEquality() && curr_lit->Positive()) num_of_pos_eq++;
      };
    };
    
   inline long MinFactorWeightUpperBound() const { return Weight(); }; 
   inline long MinPosEqLHSWeight()
    { 
     long res = 0;
     for (const TMP_LITERAL* const * lit = literal; lit < next_literal; lit++) 
      if ((*lit)->IsEquality() && (*lit)->Positive()) res += (*lit)->PosEqMinLHSWeight();
     return res; 
    };
    
  private:
   inline void CollectStat(STAT& stat);
      
  public: 
   inline bool Eliminate(TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool* eliminated);
   inline void KBOrder_NEQ(TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits);
   inline void KBOrder_NS_NEQ(TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits);
   inline void KBOrder_NS_EQ(TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits);
   inline void Order(TMP_LIT_LIST& result,const VAR_WEIGHTING& var_weighting);
   inline void CreateVarWeighting(VAR_WEIGHTING& var_weighting);
   inline void RenameVariables(INT_VAR_RENAMING& var_renaming);
   inline void CollectGround(TMP_LIT_LIST& component);
   inline void CollectNextComponent(TMP_LIT_LIST& component);

  
   inline bool Weed();

  public: // for debugging
   inline void Output()
    {
     for (TMP_LITERAL** lit = literal; lit < next_literal; lit++)
      {
       (*lit)->Output();
       ::Output(" % ");
      };
    }; 
 }; // class TMP_LIT_LIST

class TMP_LIT_LIST::CONCAT // only for nonempty lists
 {
  private:
   TMP_LITERAL** next_lit;
   TMP_LITERAL** last_lit;
   FLAT_TERM::ITER iter;
  public:
   CONCAT() : iter() {};
   CONCAT(TMP_LIT_LIST& lst) 
    : next_lit(lst.First() + 1),
      last_lit(lst.Last()),
      iter((*(lst.First()))->FlatTerm())
     {}; 
   inline void Reset(TMP_LIT_LIST& lst)
    {
     next_lit = lst.First() + 1;
     last_lit = lst.Last();
     iter.Reset((*(lst.First()))->FlatTerm());
    };
   inline bool Empty() const
    {
     return ((next_lit > last_lit) && (iter.EndOfTerm()));
    };
   inline bool Nonempty() const
    {
     return ((next_lit <= last_lit) || (iter.NotEndOfTerm()));
    };  
   inline void NextSym() 
    {
     iter.NextSym();
     if (iter.EndOfTerm())
      if (next_lit <= last_lit)
       {
        iter.Reset((*next_lit)->FlatTerm());
        next_lit++;
       }; 
    };
   inline TERM Symbol() const { return iter.CurrPos()->Symbol(); }; 
 }; // class TMP_LIT_LIST::CONCAT


//*************** Definitions for TMP_LIT_LIST: ****************************

inline bool TMP_LIT_LIST::Weed()
 {
  for (TMP_LITERAL** lit = Last(); lit >= First(); lit--)
   {
    TERM header = (*lit)->Header();
    if (header.IsEquality())
     {
      if ((*lit)->EqEqualArgs())
       {
        if (header.Positive()) return false; // tautology
        ReplaceByTop(lit); 
        goto next_lit;
       };   
      for (TMP_LITERAL** another_lit = First(); another_lit < lit; another_lit++)
       if ((*another_lit)->IsEquality())
        {
         if ((*another_lit)->EqualEqLit(**lit))
          {
           ReplaceByTop(lit); 
           goto next_lit; 
          };         
         if ((*another_lit)->ComplementaryEqLit(**lit)) return false; // tautology          
        }; 

      if ((*lit)->IsUnorderedEq()) (*lit)->EqOrderArgs(); 
                      
     }
    else // not an equality
     {
      for (TMP_LITERAL** another_lit = First(); another_lit < lit; another_lit++)
       {
        if ((*another_lit)->Equal(**lit))
          {
           ReplaceByTop(lit); 
           goto next_lit; 
          };         
         if ((*another_lit)->Complementary(**lit)) return false; // tautology         
       };
     };
    next_lit: {}; // nothing to do here, just the end of the loop body
   };
  return true;
 }; // inline bool TMP_LIT_LIST::Weed()


inline void TMP_LIT_LIST::CollectStat(STAT& stat)
 {
  for (TMP_LITERAL** lit = literal; lit < next_literal; lit++)
   {
    if ((*lit)->IsEquality()) { stat.equations++; }
    else
     {
      if (stat.noneq_lits)
       {
        if (HeaderWeighting.Compare((*lit)->Header(),stat.max_noneq_header) == GREATER)
         stat.max_noneq_header = (*lit)->Header();
       }
      else stat.max_noneq_header = (*lit)->Header(); 
      stat.noneq_lits++;
     };
    if ((*lit)->Negative()) stat.negative++;
   };
 }; // void TMP_LIT_LIST::CollectStat(STAT& stat)


inline bool TMP_LIT_LIST::Eliminate(TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool* eliminated)
 {  
  for (TMP_LITERAL** lit = Last(); lit >= literal; lit--) 
   if (eliminated[(*lit)->Header().Functor()])
    {
     max_lits.Push(*lit);    
     ReplaceByTop(lit);
     while (Nonempty()) nonmax_lits.Push(Pop());
     return true;
    };        
  return false;
 }; //  bool TMP_LIT_LIST::Eliminate(TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool* eliminated)



inline void TMP_LIT_LIST::KBOrder_NS_NEQ(TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits)
 {
  DCONTROL10("BEGIN in TMP_LIT_LIST::KBOrder_NS_NEQ(..)");
  if (Length() == 1) { max_lits.Push(Pop()); return; };
     
  // find the greatest header
  TERM max_header = (*(Last()))->Header();
  TMP_LITERAL* neg_sel_cand;
  if (max_header.Polarity() == POL_NEG)
   {
    neg_sel_cand = Pop();
    goto acomplish_negative_selection;
   };
  
  for (TMP_LITERAL** lit = Last() - 1; lit >= literal; lit--)
   {
    if ((*lit)->Negative()) 
     {
      neg_sel_cand = *lit;
      do { nonmax_lits.Push(Pop()); } while (next_literal > lit + 1);
      Pop();
      goto acomplish_negative_selection;
     };
    
    switch (HeaderWeighting.Compare((*lit)->Header(),max_header))
     {
      case GREATER :
       max_header = (*lit)->Header();
       do { nonmax_lits.Push(Pop()); } while (next_literal > lit + 1);
       break; 
      case LESS :
       nonmax_lits.Push(*lit);
       ReplaceByTop(lit);
       break;
      case EQUAL : break;
      case INCOMPARABLE : 
       #ifdef _DEBUG10_
        DebugError("E0 in TMP_LIT_LIST::KBOrder_NS_NEQ(..)");
       #endif 
       break; 
     };
   };
      
  DCONTROL10("A20 in TMP_LIT_LIST::KBOrder_NS_NEQ(..)");
  
  TMP_LITERAL* cand; 
  check_length: 
   DCONTROL10("A21 in TMP_LIT_LIST::KBOrder_NS_NEQ(..)");
   if (Empty()) return;
   if (Length() == 1) { max_lits.Push(Pop()); return; };
   cand = Pop();
   for (TMP_LITERAL** lit = Last(); lit >= literal; lit--)
    {
     DCONTROL10("A25 in TMP_LIT_LIST::KBOrder_NS_NEQ(..)");
     switch ((*lit)->KBCompareArgs(*cand))
      {
       case GREATER : nonmax_lits.Push(cand); goto check_length;
       case LESS : nonmax_lits.Push(*lit); ReplaceByTop(lit); break;
       case INCOMPARABLE : break;
       case EQUAL : 
        #ifdef _DEBUG10_
         DebugError("E1 in TMP_LIT_LIST::KBOrder_NS_NEQ(..)");
        #endif
        break; 
      };
     DCONTROL10("A28 in TMP_LIT_LIST::KBOrder_NS_NEQ(..)"); 
    };
   max_lits.Push(cand);
   goto check_length; 
   
  acomplish_negative_selection:
   if (Empty()) { max_lits.Push(neg_sel_cand); return; };
   TMP_LITERAL* lit = Pop();
   if ((lit->Negative()) && (lit->Weight() > neg_sel_cand->Weight()))
    {
     nonmax_lits.Push(neg_sel_cand);
     neg_sel_cand = lit;
    }
   else nonmax_lits.Push(lit); 
   goto acomplish_negative_selection;
 }; // void TMP_LIT_LIST::KBOrder_NS_NEQ(TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits)



inline void TMP_LIT_LIST::KBOrder_NEQ(TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits)
 {

  DCONTROL10("BEGIN in TMP_LIT_LIST::KBOrder_NEQ(..)");
  if (Length() == 1) { max_lits.Push(Pop()); return; };
     
  // find the greatest header
  TERM max_header = (*(Last()))->Header();
  for (TMP_LITERAL** lit = Last() - 1; lit >= literal; lit--)
   {
    switch (HeaderWeighting.Compare((*lit)->Header(),max_header))
     {
      case GREATER :
       max_header = (*lit)->Header();
       do { nonmax_lits.Push(Pop()); } while (next_literal > lit + 1);
       break; 
      case LESS :
       nonmax_lits.Push(*lit);
       ReplaceByTop(lit);
       break;
      case EQUAL :
       break;
      case INCOMPARABLE : 
       #ifdef _DEBUG10_
        DebugError("E0 in TMP_LIT_LIST::KBOrder_NEQ(..)");
       #endif
       break; 
     };
   };
      
  DCONTROL10("A20 in TMP_LIT_LIST::KBOrder_NEQ(..)");
  
  TMP_LITERAL* cand; 
  check_length: 
   DCONTROL10("A21 in TMP_LIT_LIST::KBOrder_NEQ(..)");
   if (Empty()) 
    {

     //::Output("EMPTY!!!\n"); // debug    
     //::Output("MAX:  "); max_lits.Output(); ::Output("\n"); // debug 
     //::Output("NONMAX:  "); nonmax_lits.Output(); ::Output("\n\n\n"); // debug 
     

     return;
    };
   if (Length() == 1) 
    {
     max_lits.Push(Pop()); 

     //::Output("MAX:  "); max_lits.Output(); ::Output("\n"); // debug 
     //::Output("NONMAX:  "); nonmax_lits.Output(); ::Output("\n\n\n"); // debug 



     return; 
    };
   cand = Pop();
   for (TMP_LITERAL** lit = Last(); lit >= literal; lit--)
    {
     DCONTROL10("A25 in TMP_LIT_LIST::KBOrder_NEQ(..)");
     switch ((*lit)->KBCompareArgs(*cand))
      {
       case GREATER : nonmax_lits.Push(cand); goto check_length;
       case LESS : nonmax_lits.Push(*lit); ReplaceByTop(lit); break;
       case INCOMPARABLE : break;
       case EQUAL : 
        #ifdef _DEBUG10_
         DebugError("E1 in TMP_LIT_LIST::KBOrder_NEQ(..)");
        #endif
        break; 
      };
     DCONTROL10("A28 in TMP_LIT_LIST::KBOrder_NEQ(..)"); 
    };
   max_lits.Push(cand);
   goto check_length; 
 }; // void TMP_LIT_LIST::KBOrder_NEQ(TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits)

 /************  Stable version: ********/ 

inline void TMP_LIT_LIST::KBOrder_NS_EQ(TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits)
 {
  DCONTROL10("BEGIN in TMP_LIT_LIST::KBOrder_NS_EQ(..)");
  if (Length() == 1) { max_lits.Push(Pop()); return; };

  STAT stat;
  CollectStat(stat);
  if (stat.negative)
   {
    // find first negative literal
    TMP_LITERAL* cand = Pop();
    while (cand->Positive())
     {
      nonmax_lits.Push(cand);
      CHECK10(Nonempty(),"A10 in MP_LIT_LIST::KBOrder_NS_EQ(..)");
      cand = Pop(); 
     };
    long cand_weight = cand->Weight();
    TMP_LITERAL* next_lit;
    while (Nonempty())
     {
      next_lit = Pop();
      if ((next_lit->Negative()) && (next_lit->Weight() > cand_weight))  
       {
        nonmax_lits.Push(cand);
        cand = next_lit;
        cand_weight = cand->Weight();
       }
      else nonmax_lits.Push(next_lit);
     };
    max_lits.Push(cand);    
   }
  else // no negative literals
   {
    if (stat.equations) 
     {
      TMP_LITERAL* cand;
      check_length:
       if (Empty()) return;
       if (Length() == 1) { max_lits.Push(Pop()); return; };
       cand = Pop();
       for (TMP_LITERAL** lit = Last(); lit >= literal; lit--)
        {
         switch ((*lit)->KBComparePos(*cand))
          {
           case GREATER :
            nonmax_lits.Push(cand); goto check_length;
           case LESS :
            nonmax_lits.Push(*lit); ReplaceByTop(lit); break;
           case INCOMPARABLE : break;
           case EQUAL : 
            #ifdef _DEBUG10_
             DebugError("E1 in TMP_LIT_LIST::KBOrder_NS_EQ(..)");
            #endif
            break; 
          };
        };
       max_lits.Push(cand);
       goto check_length; 
     }
    else // no negative literals, no equations
     KBOrder_NEQ(max_lits,nonmax_lits); 
   };  
 }; // void TMP_LIT_LIST::KBOrder_NS_EQ(TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits)

 /***********************/

/*************** experimental version: *************

inline void TMP_LIT_LIST::KBOrder_NS_EQ(TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits)
 {
  DCONTROL10("BEGIN in TMP_LIT_LIST::KBOrder_NS_EQ(..)");
  if (Length() == 1) { max_lits.Push(Pop()); return; };

  STAT stat;
  CollectStat(stat);
  if (stat.negative)
   {
    // find first negative literal
    TMP_LITERAL* cand = Pop();
    while (cand->Positive())
     {
      nonmax_lits.Push(cand);
      CHECK10(Nonempty(),"A10 in MP_LIT_LIST::KBOrder_NS_EQ(..)");
      cand = Pop(); 
     };
    long cand_num_of_red = cand->NumOfRedexes();
    TMP_LITERAL* next_lit;
    while (Nonempty())
     {
      next_lit = Pop();
      if ((next_lit->Negative()) && (next_lit->NumOfRedexes() < cand_num_of_red))  
       {
        nonmax_lits.Push(cand);
        cand = next_lit;
        cand_num_of_red = cand->NumOfRedexes();
       }
      else nonmax_lits.Push(next_lit);
     };
    max_lits.Push(cand);    
   }
  else // no negative literals
   {
    if (stat.equations) 
     {
      TMP_LITERAL* cand;
      check_length:
       if (Empty()) return;
       if (Length() == 1) { max_lits.Push(Pop()); return; };
       cand = Pop();
       for (TMP_LITERAL** lit = Last(); lit >= literal; lit--)
        {
         switch ((*lit)->KBComparePos(*cand))
          {
           case GREATER :
            nonmax_lits.Push(cand); goto check_length;
           case LESS :
            nonmax_lits.Push(*lit); ReplaceByTop(lit); break;
           case INCOMPARABLE : break;
           case EQUAL : 
            #ifdef _DEBUG10_
             DebugError("E1 in TMP_LIT_LIST::KBOrder_NS_EQ(..)");
            #endif
            break; 
          };
        };
       max_lits.Push(cand);
       goto check_length; 
     }
    else // no negative literals, no equations
     KBOrder_NEQ(max_lits,nonmax_lits); 
   };  
 }; // void TMP_LIT_LIST::KBOrder_NS_EQ(TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits)

************************************/


inline void TMP_LIT_LIST::Order(TMP_LIT_LIST& result,const VAR_WEIGHTING& var_weighting)
 {
  DCONTROL10("BEGIN in TMP_LIT_LIST::Order(..)");
  
  // in the result greater literals will be closer to the top
  if (Length() == 1) 
   { 
    result.Push(Pop()); 
    DCONTROL10("END1 in TMP_LIT_LIST::Order(..)");
    return;
   };
  if (Length() == 2) 
   {
    TMP_LITERAL* fst = Pop();
    TMP_LITERAL* snd = Pop();
    
    DCONTROL10("A10 in TMP_LIT_LIST::Order(..)");
    if (fst->Compare(*snd,var_weighting) == LESS) 
     {
      DCONTROL10("A11 in TMP_LIT_LIST::Order(..)");
      result.Push(fst); 
      result.Push(snd); 
      DCONTROL10("END2 in TMP_LIT_LIST::Order(..)");
      return;
     };
    DCONTROL10("A12 in TMP_LIT_LIST::Order(..)"); 
    result.Push(snd); 
    result.Push(fst);
    DCONTROL10("END3 in TMP_LIT_LIST::Order(..)");
    return; 
   };
  TMP_LITERAL* center = Pop();  
  TMP_LIT_LIST smaller;
  TMP_LIT_LIST greater; 
  do
   {
    TMP_LITERAL* lit = Pop();
    if (lit->Compare(*center,var_weighting) == LESS) { smaller.Push(lit); } else { greater.Push(lit); };
   }
  while (Nonempty()); 
  if (smaller.Length()) smaller.Order(result,var_weighting);
  result.Push(center);
  if (greater.Length()) greater.Order(result,var_weighting);
  DCONTROL10("END4 in TMP_LIT_LIST::Order(..)");
 }; // void TMP_LIT_LIST::Order(TMP_LIT_LIST& result,const VAR_WEIGHTING& var_weighting)

inline void TMP_LIT_LIST::CreateVarWeighting(VAR_WEIGHTING& var_weighting)
 {
  var_weighting.Reset();
  for (TMP_LITERAL** lit = literal; lit < next_literal; lit++)
   var_weighting.AnotherTermVariables((*lit)->Variables()); 
 }; // void TMP_LIT_LIST::CreateVarWeighting(VAR_WEIGHTING& var_weighting)

inline void TMP_LIT_LIST::RenameVariables(INT_VAR_RENAMING& var_renaming)
 {
  for (TMP_LITERAL** lit = Last(); lit >= literal; lit--)
   // order is significant!
   (*lit)->RenameVariables(var_renaming);
 }; // void TMP_LIT_LIST::RenameVariables(INT_VAR_RENAMING& var_renaming)

inline void TMP_LIT_LIST::CollectGround(TMP_LIT_LIST& component)
 {
  component.Reset();
  for (TMP_LITERAL** lit = Last(); lit >= literal; lit--) // order is significant!
   if ((*lit)->Ground()) 
    {
     component.Push(*lit);
     ReplaceByTop(lit); 
    }; 
 }; // void TMP_LIT_LIST::CollectGround(TMP_LIT_LIST& component)

inline void TMP_LIT_LIST::CollectNextComponent(TMP_LIT_LIST& component)
 {
  component.Reset();
  static VAR_MULTISET variables;
  if (Length() == 1) { component.Push(Pop()); return; };  
  TMP_LITERAL* fst = Pop();
  component.Push(fst);
  variables = fst->Variables();
  
  int prev_num_of_different;
  new_search:
   prev_num_of_different = variables.NumOfDifferent();
   for (TMP_LITERAL** lit = Last(); lit >= literal; lit--) // order is significant
    {
     if ((*lit)->Variables().IntersectionNonempty(variables)) 
      {
       component.Push(*lit);
       variables.Add((*lit)->Variables());
       ReplaceByTop(lit);
      };
    };
   if ((Nonempty()) && (variables.NumOfDifferent() != prev_num_of_different))
    goto new_search;
 }; // void TMP_LIT_LIST::CollectNextComponent(TMP_LIT_LIST& component)

//***********************************************************************

