#define __VARIABLES_H
#ifndef __DEBUG_H
 #include "debug.h"
#endif
#ifndef __VAMPIRE_H
 #include "vampire.h"
#endif
#ifndef __TERM_H
 #include "term.h"
#endif
#ifndef __OUTPUT_H
 #include "output.h"
#endif
 
class SUBSTITUTION;
 
extern SUBSTITUTION GlobalSubst;
 
class VAR
 {
  private: // static data
   static VAR var[MAX_NUMBER_OF_INDICES*MAX_NUMBER_OF_VARS];
   static unsigned offset[MAX_NUMBER_OF_INDICES];
   static const void* end_of_vars;
  private: // structure
   void* binding;
  public:
   VAR() : binding(0) {};
  public:
   static inline void Reset() 
    {       
     for (int i = 0; i<MAX_NUMBER_OF_INDICES; i++)
      {
       offset[i] = i*MAX_NUMBER_OF_VARS;
      };
    };
   inline void Bind(void* bind) { binding = bind; };
   inline void Unbind() { binding = 0; };
   inline void* Binding() const { return binding; };
   static inline VAR* Base(int v) { return var+v; };
   static inline unsigned Offset(int index) { return offset[index]; }; 
   static inline VAR* Var(int v,int index) 
    { return var + v + offset[index]; };
   static inline VAR* First() { return var; };
   static inline VAR* First(int index) { return var + offset[index]; };
   static inline VAR* Last(int index) { return (var + offset[index] + MAX_NUMBER_OF_VARS - 1); };
   inline VAR* Next() { return this+1; };
   inline VAR* Previous() { return this-1; };
   static inline int Index(unsigned var_offset) 
    {
     return (var_offset/MAX_NUMBER_OF_VARS);
    };
   inline int Index() const { return (this - var)/MAX_NUMBER_OF_VARS; }; 
   inline void NumAndIndex(int& num,int& index) const
    {
     unsigned pos = this - var;
     index = pos/MAX_NUMBER_OF_VARS;
     num = pos%MAX_NUMBER_OF_VARS;
    };
   
   static inline bool IsVar(void* ptr) 
    { return ((ptr >= ((void*)var)) && (ptr<((void*)end_of_vars))); };
   
   
   /************** Can be more efficient variant:
   static inline bool IsVar(void* ptr) 
    { 
     return ((ptr<((void*)end_of_vars)) && (ptr >= ((void*)var))); 
    }; 
   *****************************************************/ 
    
   inline void* Unref() 
    {
     void* res = Binding();
     if (res)
      {
       check_binding:
        if (IsVar(res))
         {
          void* tmp = ((VAR*)res)->Binding();
          if (tmp)
           {
            res = tmp;
            goto check_binding;
           }
          else
           {
            return res; // free variable
           };
         }
        else // complex term 
         {
          return res;
         };
      }
     else
      {
       return this;
      };
    }; 
   inline void* Unref(bool& free_var) const
    {
     void* res = Binding();
     if (res)
      {
       check_binding:
        if (IsVar(res))
         {
          void* tmp = ((VAR*)res)->Binding();
          if (tmp)
           {
            res = tmp;
            goto check_binding;
           }
          else
           {
            free_var = true;
            return res; // free variable
           };
         }
        else // complex term 
         {
          free_var = false;
          return res;
         };
      }
     else
      {
       free_var = true;
       return (void*)this;
      };
    };
  public: // methods for debugging
   inline void Output() const
    {
     int num;
     int index;
     NumAndIndex(num,index);
     OutputVar(num);
     ::Output("{"); 
     ::Output(index);
     ::Output("}");
    };
   inline void OutputBinding(void (*OutputTerm)(void* t))
    {
     Output();
     ::Output(" := ");
     if (Binding())
      {
       if (VAR::IsVar(Binding()))
        {
         ((VAR*)Binding())->Output();
        }
       else { OutputTerm(Binding()); };
      }
     else { ::Output("NULL"); };
    };
 }; // class VAR

class SUBSTITUTION
 {
  private: 
   VAR* bound[MAX_NUMBER_OF_VARS*MAX_NUMBER_OF_INDICES];
   VAR** next;
   VAR** section[MAX_NUMBER_OF_VARS*MAX_NUMBER_OF_INDICES];
   VAR*** next_section;
  public:
   SUBSTITUTION() 
    {
     next = bound;
     next_section = section;
    };
   inline void UnbindAll() 
    {
     while (next != bound)
      {
       next--;
       (*next)->Unbind();
      };
     next_section = section;
    };
   inline void Bind(VAR* v,void* val) 
    {
     CHECK2(v != ((VAR*)val),"SAME VAR in SUBSTITUTION::Bind()"); 
     v->Bind(val);
     *next = v;
     next++;
    };
   
   inline void BindAlone(VAR* v,void* val)
    {
     Bind(v,val);
    }; 
   
   inline void BindInSection(VAR* v,void* val)
    {
     Bind(v,val);
    };
    
   inline VAR* LastBound() const { return *(next-1); };
   inline void UnbindLast()
    {
     
     next--;
     (*next)->Unbind();
     
     CHECK2(SectionsOK(),"A1 in SUBSTITUTION::UnbindLast()");
    };
   inline void MarkSection() 
    {
     *next_section = next; 
     next_section++;
     CHECK2(SectionsOK(),"A1 in SUBSTITUTION::MarkSection()");
    };
   inline VAR*** LastMark() const
    {
     CHECK2(next_section > section,"A1 in SUBSTITUTION::LastMark()");
     return next_section-1; 
    };
   
   inline void UnbindDownto(VAR*** mark)
    {
     VAR** m = *mark;
     next_section = mark;
     while (next != m)
      {
       next--;
       (*next)->Unbind();
      };
     CHECK2(SectionsOK(),"A1 in SUBSTITUTION::UnbindDownto()");
    };

   inline void UnbindLastSection()
    {
     CHECK2(next_section > section,"A1 in SUBSTITUTION::UnbindLastSection()");
     next_section--;
     VAR** section_mark = *next_section;
     
     CHECK2(section_mark >= bound,"A2 in SUBSTITUTION::UnbindLastSection()");
     CHECK2(section_mark <= next,"A3 in SUBSTITUTION::UnbindLastSection()");
     
     while (next != section_mark)
      {
       next--;
       (*next)->Unbind();
      };
     CHECK2(SectionsOK(),"B1 in SUBSTITUTION::UnbindLastSection()"); 
    };
   inline VAR*** MarkState()
    {
     MarkSection();
     return LastMark();
    };
 
  public:
   inline bool VariablesAffected(VAR* first,VAR* last)
    {
     for (VAR** bv = bound; bv < next; bv++)
      if (((*bv) >= first) && ((*bv) <= last)) return true;    
     return false;
    };
   inline bool VariablesAffected(int index)
    { 
     return VariablesAffected(VAR::First(index),VAR::Last(index));
    };

  public: // methods for debugging
   inline bool SectionsOK() const
    {
     VAR*** sect = next_section - 1;
     if (sect >= ((VAR***)section))
      {
       if ((*sect) > next)
        {
         
         //GlobalSubst.Output(OutputPrefix); // debug
         
         return false;
        };
      };
     return true;     
    };
   inline unsigned NumOfBindings() const 
    { return (next - (VAR**)bound); };
   inline void Output(void (*OutputTerm)(void* t))
    {
     for(VAR** p = bound; p<next; p++)
      {
       (*p)->Output();
       ::Output(" := ");
       void* bind = (*p)->Binding();
       if (VAR::IsVar(bind))
        {
         ((VAR*)bind)->Output();
        }
       else
        {
         OutputTerm(bind);
        };
       ::Output("\n");
      };
    };
 }; // class SUBSTITUTION


class LARGE_VAR_TABLE
 {
  private:
   void* table[MAX_NUMBER_OF_INDICES*MAX_NUMBER_OF_VARS];
   unsigned offset; 
  public:
   LARGE_VAR_TABLE() { offset = table - (void**)(VAR::First()); };
   inline void Initialize(void* val)
    {
     for (int i = 0; i< MAX_NUMBER_OF_INDICES*MAX_NUMBER_OF_VARS; i++)
      {
       table[i] = val;
      };
    };
   inline void** Entry(const VAR* const v) const { return ((void**)v) + offset; };
 }; // class LARGE_VAR_TABLE
 
class SMALL_VAR_TABLE
 {
  private:
   void* table[MAX_NUMBER_OF_VARS];
   long offset; 
  public:
   SMALL_VAR_TABLE() { offset = table - (void**)(VAR::First()); };
   inline void SetIndex(int ind)
    {
     offset = (table - (void**)(VAR::First())) - VAR::Offset(ind);
    };
   inline void Initialize(void* val)
    {
     for (int i = 0; i< MAX_NUMBER_OF_VARS; i++)
      {
       table[i] = val;
      };
    };
   inline void** Entry(const VAR* const v) const { return ((void**)v) + offset; };
 }; // class SMALL_VAR_TABLE 
 
class LARGE_VAR_COUNTER
 {
  private:
   LARGE_VAR_TABLE counter;
   VAR* registered[MAX_NUMBER_OF_VARS*MAX_NUMBER_OF_INDICES]; // stack
   VAR** next;
   VAR** iter;
  public:
   LARGE_VAR_COUNTER() : counter() 
    { 
     counter.Initialize((void*)(0L));
     next = registered; 
    };
   inline void Reset()
    {
     while (next != registered)
      {
       next--; 
       *(counter.Entry(*next)) = (void*)(0L);
      };
    };
  private:  
   inline void Unregister(const VAR* v)
    {
     CHECK2(!Score(v),"A0 in LARGE_VAR_COUNTER::Unregister(const VAR* v)");  
     for (VAR** ptr = registered; ptr < next; ptr++)  
      {
       if (v == (*ptr)) { next--; *ptr = *next; return; };
      };
    };

  public:
   inline long Register(const VAR* const v) 
    // returns previous value of the counter for the variable
    {
     long res = (long)(*(counter.Entry(v)));
     *(counter.Entry(v)) = (void*)(res + 1);
     if (!res) { *next = (VAR*)v; next++; }
     else 
      if (!Score(v)) 
       {
        // variable with zero score must be ungeristered
        Unregister(v);
       };
     return res;      
    };

   inline long Register(long coefficient,const VAR* const v) 
    // returns previous value of the counter for the variable
    {
     long res = (long)(*(counter.Entry(v)));
     *(counter.Entry(v)) = (void*)(res + coefficient);
     if (!res) { *next = (VAR*)v; next++; }
     else 
      if (!Score(v)) 
       {
        // variable with zero score must be ungeristered
        Unregister(v);
       };
     return res;      
    };

   inline void MultiplyBy(long coefficient)
    {
     if (coefficient)
      {
       for (VAR** ptr = registered; ptr < next; ptr++)  
	{
         *(counter.Entry(*ptr)) = (void*)(coefficient*Score(*ptr)); 
        };
      }
     else Reset();
    };
 
   inline long Score(const VAR* const v) const
    {
     return (long)(*(counter.Entry(v)));
    };

   inline int NumOfRegistered() const { return (next - registered); };

   inline void ResetIteration()
    {
     iter = registered;
    };
   inline VAR* NextRegistered()
    {
     if (iter == next) { return (VAR*)0; };
     VAR* res = *iter;
     iter++;
     return res;
    };
 }; // class LARGE_VAR_COUNTER

class SMALL_VAR_COUNTER
 {
  private:
   SMALL_VAR_TABLE counter;
   VAR* registered[MAX_NUMBER_OF_VARS]; // stack
   VAR** next;
   VAR** iter;
  public:
   SMALL_VAR_COUNTER() : counter() 
    { 
     counter.Initialize((void*)((unsigned)0));
     next = registered; 
    };
   inline void SetIndex(int ind) { counter.SetIndex(ind); }; 
   inline void Reset()
    {
     while (next != registered)
      {
       next--; 
       *(counter.Entry(*next)) = (void*)((unsigned)0);
      };
    };
   inline unsigned Register(const VAR* const v) 
    // returns previous value of the counter for the variable
    {
     unsigned res = (unsigned)(*(counter.Entry(v)));
     *(counter.Entry(v)) = (void*)(res + 1);
     if (!res) { *next = (VAR*)v; next++; };
     return res;      
    };
   
   inline unsigned Score(const VAR* const v) const
    {
     return (unsigned)(*(counter.Entry(v)));
    };
    
   inline void ResetIteration()
    {
     iter = registered;
    };
   inline VAR* NextRegistered()
    {
     if (iter == next) { return (VAR*)0; };
     VAR* res = *iter;
     iter++;
     return res;
    };
 }; // class SMALL_VAR_COUNTER

class LARGE_VAR_RENAMING
 {
  private:
   LARGE_VAR_TABLE renaming;
   VAR* next_var;
   VAR* renamed[MAX_NUMBER_OF_VARS*MAX_NUMBER_OF_INDICES]; // stack
   VAR** next;
  public: 
   LARGE_VAR_RENAMING() : renaming()
    {
     renaming.Initialize((void*)0);
     next_var = VAR::First();
     next = renamed;
    };
   inline void Reset()
    {
     next_var = VAR::First();
     while (next != renamed)
      {
       next--; 
       *(renaming.Entry(*next)) = (void*)0;
      };
    };
   inline VAR* Rename(VAR* var)
    {
     VAR* res = (VAR*)(*(renaming.Entry(var)));
     if (!res) 
      {
       res = next_var;
       *(renaming.Entry(var)) = (void*)res;
       next_var++;
       *next = var;
       next++;
      };
     return res;
    };
     
  public: // for debugging
   inline void Output()
    {
     ::Output("VARIABLE RENAMING:\n");
     for (VAR** r = renamed; r < next; r++)
      {
       (*r)->Output();
       ::Output(" -> ");
       ((VAR*)(*(renaming.Entry(*r))))->Output();
       ::Output("\n");
      };
     ::Output("END OF VARIABLE RENAMING.\n");
    };
   
 }; // class LARGE_VAR_RENAMING

class VAR_TO_INT_RENAMING
 {
  private:
   LARGE_VAR_TABLE renaming;
   int next_var;
   VAR* renamed[MAX_NUMBER_OF_VARS*MAX_NUMBER_OF_INDICES]; // stack
   VAR** next;
  public: 
   VAR_TO_INT_RENAMING() : renaming()
    {
     renaming.Initialize((void*)((int)0));
     next_var = 1;
     next = renamed;
    };
   inline void Reset()
    {
     next_var = 1;
     while (next != renamed)
      {
       next--; 
       *(renaming.Entry(*next)) = (void*)((int)0);
      };
    };
   inline int Rename(VAR* var)
    {
     int res = (int)(*(renaming.Entry(var)));
     if (!res) 
      {
       res = next_var;
       *(renaming.Entry(var)) = (void*)res;
       next_var++;
       *next = var;
       next++;
      };
     return res;
    };
     
  public: // for debugging
   inline void Output()
    {
     ::Output("VAR_TO_INT_RENAMING:\n");
     for (VAR** r = renamed; r < next; r++)
      {
       (*r)->Output();
       ::Output(" -> ");
       ::OutputVar((int)(*(renaming.Entry(*r))));
       ::Output("\n");
      };
     ::Output("END OF VAR_TO_INT_RENAMING.\n");
    };
   
 }; // class VAR_TO_INT_RENAMING



class INT_VAR_RENAMING
 {
  private:
   int renaming[MAX_NUMBER_OF_VARS];
   int next_var;
   int renamed[MAX_NUMBER_OF_VARS]; // stack
   int* next;
  public: 
   INT_VAR_RENAMING() 
    {
     for (int v = 0; v < MAX_NUMBER_OF_VARS; v++)
      {
       renaming[v] = -1;
      };
     next_var = 0;
     next = renamed;
    };
   inline void Reset()
    {
     next_var = 0;
     while (next != renamed)
      {
       next--; 
       renaming[*next] = -1;
      };
    };
   inline int Rename(const int var)
    {
     int res = renaming[var];
     if (res < 0) 
      {
       res = next_var;
       renaming[var] = res;
       next_var++;
       *next = var;
       next++;
      };
     return res;
    };
  public: // for debugging
   inline void Output()
    {
     ::Output("VARIABLE RENAMING:\n");
     for (int* r = renamed; r < next; r++)
      {
       OutputVar(*r);
       ::Output(" -> ");
       OutputVar(renaming[*r]);
       ::Output("\n");
      };
     ::Output("END OF VARIABLE RENAMING.\n");
    };
   
 }; // class INT_VAR_RENAMING

class VAR_MULTISET
 {
  private:
   STACK_WITH_ITERATOR<int,MAX_NUMBER_OF_VARS> registered;
   int occurences[MAX_NUMBER_OF_VARS];
   int size;
  public:
   VAR_MULTISET() 
    : registered("MAX_NUMBER_OF_VARS in VAR_MULTISET::registered"),
      size(0)
    {
     for (int v = 0; v < MAX_NUMBER_OF_VARS; v++) { occurences[v] = 0; };
    };
   
   inline VAR_MULTISET& operator=(VAR_MULTISET& vms)
    {
     Reset();
     registered = vms.registered;  
     vms.ResetIterator();
     const int* var = vms.NextVar();
     while (var)
      {
       occurences[*var] = vms.Occurences(*var);
       var = vms.NextVar();
      };
     size = vms.Size();
     return *this;
    };  
   
   inline void Reset()
    {
     while (!(registered.Empty())) { occurences[registered.Pop()] = 0; };
     size = 0;
    };
   inline void Register(const int v)
    {
     if (!(occurences[v])) { registered.Push(v); };
     occurences[v]++;
     size++;
    };
   inline void Register(const int v,const int new_occs)
    {
     if (!(occurences[v])) { registered.Push(v); };
     occurences[v] += new_occs;
     size += new_occs;
    };
   
   inline void Unregister(int v) // v must be in the multiset!
    {
     occurences[v] = 0;
     int top_var = registered.Pop();
     if (v != top_var) 
      {
       registered.ResetIterator();
       int* next_var = registered.Next();
       while ((*next_var) != v) next_var = registered.Next();
       *next_var = top_var; 
      };      
    };
    
   inline int Occurences(const int v) const { return occurences[v]; };
   inline int Size() const { return size; };
   inline int NumOfDifferent() const { return registered.NumberOfElements(); };
   inline bool Belongs(const int v) const { return (bool)(Occurences(v)); };
   
   inline void ResetIterator() { registered.ResetIterator(); };
   inline const int* NextVar() { return registered.Next(); };
   
   inline void Add(VAR_MULTISET& vms)
    {
     vms.ResetIterator();
     const int* var = vms.NextVar();
     while (var)
      {
       Register(*var,vms.Occurences(*var));
       var = vms.NextVar();
      };
    };
   
   inline void CollectDifference(VAR_MULTISET& vms1,VAR_MULTISET& vms2)
    {
     vms1.ResetIterator();
     int diff;
     for (const int* var = vms1.NextVar(); var; var = vms1.NextVar())
      {
       diff = vms1.Occurences(*var) - vms2.Occurences(*var);
       if (diff > 0) Register(*var,diff);
      }; 
    }; 
   
   inline bool IntersectionNonempty(VAR_MULTISET& vms) 
    {
     if ((!(Size())) || (!(vms.Size()))) { return false; };
     if (NumOfDifferent() <= vms.NumOfDifferent())
      {
       registered.ResetIterator();  
       int* var = registered.Next();
       do
        {
         if (vms.Belongs(*var)) { return true; };
         var = registered.Next();
        }
       while (var);
      }
     else
      {
       vms.registered.ResetIterator();  
       int* var = vms.registered.Next();
       do
        {
         if (Belongs(*var)) { return true; };
         var = vms.registered.Next();
        }
       while (var);
      };
     return false;
    };
   
   inline bool Equal(const VAR_MULTISET& vms)
    {
     int* var;
     if ((Size() == vms.Size()) && (NumOfDifferent() == vms.NumOfDifferent()))
      {
       registered.ResetIterator();
       next_var:
        var = registered.Next();
        if (var)
         {
          if (Occurences(*var) != vms.Occurences(*var)) { return false; };
          goto next_var; 
         }
        else { return true; };
      }
     else { return false; };
    }; // bool Equal(const VAR_MULTISET& vms)
   
   inline bool GreaterOrEqual(VAR_MULTISET& vms)
    {
     int* var;
     if (Size() == vms.Size())
      {
       if (NumOfDifferent() != vms.NumOfDifferent()) { return false; };
       registered.ResetIterator();
       eq_next_var:
        var = registered.Next();
        if (var)
         {
          if (Occurences(*var) != vms.Occurences(*var)) { return false; };
          goto eq_next_var; 
         }
        else { return true; }; // equal multisets
      }
     else
      {
       if (Size() > vms.Size())
        {
         if (NumOfDifferent() < vms.NumOfDifferent()) { return false; };
         vms.registered.ResetIterator(); 
         greater_next_var:
          var = vms.registered.Next();
          if (var)
           {
            if (Occurences(*var) < vms.Occurences(*var)) { return false; };
            goto greater_next_var; 
           }
          else { return true; }; // greater
        }
       else // Size() < vms.Size()
        {
         return false; 
        };
      };
     
    }; // bool GreaterOrEqual(VAR_MULTISET& vms)
   
   inline COMPARISON Compare(VAR_MULTISET& vms) 
    {
     int* var;
     if (Size() == vms.Size())
      {
       if (NumOfDifferent() != vms.NumOfDifferent()) { return INCOMPARABLE; };
       registered.ResetIterator();
       eq_next_var:
        var = registered.Next();
        if (var)
         {
          if (Occurences(*var) != vms.Occurences(*var)) { return INCOMPARABLE; };
          goto eq_next_var; 
         }
        else { return EQUAL; };
      }
     else
      {
       if (Size() < vms.Size())
        {
         if (NumOfDifferent() > vms.NumOfDifferent()) { return INCOMPARABLE; };
         registered.ResetIterator();
         less_next_var:
          var = registered.Next();
          if (var)
           {
            if (Occurences(*var) > vms.Occurences(*var)) { return INCOMPARABLE; };
            goto less_next_var; 
           }
          else { return LESS; };
        }
       else
        {
         if (NumOfDifferent() < vms.NumOfDifferent()) { return INCOMPARABLE; };
         vms.registered.ResetIterator(); 
         greater_next_var:
          var = vms.registered.Next();
          if (var)
           {
            if (Occurences(*var) < vms.Occurences(*var)) { return INCOMPARABLE; };
            goto greater_next_var; 
           }
          else { return GREATER; };
        };
      };
    };
  public: // for debugging
   inline void Output()
    {
     if (Size())
      {
       ResetIterator();
       const int* var = NextVar();
       
       ///cout << "size = " << Size() << "\n"; // debug
       //cout << "different = " << NumOfDifferent() << "\n"; // debug
       //cout << (long)var << "\n"; // debug
       
       ::Output(Occurences(*var));
       ::Output("*");
       ::OutputVar(*var);
       var = NextVar();
       while (var)
        {
         ::Output(" + ");
         ::Output(Occurences(*var));
         ::Output("*");
         ::OutputVar(*var);
         var = NextVar();
        };
      };
    };
 }; // class VAR_MULTISET
 
class VAR_WEIGHT
 {
  private: // structure
   int num_of_occurences;
   int different_terms; 
  public:
   VAR_WEIGHT() : num_of_occurences(0), different_terms(0) {};
   inline void Reset() { num_of_occurences = 0; different_terms = 0; };
   inline bool Nonnull() { return (num_of_occurences || different_terms); };
   inline void MoreOccurences(const int n) { num_of_occurences += n; };
   inline void AnotherTerm() { different_terms++; };
   inline COMPARISON Compare(const VAR_WEIGHT& w) const
    {
     if (num_of_occurences == w.num_of_occurences)
      {
       if (different_terms == w.different_terms) { return EQUAL; }
       else
        { 
         if (different_terms > w.different_terms) { return GREATER; }
         else { return LESS; };
        };
      }
     else
      {
       if (num_of_occurences > w.num_of_occurences) { return GREATER; }
       else { return LESS; };
      }; 
    };
 }; // class VAR_WEIGHT
 
class VAR_WEIGHTING
 {
  private: 
   VAR_WEIGHT weight[MAX_NUMBER_OF_VARS];
   STACK<int,MAX_NUMBER_OF_VARS> registered;
  public:
   VAR_WEIGHTING() : registered("MAX_NUMBER_OF_VARS in VAR_WEIGHTING::registered")
    {
     for (int v = 0; v < MAX_NUMBER_OF_VARS; v++) { weight[v].Reset(); };
    };
   inline void Reset()
    {
     while (!(registered.Empty())) { weight[registered.Pop()].Reset(); };
    };
   inline void MoreOccurences(const int var,const int occ)
    {
     if (!(weight[var].Nonnull())) { registered.Push(var); };
     weight[var].MoreOccurences(occ);
    };
   inline void AnotherTerm(const int var)
    {
     if (!(weight[var].Nonnull())) { registered.Push(var); };
     weight[var].AnotherTerm();
    };
   inline COMPARISON Compare(const int var1,const int var2) const
    {
     return weight[var1].Compare(weight[var2]);
    };
   inline void AnotherTermVariables(VAR_MULTISET& vars)
    {
     vars.ResetIterator();
     const int* next_var = vars.NextVar();
     while (next_var)
      {
       AnotherTerm(*next_var);
       MoreOccurences(*next_var,vars.Occurences(*next_var));
       next_var = vars.NextVar();
      };
    };
 }; // class VAR_WEIGHTING  
 
 

//**************************************************************************  






