#define __WEIGHT_FORMULA_H

#ifndef __DEBUG_H
 #include "debug.h"
#endif
#ifndef __VAMPIRE_H
 #include "vampire.h"
#endif
#ifndef __PREFIX_H
 #include "prefix.h"
#endif
#ifndef __VARIABLES_H
 #include "variables.h"
#endif
#ifndef __OUTPUT_H
 #include "output.h"
#endif

//************************************************************************

class LARGE_WEIGHT_FORMULA
 {
  // this class has heavy constructor, objects must be static!
  private:
   LARGE_VAR_COUNTER var_counter;
   long const_part;
  public:
   LARGE_WEIGHT_FORMULA() : var_counter(), const_part(0L) {};
   inline void Reset() { var_counter.Reset(); const_part = 0L; };      
   inline void RegisterVar(long coefficient,const VAR* v) { var_counter.Register(coefficient,v); };
   inline void RegisterFunc(long coefficient) { const_part += coefficient; };

   inline void MultiplyBy(long coefficient) 
    {
     const_part = coefficient*const_part;
     var_counter.MultiplyBy(coefficient); 
    };  
   
   inline void InverseSign() { MultiplyBy(-1L); };
 
   inline void CompileSimple(long coefficient,PREFIX_SYM* t) // without substitution
    {
     if (t->Head().IsVariable()) 
      {
       RegisterVar(coefficient,t->Head().Var());
       return;             
      };

     PREFIX::SIMPLE_ITER iter;
     iter.Reset(t);
     do
      {
       PREF_SYM sym = iter.CurrPos()->Head();
       if (sym.IsVariable())
        {
         RegisterVar(coefficient,sym.Var());
        }
       else RegisterFunc(coefficient); 
       iter.SkipSym();
      }
     while (!iter.Empty());
    }; // void CompileSimple(long coefficient,PREFIX_SYM* t)

 
   inline void Compile(long coefficient,PREFIX_SYM* t)
    {
     if (t->Head().IsVariable()) 
      {
       bool free_var;
       void* bind = t->Head().Var()->Unref(free_var);
       if (free_var)
        {
         RegisterVar(coefficient,(const VAR*)bind);
         return;
        }
       else { t = (PREFIX_SYM*)bind; }; 
      };

     PREFIX::ITER iter;
     iter.Reset(t);
     do
      {
       if (iter.CurrVar())
        {
         RegisterVar(coefficient,iter.CurrVar());
        }
       else RegisterFunc(coefficient);  
       iter.SkipSym();
      }
     while (!iter.Empty());
    }; // void Compile(long coefficient,PREFIX_SYM* t)


   inline void CompileSubst(long coefficient,const VAR* var)
    {
     bool free_var;
     void* bind = var->Unref(free_var);
     if (free_var)
      {
       RegisterVar(coefficient,(const VAR*)bind);
      }
     else Compile(coefficient,(PREFIX_SYM*)bind);
    };   

   inline void Substitute(LARGE_WEIGHT_FORMULA& result)
    {
     result.Reset();
     result.RegisterFunc(const_part);
     var_counter.ResetIteration();
     for (VAR* var = var_counter.NextRegistered(); var; var = var_counter.NextRegistered())
      {
       result.CompileSubst(var_counter.Score(var),var); 
      };
    }; // void Substitute(LARGE_WEIGHT_FORMULA& result)
   
   inline bool Zero() const { return (!const_part) && (!var_counter.NumOfRegistered()); };
   inline bool Negative() 
    {
     if (!var_counter.NumOfRegistered()) return (const_part < 0);
     if (const_part > 0)
      {
       //return false; // debug   

       long rest_of_const_part = const_part;
       long score;
       
       var_counter.ResetIteration(); 
       for (VAR* var = var_counter.NextRegistered(); var; var = var_counter.NextRegistered())
        {
         score = var_counter.Score(var);
         if (score > 0) return false; 
         rest_of_const_part+=score;
        };
       return (rest_of_const_part < 0);
      };
     // here const_part <= 0
     var_counter.ResetIteration();
     for (VAR* var = var_counter.NextRegistered(); var; var = var_counter.NextRegistered())
      if (var_counter.Score(var) > 0) return false;
     return true;
    };

   inline bool Positive() 
    {
     if (!var_counter.NumOfRegistered()) return (const_part > 0);
     if (const_part < 0)
      {
       //return false; // debug   

       long rest_of_const_part = const_part;
       long score;
       
       var_counter.ResetIteration(); 
       for (VAR* var = var_counter.NextRegistered(); var; var = var_counter.NextRegistered())
        {
         score = var_counter.Score(var);
         if (score < 0) return false; 
         rest_of_const_part+=score;
        };
       return (rest_of_const_part > 0);
      };
     // here const_part >= 0
     var_counter.ResetIteration();
     for (VAR* var = var_counter.NextRegistered(); var; var = var_counter.NextRegistered())
      if (var_counter.Score(var) < 0) return false;
     return true;
    }; // bool Positive() 

  public: // for debugging  
   inline void Output() 
    {
     var_counter.ResetIteration();
     ::Output(const_part);
     for (const VAR* var = var_counter.NextRegistered();var;var = var_counter.NextRegistered())
      {
       long score = var_counter.Score(var);
       if (score < 0)
        {
         ::Output(" - ");
         ::Output(score*(-1));
        }
       else
        { 
         ::Output(" + ");
         ::Output(score);
        };
       var->Output();
      };
    };  
  inline void OutputWithSubst()
   {
    Output(); ::Output("  WHERE\n");
    var_counter.ResetIteration();
    //for (const VAR* var = var_counter.NextRegistered();var;var = var_counter.NextRegistered())
    // { 
    //  ::Output("     "); ((VAR*)var)->OutputBinding(OutputPrefix); ::Output("\n");    
    // };
    GlobalSubst.Output(OutputPrefix);
   };
 }; // class LARGE_WEIGHT_FORMULA


class ORD_FORMULA
 {
  // this class has heavy constructor, objects must be static!
  private:
   LARGE_WEIGHT_FORMULA difference;
   LARGE_WEIGHT_FORMULA subst_diff;
   PREFIX_SYM* term1;
   PREFIX_SYM* term2;
  public:
   ORD_FORMULA() : difference(), subst_diff() {};
   inline void Reset() { difference.Reset(); };
   inline void Compile(PREFIX_SYM* t1,PREFIX_SYM* t2)
    {
     // t1 and t2 must be complex here, this is used in lexicographic comparison 
     term1 = t1;
     term2 = t2;
     difference.CompileSimple(1L,t1);
     difference.CompileSimple(-1L,t2);
    };
 
   inline void Inverse() 
    {
     PREFIX_SYM* tmp = term1;
     term1 = term2;
     term2 = tmp;
     difference.InverseSign(); 
    };
  private:
   inline bool LexLessOrEq() 
    {
     PREFIX::ITER iter1;
     PREFIX::ITER iter2;
     iter1.Reset(term1);
     iter2.Reset(term2);
     TERM func1;
     TERM func2;
     do
      {
       if (iter1.CurrVar())
        {
         if (iter2.CurrVar())
          {
           if (iter1.CurrVar() != iter2.CurrVar()) return false;
          }
         else return false;
        } 
       else
        if (iter2.CurrVar())
         {
          return false;
         }
        else // both are functors
         {
          func1 = iter1.CurrSubterm()->Head().Func();     
          func2 = iter2.CurrSubterm()->Head().Func(); 
          if (func1 != func2) return (func1.CompareFunctor(func2) == LESS); 
         }; 
       iter1.SkipSym();
       iter2.SkipSym();
      }      
     while (!iter1.Empty());
     return true; // equal
    }; // bool LexLessOrEq() 

   inline bool LexGreater() 
    {
     PREFIX::ITER iter1;
     PREFIX::ITER iter2;
     iter1.Reset(term1);
     iter2.Reset(term2);
     TERM func1;
     TERM func2;
     do
      {
       if (iter1.CurrVar())
        {
         if (iter2.CurrVar())
          {
           if (iter1.CurrVar() != iter2.CurrVar()) return false; // incomparable
          }
         else return false; // incomparable
        } 
       else
        if (iter2.CurrVar())
         {
          return false; // incomparable
         }
        else // both are functors
         {
          func1 = iter1.CurrSubterm()->Head().Func();     
          func2 = iter2.CurrSubterm()->Head().Func(); 
          if (func1 != func2) return (func1.CompareFunctor(func2) == GREATER); 
         }; 
       iter1.SkipSym();
       iter2.SkipSym();
      }      
     while (!iter1.Empty());
     return false; // equal
    }; // bool LexGreater() 

  public:
   inline bool Greater()
    {
     difference.Substitute(subst_diff); 
     if (subst_diff.Zero()) return LexGreater();
     return subst_diff.Positive();
    };
   inline bool LessOrEqual() 
    {
     difference.Substitute(subst_diff); 
     if (subst_diff.Zero()) return LexLessOrEq();
     return subst_diff.Negative();
    }; 
   
  public: // for debugging
   inline void Output()
    {
     OutputPrefix(term1); ::Output(" > "); ::OutputPrefix(term2); ::Output("\n");
     ::Output("DIFF.FORMULA:   "); difference.Output(); ::Output("\n");
     ::Output("COMPUTED DIFF.: "); subst_diff.Output(); ::Output("\n");
    }; 
   inline void OutputWithSubst()
    {
     OutputPrefix(term1); ::Output(" > "); ::OutputPrefix(term2); ::Output("\n");
     ::Output("DIFF.FORMULA:   "); difference.OutputWithSubst(); ::Output("\n");
     ::Output("COMPUTED DIFF.: "); subst_diff.Output(); ::Output("\n");
    };
 }; // class ORD_FORMULA


class ORD_CONSTR
 {
  private:
   ORD_FORMULA formula; 
   bool always_valid;
  public:
   ORD_CONSTR() : formula() {};
   inline void Compile(PREFIX_SYM* t1,PREFIX_SYM* t2) 
    { 
     always_valid = false;
     formula.Reset();
     formula.Compile(t1,t2);
    };
   inline void MakeAlwaysValid() { always_valid = true; };
   inline void Inverse() { formula.Inverse(); };
   inline bool Valid() { return (always_valid || (!formula.LessOrEqual())); };  
   inline bool Violated() { return ((!always_valid) && formula.LessOrEqual()); };
   inline bool Trivial() { return (always_valid || formula.Greater()); };
  public: // for debugging
   inline void Output()
    {
     if (always_valid) { ::Output("ALWAYS VALID.\n"); } else formula.OutputWithSubst(); 
    };  
 }; // class ORD_CONSTR






//***********************************************************************

class WEIGHT_FORMULA
 {
  // this class has heavy constructor, objects must be static!
  private:
   SMALL_VAR_COUNTER var_counter;
   long const_part;
   int index;
  public:
   WEIGHT_FORMULA() 
    : var_counter(),
      const_part(0),
      index(0)
   {};
   inline void SetIndex(int ind) { index = ind; var_counter.SetIndex(ind); };
   inline void Reset()
    {
     var_counter.Reset();
     const_part = 0;
    };  
   inline void RegVar(const VAR* v) { var_counter.Register(v); };
   inline void RegFunc() { const_part++; }; 
   inline long Compute() 
    {
     long res = const_part;
     var_counter.ResetIteration();
     for (const VAR* var = var_counter.NextRegistered();var;var = var_counter.NextRegistered())
      {
       bool free_var;
       void* bind = var->Unref(free_var);
       long w = (free_var)? 1 : PREFIX::Weight((PREFIX_SYM*)bind); 
       res += var_counter.Score(var)*w;
      };
     return res; 
    };
   inline bool GreaterThan(long x)
    {
     x -= const_part;
     if (x < 0) return true;
     var_counter.ResetIteration();
     for (const VAR* var = var_counter.NextRegistered();var;var = var_counter.NextRegistered())
      {
       bool free_var;
       void* bind = var->Unref(free_var);
       long w = (free_var)? 1 : PREFIX::Weight((PREFIX_SYM*)bind); 
       x -= var_counter.Score(var)*w;
       if (x < 0) return true;
      };
     return false; 
    }; 
   inline void Compile(TERM* t)
    {
     // t must be complex here
     TERM::ITERATOR iter;
     iter.Reset(t);
     do
      {
       if (iter.CurrentSymIsVar())
        {
         RegVar(VAR::Var(iter.Symbol().Var(),index));
        }
       else RegFunc();   
      }
     while (iter.Next()); 
    }; 
   inline void Compile(TERM t) 
    {
     if (t.IsVariable())
      {
       RegVar(VAR::Var(t.Var(),index));
      }
     else Compile(t.First()); 
    };
   inline void Compile(PREFIX_SYM* t)
    {
     if (t->Head().IsVariable()) 
      {
       CHECK18(t->Head().Var()->Index() == index,"A0 in WEIGHT_FORMULA::Compile(PREFIX_SYM* t)");
       RegVar(t->Head().Var()); 
       return; 
      };
     PREFIX::SIMPLE_ITER iter;
     iter.Reset(t);
     do
      {
       PREF_SYM sym = iter.CurrPos()->Head();
       if (sym.IsVariable())
        {
         CHECK18(sym.Var()->Index() == index,"A1 in WEIGHT_FORMULA::Compile(PREFIX_SYM* t)");
         RegVar(sym.Var());
        }
       else RegFunc(); 
       iter.SkipSym();
      }
     while (!iter.Empty()); 
    }; 
  public: // for debugging  
   inline void Output() 
    {
     var_counter.ResetIteration();
     ::Output(const_part);
     for (const VAR* var = var_counter.NextRegistered();var;var = var_counter.NextRegistered())
      {
       ::Output(" + ");
       ::Output((long)(var_counter.Score(var)));
       var->Output();
      };
    };
    
 }; // class WEIGHT_FORMULA



//************************************************************************
